/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Bank_Dummy.h"

#include "Bank_DRAM.h"			// DRAMBank::ValidAddress
#include "Bank_SRAM.h"			// SRAMBank::GetMemoryStart
#include "CPU_REG.h"			// Software::BusError
#include "EmPalmHeap.h"			// EmPalmHeap::GetHeapByPtr
#include "RAM_ROM.h"			// Memory::InitializeBanks


static Bool PrvLooksLikeA5Access(emuptr iAddress, long size, Bool forRead);


// ===========================================================================
//		 Dummy Bank Accessors
// ===========================================================================
// Dummy banks are non-existent blocks of memory.  Dummy bank accessors do
// not do anything.

static AddressBank	gAddressBank =
{
	DummyBank::GetLong,
	DummyBank::GetWord,
	DummyBank::GetByte,
	DummyBank::SetLong,
	DummyBank::SetWord,
	DummyBank::SetByte,
	DummyBank::GetRealAddress,
	DummyBank::ValidAddress,
	DummyBank::GetMetaAddress,
	DummyBank::AddOpcodeCycles
};


inline Bool HackForHwrGetRAMSize (emuptr address)
{
//	if ((address & 0xFF000000) == SRAMBank::GetMemoryStart ())
	if (address == SRAMBank::GetMemoryStart () + gRAMBank_Size)
		return true;

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	DummyBank::Initialize
 *
 * DESCRIPTION:	Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void DummyBank::Initialize (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	DummyBank::Reset
 *
 * DESCRIPTION:	Standard reset function.  Sets the sub-system to a
 *				default state.  This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void DummyBank::Reset (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	DummyBank::Save
 *
 * DESCRIPTION:	Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void DummyBank::Save (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	DummyBank::Load
 *
 * DESCRIPTION:	Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void DummyBank::Load (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	DummyBank::Dispose
 *
 * DESCRIPTION:	Standard dispose function.  Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void DummyBank::Dispose (void)
{
}


/***********************************************************************
 *
 * FUNCTION:    DummyBank::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void DummyBank::SetBankHandlers (void)
{
	Memory::InitializeBanks (gAddressBank, 0, 0xFFFF);
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetLong
// ---------------------------------------------------------------------------

uint32 DummyBank::GetLong (emuptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (long)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (long), true))
			InvalidAccess (iAddress, sizeof (long), true);
	}
#endif

	InvalidAccess (iAddress, sizeof (long), true);
	return ~0;
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetWord
// ---------------------------------------------------------------------------

uint32 DummyBank::GetWord (emuptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (short)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (short), true))
			InvalidAccess (iAddress, sizeof (short), true);
	}
#endif

	InvalidAccess (iAddress, sizeof (short), true);
	return ~0;
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetByte
// ---------------------------------------------------------------------------

uint32 DummyBank::GetByte (emuptr iAddress)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return 0;

#if (VALIDATE_DUMMY_GET)
	if (gMemAccessFlags.fValidate_DummyGet && !ValidAddress (iAddress, sizeof (char)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (short), true))
			InvalidAccess (iAddress, sizeof (char), true);
	}
#endif

	InvalidAccess (iAddress, sizeof (char), true);
	return ~0;
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetLong
// ---------------------------------------------------------------------------

void DummyBank::SetLong (emuptr iAddress, uint32)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (long)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (long), false))
			InvalidAccess (iAddress, sizeof (long), false);
	}
#endif

	InvalidAccess (iAddress, sizeof (long), true);
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetWord
// ---------------------------------------------------------------------------

void DummyBank::SetWord (emuptr iAddress, uint32)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (short)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (short), false))
			InvalidAccess (iAddress, sizeof (short), false);
	}
#endif

	InvalidAccess (iAddress, sizeof (short), true);
}


// ---------------------------------------------------------------------------
//		 DummyBank::SetByte
// ---------------------------------------------------------------------------

void DummyBank::SetByte (emuptr iAddress, uint32)
{
	// Hack to keep HwrGetRAMSize working: it runs off
	// the end of RAM while testing it.

	if (HackForHwrGetRAMSize (iAddress))
		return;

#if (VALIDATE_DUMMY_SET)
	if (gMemAccessFlags.fValidate_DummySet && !ValidAddress (iAddress, sizeof (char)))
	{
		if (!PrvLooksLikeA5Access (iAddress, sizeof (char), false))
			InvalidAccess (iAddress, sizeof (char), false);
	}
#endif

	InvalidAccess (iAddress, sizeof (char), true);
}


// ---------------------------------------------------------------------------
//		 DummyBank::ValidAddress
// ---------------------------------------------------------------------------

int DummyBank::ValidAddress (emuptr, uint32)
{
	int	result = false;

	return result;
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetRealAddress
// ---------------------------------------------------------------------------

uint8* DummyBank::GetRealAddress (emuptr iAddress)
{
	return (uint8*) iAddress;
}


// ---------------------------------------------------------------------------
//		 DummyBank::GetMetaAddress
// ---------------------------------------------------------------------------

uint8* DummyBank::GetMetaAddress (emuptr iAddress)
{
	UNUSED_PARAM(iAddress)

	static uint8	dummyBits[4] = {0, 0, 0, 0};

	return dummyBits;
}


// ---------------------------------------------------------------------------
//		 DummyBank::AddOpcodeCycles
// ---------------------------------------------------------------------------

void DummyBank::AddOpcodeCycles (void)
{
}


// ---------------------------------------------------------------------------
//		 DummyBank::InvalidAccess
// ---------------------------------------------------------------------------

void DummyBank::InvalidAccess (emuptr iAddress, long size, Bool forRead)
{
	UNUSED_PARAM(iAddress)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 PrvLooksLikeA5Access
// ---------------------------------------------------------------------------

Bool PrvLooksLikeA5Access(emuptr iAddress, long size, Bool forRead)
{
	// The OS sets the high bit of the A5 register when calling PilotMain
	// with a launch code that doesn't allow global variable access.

	if ((iAddress & 0x80000000) != 0 &&
		(m68k_areg (regs, 5) & 0x80000000) != 0)
	{
		emuptr	strippedAddress = iAddress & 0x7FFFFFFF;
		emuptr	strippedA5 = m68k_areg (regs, 5) & 0x7FFFFFFF;

		// See if the stripped test address and the real A5 both reside in
		// the dynamic heap.

		if (DRAMBank::ValidAddress (strippedA5, 1) &&
			DRAMBank::ValidAddress (strippedAddress, size))
		{
			// Now see if they both point into the same chunk.

			const EmPalmHeap*	heap1 = EmPalmHeap::GetHeapByPtr (strippedA5);
			const EmPalmHeap*	heap2 = EmPalmHeap::GetHeapByPtr (strippedAddress);

			if (heap1 != heap2)
				return false;

			const EmPalmChunk*	chunk1 = heap1->GetChunkBodyContaining (strippedA5);
			const EmPalmChunk*	chunk2 = heap1->GetChunkBodyContaining (strippedAddress);

			if ((chunk1 != NULL) && (chunk1 == chunk2))
			{
				// All tests pass, so report an attempt to utilize A5
				// in a verbotten situation.

				int button = Errors::ReportA5Access (iAddress, size, forRead);
				Emulator::HandleDlgButton (button, m68k_getpc ());
				return true;	// (I don't think we ever get here...)
			}
		}
	}

	return false;
}
