/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Bank_Mapped.h"

#include "CPU_REG.h"			// Software::BusError
#include "RAM_ROM.h"			// Memory::InitializeBanks

#include <vector>


// ===========================================================================
//		 Dummy Bank Accessors
// ===========================================================================
// Dummy banks are non-existent blocks of memory.  Dummy bank accessors do
// not do anything.

static AddressBank	gAddressBank =
{
	MappedBank::GetLong,
	MappedBank::GetWord,
	MappedBank::GetByte,
	MappedBank::SetLong,
	MappedBank::SetWord,
	MappedBank::SetByte,
	MappedBank::GetRealAddress,
	MappedBank::ValidAddress,
	MappedBank::GetMetaAddress,
	MappedBank::AddOpcodeCycles
};

struct MapRange
{
	Bool		Contains (const void* addr)
	{
		const char*	begin = (const char*) this->realAddress;
		const char*	end = begin + this->size;

		return ((addr >= begin) && (addr < end));
	}

	Bool		Contains (emuptr addr)
	{
		emuptr	begin = this->mappedAddress;
		emuptr	end = begin + this->size;

		return ((addr >= begin) && (addr < end));
	}

	const void*	realAddress;		// Address in host's space
	emuptr		mappedAddress;		// Address that emulated code sees
	uint32		size;
};
typedef vector<MapRange>		MapRangeList;

static MapRangeList				gMappedRanges;
static MapRangeList::iterator	gLastIter;

const emuptr					kMemoryStart = 0x60000000;

static MapRangeList::iterator	PrvGetMappingInfo (const void* addr);
static MapRangeList::iterator	PrvGetMappingInfo (emuptr addr);
static emuptr					PrvEnsureAligned (emuptr candidate, const void* addr);
static void						PrvInvalidateCache (void);
static void						PrvCheckRanges (void);


/***********************************************************************
 *
 * FUNCTION:	MappedBank::Initialize
 *
 * DESCRIPTION:	Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void MappedBank::Initialize (void)
{
	gMappedRanges.clear ();
	::PrvInvalidateCache ();
}


/***********************************************************************
 *
 * FUNCTION:	MappedBank::Reset
 *
 * DESCRIPTION:	Standard reset function.  Sets the sub-system to a
 *				default state.  This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void MappedBank::Reset (void)
{
	// Note: I used to clear out the mapped ranges in this function.
	// However, this didn't work out too well.  While manually resetting
	// the device (as occurs at the end of a debugging session), there
	// would be several StMemoryMapper objects instantiated.  After the
	// reset process, these StMemoryMappers would try to destruct
	// themselves, calling MappedBank::UnmapPhysicalMemory in the
	// process.  But since the mapped ranges had been cleared, that
	// function would throw an assert.  So....I can't think of a reason
	// for clearing out the ranges en masse,  so let's no do that anymore.
}


/***********************************************************************
 *
 * FUNCTION:	MappedBank::Save
 *
 * DESCRIPTION:	Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void MappedBank::Save (SessionFile&)
{
	// The only state we have to save is in gMappedRanges.  We shouldn't
	// have to actually save this because there shouldn't be anything in
	// there at the moment when we save files.  We map in memory ranges
	// at the following times:
	//
	//		* Calling system functions
	//		* Loading .prc, etc., files
	//		* Mapping in environment strings
	//
	// The first two should not be in "effect" at the time we save a file.
	// The last one we'll leave up to the HostControl system to re-establish.
}


/***********************************************************************
 *
 * FUNCTION:	MappedBank::Load
 *
 * DESCRIPTION:	Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void MappedBank::Load (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	MappedBank::Dispose
 *
 * DESCRIPTION:	Standard dispose function.  Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void MappedBank::Dispose (void)
{
	gMappedRanges.clear ();
	::PrvInvalidateCache ();
}


/***********************************************************************
 *
 * FUNCTION:    MappedBank::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void MappedBank::SetBankHandlers (void)
{
	Memory::InitializeBanks (gAddressBank, bankindex (kMemoryStart), 0x1000);
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetLong
// ---------------------------------------------------------------------------

uint32 MappedBank::GetLong (emuptr iAddress)
{
#if HAS_PROFILING
	CYCLE_GETLONG (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return ~0;
	}

	return (((uint32) p[0]) << 24) | (((uint32) p[1]) << 16) | (((uint32) p[2]) << 8) | p[3];
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetWord
// ---------------------------------------------------------------------------

uint32 MappedBank::GetWord (emuptr iAddress)
{
#if HAS_PROFILING
	CYCLE_GETWORD (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return ~0;
	}

	return (((uint32) p[0]) << 8) | p[1];
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetByte
// ---------------------------------------------------------------------------

uint32 MappedBank::GetByte (emuptr iAddress)
{
#if HAS_PROFILING
	CYCLE_GETBYTE (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return ~0;
	}

	return p[0];
}


// ---------------------------------------------------------------------------
//		 MappedBank::SetLong
// ---------------------------------------------------------------------------

void MappedBank::SetLong (emuptr iAddress, uint32 iLongValue)
{
#if HAS_PROFILING
	CYCLE_PUTLONG (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return;
	}

	p[0] = (uint8) (iLongValue >> 24);
	p[1] = (uint8) (iLongValue >> 16);
	p[2] = (uint8) (iLongValue >> 8);
	p[3] = (uint8) (iLongValue >> 0);
}


// ---------------------------------------------------------------------------
//		 MappedBank::SetWord
// ---------------------------------------------------------------------------

void MappedBank::SetWord (emuptr iAddress, uint32 iWordValue)
{
#if HAS_PROFILING
	CYCLE_PUTWORD (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return;
	}

	p[0] = (uint8) (iWordValue >> 8);
	p[1] = (uint8) (iWordValue >> 0);
}


// ---------------------------------------------------------------------------
//		 MappedBank::SetByte
// ---------------------------------------------------------------------------

void MappedBank::SetByte (emuptr iAddress, uint32 iByteValue)
{
#if HAS_PROFILING
	CYCLE_PUTBYTE (WAITSTATES_DUMMYBANK);
#endif

	uint8*	p = GetRealAddress (iAddress);

	if (p == NULL)
	{
		return;
	}

	p[0] = (uint8) (iByteValue >> 0);
}


// ---------------------------------------------------------------------------
//		 MappedBank::ValidAddress
// ---------------------------------------------------------------------------

int MappedBank::ValidAddress (emuptr iAddress, uint32)
{
	uint8*	realAddress = GetRealAddress (iAddress);
	int		result = realAddress != NULL;

	return result;
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetRealAddress
// ---------------------------------------------------------------------------

uint8* MappedBank::GetRealAddress (emuptr iAddress)
{
	MapRangeList::iterator	iter = ::PrvGetMappingInfo (iAddress);

	if (iter == gMappedRanges.end ())
		return NULL;

	return ((uint8*) iter->realAddress) + (iAddress - iter->mappedAddress);
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetMetaAddress
// ---------------------------------------------------------------------------

uint8* MappedBank::GetMetaAddress (emuptr iAddress)
{
	UNUSED_PARAM(iAddress)

	static uint8	dummyBits[4] = {0, 0, 0, 0};

	return dummyBits;
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetEmulatedAddress
// ---------------------------------------------------------------------------

emuptr MappedBank::GetEmulatedAddress (const void* iAddress)
{
	MapRangeList::iterator	iter = ::PrvGetMappingInfo (iAddress);

	if (iter == gMappedRanges.end ())
		return UAE_NULL;

	return iter->mappedAddress + ((char*) iAddress - (char*) iter->realAddress);
}


// ---------------------------------------------------------------------------
//		 MappedBank::AddOpcodeCycles
// ---------------------------------------------------------------------------

void MappedBank::AddOpcodeCycles (void)
{
}


// ---------------------------------------------------------------------------
//		 MappedBank::InvalidAccess
// ---------------------------------------------------------------------------

void MappedBank::InvalidAccess (emuptr iAddress, long size, Bool forRead)
{
	UNUSED_PARAM(iAddress)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 MappedBank::MapPhysicalMemory
// ---------------------------------------------------------------------------
// Maps a range of physical memory to appear at the same location of the
// emulated Palm OS's virtual memory.

void MappedBank::MapPhysicalMemory (const void* addr, uint32 size)
{
	if (addr == NULL)
		return;

	emuptr	candidate = ::PrvEnsureAligned (kMemoryStart, addr);

	// If the list is empty, add the item

	if (gMappedRanges.size () == 0)
	{
		MapRange	range;

		range.realAddress	= addr;
		range.mappedAddress	= candidate;
		range.size			= size;

		gLastIter = gMappedRanges.insert (gMappedRanges.begin (), range);
		::PrvCheckRanges ();

//		::PrvInvalidateCache ();

		return;
	}

	// Find an available space for the range.

	MapRangeList::iterator	iter	= gMappedRanges.begin ();

	while (iter != gMappedRanges.end ())
	{
		// If there's room in front of the current block...

		if (iter->mappedAddress >= candidate + size)
		{
			// ...insert a map record in front of the current block.

			MapRange	range;

			range.realAddress	= addr;
			range.mappedAddress	= candidate;
			range.size			= size;

			gLastIter = gMappedRanges.insert (iter, range);
			::PrvCheckRanges ();

//			::PrvInvalidateCache ();

			return;
		}

		// No room at the inn...move onto the next block.

		candidate =  ::PrvEnsureAligned (iter->mappedAddress + iter->size, addr);
		++iter;
	}

	// Couldn't find room at the beginning or the middle of the list. Add to the end.

	MapRangeList::iterator	last = gMappedRanges.end ();
	iter = last - 1;

	MapRange	range;

	range.realAddress	= addr;
	range.mappedAddress	= ::PrvEnsureAligned (iter->mappedAddress + iter->size, addr);
	range.size			= size;

	gLastIter = gMappedRanges.insert (last, range);
	::PrvCheckRanges ();

//	::PrvInvalidateCache ();
}


// ---------------------------------------------------------------------------
//		 MappedBank::UnmapPhysicalMemory
// ---------------------------------------------------------------------------
// Unmaps a range of physical memory from appearing at the same location of
// the emulated Palm OS's virtual memory.

void MappedBank::UnmapPhysicalMemory (const void* addr)
{
	if (addr == NULL)
		return;

	MapRangeList::iterator	iter = ::PrvGetMappingInfo (addr);

	assert (iter != gMappedRanges.end ());

	if (iter != gMappedRanges.end ())
	{
		gMappedRanges.erase (iter);
		::PrvCheckRanges ();

		::PrvInvalidateCache ();
	}
}


// ---------------------------------------------------------------------------
//		 MappedBank::GetMappingInfo
// ---------------------------------------------------------------------------

void MappedBank::GetMappingInfo	(emuptr addr, void** start, uint32* len)
{
	MapRangeList::iterator	iter = ::PrvGetMappingInfo (addr);

	if (iter != gMappedRanges.end ())
	{
		if (start)
			*start = (void*) iter->realAddress;

		if (len)
			*len = iter->size;
	}
	else
	{
		if (start)
			*start = NULL;

		if (len)
			*len = 0;
	}
}


// ---------------------------------------------------------------------------
//		 PrvGetMappingInfo
// ---------------------------------------------------------------------------

MapRangeList::iterator PrvGetMappingInfo (const void* addr)
{
	if (gLastIter != gMappedRanges.end () && gLastIter->Contains (addr))
		return gLastIter;

	MapRangeList::iterator	iter = gMappedRanges.begin ();
	while (iter != gMappedRanges.end ())
	{
		if (iter->Contains (addr))
			break;

		++iter;
	}

	gLastIter = iter;

	return iter;
}


MapRangeList::iterator PrvGetMappingInfo (emuptr addr)
{
	if (gLastIter != gMappedRanges.end () && gLastIter->Contains (addr))
		return gLastIter;

	MapRangeList::iterator	iter = gMappedRanges.begin ();
	while (iter != gMappedRanges.end ())
	{
		if (iter->Contains (addr))
			break;

		++iter;
	}

	gLastIter = iter;

	return iter;
}


emuptr PrvEnsureAligned (emuptr candidate, const void* addr)
{
	// Make sure that "candidate" -- which is the candidate
	// mapped address -- maintains the same alignment as
	// the incoming address.

	while ((candidate & 0x03) != ((uint32) addr & 0x03))
		++candidate;

	return candidate;
}

void PrvInvalidateCache (void)
{
	gLastIter = gMappedRanges.end ();
}

void PrvCheckRanges (void)
{
#if _DEBUG
	/*
		Check for overlaps.  We iterate over the container with
		two iterators: outerIter and innerIter.  These two iterators
		will allow us to compare all possible combinations of
		elements in the container.

		When checking for overlaps, there are six cases to consider:

					*-----------*			Segment A

			*---*							Possible segment B's.
			*----------*
			*---------------------------*
						*---*
						*---------------*
									*---*

		An overlap occurs if any part of segment B crosses segment A.
		From the drawing, we can see that the first and last possible
		segments B's are OK; the others overlap with segment A in
		some fashion.  Thus, Segment B is OK if it's end is less than
		Segment A's start, or its start is greater than segment A's end.
	*/

	MapRangeList::iterator	outerIter = gMappedRanges.begin ();
	while (outerIter != gMappedRanges.end ())
	{
		MapRangeList::iterator	innerIter = outerIter + 1;
		while (innerIter != gMappedRanges.end ())
		{
			char*	outerStart = (char*) outerIter->realAddress;
			char*	innerStart = (char*) innerIter->realAddress;

			char*	outerEnd = outerStart + outerIter->size;
			char*	innerEnd = innerStart + innerIter->size;

			if (innerEnd > outerStart && innerStart < outerEnd)
			{
				assert (false);
			}

			++innerIter;
		}

		++outerIter;
	}
#endif
}

