/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Bank_Regs.h"

#include "Bank_PLD.h"			// PLDBank::IsSerialPortOn
#include "Bank_ROM.h"			// ROMBank::IsPCInRAM
#include "CPU_REG.h"			// GetDevice
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "EmDevice.h"			// CreateRegs
#include "EmRegs.h"				// EmRegs
#include "RAM_ROM.h"			// gMemAccessFlags

EmRegs*	gRegs;


// ===========================================================================
//		 Register Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's register
// memory.

static AddressBank gAddressBank =
{
	RegsBank::GetLong,
	RegsBank::GetWord,
	RegsBank::GetByte,
	RegsBank::SetLong,
	RegsBank::SetWord,
	RegsBank::SetByte,
	RegsBank::GetRealAddress,
	RegsBank::ValidAddress,
	NULL,
	NULL
};


/***********************************************************************
 *
 * FUNCTION:	RegsBank::Initialize
 *
 * DESCRIPTION:	Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void RegsBank::Initialize (void)
{
	assert (!gRegs);
	gRegs = Emulator::GetDevice ().CreateRegs ();
	assert (gRegs);

	// Install the handlers for each register.

	gRegs->SetSubHandlers ();

	// Also reset here and now so that the hardware registers will be
	// set up.  We need them set up so that we know how to configure
	// our RAM from the chip selects.

	RegsBank::Reset ();
}


/***********************************************************************
 *
 * FUNCTION:	RegsBank::Reset
 *
 * DESCRIPTION:	Standard reset function.  Sets the sub-system to a
 *				default state.  This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void RegsBank::Reset (void)
{
	assert (gRegs);
	gRegs->Reset ();
}


/***********************************************************************
 *
 * FUNCTION:	RegsBank::Save
 *
 * DESCRIPTION:	Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void RegsBank::Save (SessionFile& f)
{
	assert (gRegs);
	gRegs->Save (f);
}


/***********************************************************************
 *
 * FUNCTION:	RegsBank::Load
 *
 * DESCRIPTION:	Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void RegsBank::Load (SessionFile& f)
{
	assert (gRegs);
	gRegs->Load (f);
}


/***********************************************************************
 *
 * FUNCTION:	RegsBank::Dispose
 *
 * DESCRIPTION:	Standard dispose function.  Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void RegsBank::Dispose (void)
{
	if (gRegs)
	{
		gRegs->Dispose ();
		delete gRegs;
		gRegs = NULL;
	}
}


/***********************************************************************
 *
 * FUNCTION:    RegsBank::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void RegsBank::SetBankHandlers (void)
{
	// Memory bank 0xFFFF is managed by the functions in RegsBank.

	Memory::InitializeBanks (gAddressBank, bankindex (kMemoryStart), 1);
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLong
// ---------------------------------------------------------------------------

uint32 RegsBank::GetLong (emuptr address)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGLongRead]++;
	if (address & 2)
		gMemoryAccess[kREGLongRead2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((address & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 4, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (address, 4))
	{
		InvalidAccess (address, 4, true);
	}
#endif

	assert (gRegs);
	return gRegs->GetLong (address);
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetWord
// ---------------------------------------------------------------------------

uint32 RegsBank::GetWord (emuptr address)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGWordRead]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((address & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 2, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (address, 2))
	{
		InvalidAccess (address, 2, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETWORD (WAITSTATES_REGISTERS);
#endif

	assert (gRegs);
	return gRegs->GetWord (address);
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetByte
// ---------------------------------------------------------------------------

uint32 RegsBank::GetByte (emuptr address)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGByteRead]++;
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 1, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (address, 1))
	{
		InvalidAccess (address, 1, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETBYTE (WAITSTATES_REGISTERS);
#endif

	assert (gRegs);
	return gRegs->GetByte (address);
}


// ---------------------------------------------------------------------------
//		 RegsBank::SetLong
// ---------------------------------------------------------------------------

void RegsBank::SetLong (emuptr address, uint32 value)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGLongWrite]++;
	if (address & 2)
		gMemoryAccess[kREGLongWrite2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((address & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 4, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (address, 4))
	{
		InvalidAccess (address, 4, false);
	}
#endif

	assert (gRegs);
	gRegs->SetLong (address, value);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, 4);
}


// ---------------------------------------------------------------------------
//		 RegsBank::SetWord
// ---------------------------------------------------------------------------

void RegsBank::SetWord (emuptr address, uint32 value)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGWordWrite]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((address & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 2, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (address, 2))
	{
		InvalidAccess (address, 2, false);
	}
#endif

	assert (gRegs);
	gRegs->SetWord (address, value);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, 2);
}


// ---------------------------------------------------------------------------
//		 RegsBank::SetByte
// ---------------------------------------------------------------------------

void RegsBank::SetByte (emuptr address, uint32 value)
{
#if PROFILE_MEMORY
	gMemoryAccess[kREGByteWrite]++;
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (address, 1, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (address, 1))
	{
		InvalidAccess (address, 1, false);
	}
#endif

	assert (gRegs);
	gRegs->SetByte (address, value);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, 1);
}


// ---------------------------------------------------------------------------
//		 RegsBank::ValidAddress
// ---------------------------------------------------------------------------

int RegsBank::ValidAddress (emuptr address, uint32 size)
{
	assert (gRegs);
	return gRegs->ValidAddress (address, size);;
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetRealAddress
// ---------------------------------------------------------------------------

uint8* RegsBank::GetRealAddress (emuptr address)
{
	assert (gRegs);
	return gRegs->GetRealAddress (address);
}


// ---------------------------------------------------------------------------
//		 RegsBank::InvalidAccess
// ---------------------------------------------------------------------------

void RegsBank::InvalidAccess (emuptr address, long size, Bool forRead)
{
	UNUSED_PARAM(address)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::PreventedAccess
// ---------------------------------------------------------------------------

void RegsBank::PreventedAccess (emuptr address, long size, Bool forRead)
{
	Memory::PreventedAccess (address, size, forRead, Errors::kRegisterAccess);
}


// ---------------------------------------------------------------------------
//		 RegsBank::Cycle
// ---------------------------------------------------------------------------
// Handles periodic events that need to occur when the processor cycles (like
// updating timer registers).  This function is called in two places from
// Emulator::Execute.  Interestingly, the loop runs 3% FASTER if this function
// is in its own separate function instead of being inline.

void RegsBank::Cycle (Bool sleeping)
{
	assert (gRegs);
	gRegs->Cycle (sleeping);
}


// ---------------------------------------------------------------------------
//		 RegsBank::CycleSlowly
// ---------------------------------------------------------------------------
// Handles periodic events that need to occur when the processor cycles (like
// updating timer registers).  This function is called in two places from
// Emulator::Execute.  Interestingly, the loop runs 3% FASTER if this function
// is in its own separate function instead of being inline.

void RegsBank::CycleSlowly (Bool sleeping)
{
	assert (gRegs);
	gRegs->CycleSlowly (sleeping);
}


// ---------------------------------------------------------------------------
//		 RegsBank::ButtonEvent
// ---------------------------------------------------------------------------
// Handles a Palm device button event by updating the appropriate registers.

void RegsBank::ButtonEvent (Bool buttonIsDown, int32 button)
{
	assert (gRegs);
	gRegs->ButtonEvent (buttonIsDown, button);
}


// ---------------------------------------------------------------------------
//		 RegsBank::HotSyncEvent
// ---------------------------------------------------------------------------
// Handles a HotSync button event by updating the appropriate registers.

void RegsBank::HotSyncEvent (Bool buttonIsDown)
{
	assert (gRegs);
	gRegs->HotSyncEvent (buttonIsDown);
}


// ---------------------------------------------------------------------------
//		 RegsBank::TurnSoundOff
// ---------------------------------------------------------------------------

void RegsBank::TurnSoundOff (void)
{
	assert (gRegs);
	gRegs->TurnSoundOff ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::ResetTimer
// ---------------------------------------------------------------------------

void RegsBank::ResetTimer (void)
{
	assert (gRegs);
	gRegs->ResetTimer ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::ResetRTC
// ---------------------------------------------------------------------------

void RegsBank::ResetRTC (void)
{
	assert (gRegs);
	gRegs->ResetRTC ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetInterruptLevel
// ---------------------------------------------------------------------------

int32 RegsBank::GetInterruptLevel (void)
{
	assert (gRegs);
	return gRegs->GetInterruptLevel ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetInterruptBase
// ---------------------------------------------------------------------------

int32 RegsBank::GetInterruptBase (void)
{
	assert (gRegs);
	return gRegs->GetInterruptBase ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDDepth
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDDepth (void)
{
	assert (gRegs);
	return gRegs->GetLCDDepth ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDRowBytes
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDRowBytes (void)
{
	assert (gRegs);
	return gRegs->GetLCDRowBytes ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDWidth
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDWidth (void)
{
	assert (gRegs);
	return gRegs->GetLCDWidth ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDHeight
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDHeight (void)
{
	assert (gRegs);
	return gRegs->GetLCDHeight ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDStartAddr
// ---------------------------------------------------------------------------

emuptr RegsBank::GetLCDStartAddr (void)
{
	assert (gRegs);
	return gRegs->GetLCDStartAddr ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDPalette
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDPalette (RGBType* scrPalette)
{
	assert (gRegs);
	return gRegs->GetLCDPalette (scrPalette);
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDOffset
// ---------------------------------------------------------------------------

int32 RegsBank::GetLCDOffset (void)
{
	assert (gRegs);
	return gRegs->GetLCDOffset ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetDynamicHeapSize
// ---------------------------------------------------------------------------

int32 RegsBank::GetDynamicHeapSize (void)
{
	assert (gRegs);
	return gRegs->GetDynamicHeapSize ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetROMSize
// ---------------------------------------------------------------------------

int32 RegsBank::GetROMSize (void)
{
	assert (gRegs);
	return gRegs->GetROMSize ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetROMBaseAddress
// ---------------------------------------------------------------------------

emuptr RegsBank::GetROMBaseAddress (void)
{
	assert (gRegs);
	return gRegs->GetROMBaseAddress ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::ChipSelectsConfigured
// ---------------------------------------------------------------------------

Bool RegsBank::ChipSelectsConfigured (void)
{
	assert (gRegs);
	return gRegs->ChipSelectsConfigured ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetSystemClockFrequency
// ---------------------------------------------------------------------------

int32 RegsBank::GetSystemClockFrequency (void)
{
	assert (gRegs);
	return gRegs->GetSystemClockFrequency ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetLCDOn
// ---------------------------------------------------------------------------

Bool RegsBank::GetLCDOn (void)
{
	assert (gRegs);
	return gRegs->GetLCDOn ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetBacklightOn
// ---------------------------------------------------------------------------

Bool RegsBank::GetBacklightOn (void)
{
	assert (gRegs);
	return gRegs->GetBacklightOn ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetSerialPortOn
// ---------------------------------------------------------------------------

Bool RegsBank::GetSerialPortOn (void)
{
	if (Emulator::GetDevice ().HasPLD ())
	{
		return PLDBank::IsSerialPortOn ();
	}

	assert (gRegs);
	return gRegs->GetSerialPortOn ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetCanStop
// ---------------------------------------------------------------------------

Bool RegsBank::GetCanStop (void)
{
	assert (gRegs);
	return gRegs->GetCanStop ();
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetPortInputValue
// ---------------------------------------------------------------------------

uint8 RegsBank::GetPortInputValue (int port)
{
	assert (gRegs);
	return gRegs->GetPortInputValue (port);
}


// ---------------------------------------------------------------------------
//		 RegsBank::GetPortInternalValue
// ---------------------------------------------------------------------------

uint8 RegsBank::GetPortInternalValue (int port)
{
	assert (gRegs);
	return gRegs->GetPortInternalValue (port);
}


// ---------------------------------------------------------------------------
//		 RegsBank::PortDataChanged
// ---------------------------------------------------------------------------

void RegsBank::PortDataChanged (int port, uint8 oldValue, uint8 newValue)
{
	assert (gRegs);
	gRegs->PortDataChanged (port, oldValue, newValue);
}
