/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Bank_SymbolASIC.h"

#include "Bank_ROM.h"			// ROMBank::IsPCInRAM
#include "Byteswapping.h"		// ByteswapWords
#include "CPU_REG.h"			// Software::BusError
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "Logging.h"			// LogAppendMsg
#include "Profiling.h"			// WAITSTATES_SymbolASIC
#include "RAM_ROM.h"			// gMemoryAccess
#include "SessionFile.h"		// WriteSymbolASICRegsType, ReadSymbolASICRegsType, etc.

typedef uint32 	(*ReadFunction)(emuptr address, int size);
typedef void	(*WriteFunction)(emuptr address, int size, uint32 value);

// Macro to return the Dragonball address of the specified register

#define addressof(x)	(kMemoryStart + offsetof(HwrSymbolASICType, x))


// Macros for reading/writing Dragonball registers.

#define READ_REGISTER(reg)	\
	((sizeof(gSymbolASICRegs.reg) == 1) ? do_get_mem_byte(&gSymbolASICRegs.reg) :	\
	 (sizeof(gSymbolASICRegs.reg) == 2) ? do_get_mem_word(&gSymbolASICRegs.reg) :	\
								  do_get_mem_long(&gSymbolASICRegs.reg))

#define WRITE_REGISTER(reg, value)	\
	(sizeof(gSymbolASICRegs.reg) == 1) ? do_put_mem_byte(&gSymbolASICRegs.reg, value) : \
	(sizeof(gSymbolASICRegs.reg) == 2) ? do_put_mem_word(&gSymbolASICRegs.reg, value) : \
								 do_put_mem_long(&gSymbolASICRegs.reg, value)

#define hwrSymbolASICBase		0x11000000
#define WAITSTATES_SymbolASIC	0

/*
	// Bits describing GPIOData (0xFD) and GPIODirection (0xFF)

		GPIO 7	Volume Bit 2			Output	High	Input mode, '0'
		GPIO 6	Volume Bit 1			Output	High	Input mode, '0'
		GPIO 5	Volume Bit 0			Output	High	Input mode, '0'
		GPIO 4	External +5V Enable		Output	High	Input mode, '0'
		GPIO 3	Scanner Trigger			Output	Low		Input mode, '0'
		GPIO 2	Charge Enable			Output	High	Input mode, '0'
		GPIO 1	LCD Contrast Enable		Output	Low		Input mode, '0'
		GPIO 0	Cradle Power Indicator	Input	High	Input mode, '0'

		GPIO Direction Control:
		Defaults to zero.
		Resets to Zero.
		Writing a '0' - Read only, Input
		Writing a '1' - Read or Write. Output

	Volume Control Bits : Five different volume levels can
	be achieved by using these bits. They are as follows:

		Function				GPIO 7	GPIO 6	GPIO 5
		OFF						0		0		0
		LOW						0		0		1
		MED						0		1		0
		MED-HIGH				0		1		1
		HIGH					1		0		0
		Invalid (Do not use)	1		0		1
		Invalid (Do not use)	1		1		0
		Invalid (Do not use)	1		1		1
*/

#define	hwrSymbolVolume2		0x80
#define hwrSymbolVolume1		0x40
#define hwrSymbolVolume0		0x20
#define hwrSymbolExtVEnable		0x10
#define hwrSymbolScannerTrigger	0x08
#define hwrSymbolChargeEnable	0x04
#define hwrSymbolLCDContrast	0x02
#define hwrSymbolCradlePower	0x01

/*
	// Bits describing the scanner control registers (0xA5)

		Bit #	Direction	Function					Default
		0		Output		Enable of TXRDY as IRQOUT.	0
		1		Output		Enable of TXEMP as IRQOUT	0
		2		Output		Enable of RXRDY as IRQOUT	0
		3		Output		Enable of BD as IRQOUT		0
		4		Output		Unused						0
		5		Output		Unused						0
		6		Output		Decoder LED Output			0
		7		Input		CTS Input					N/A
*/


class SymbolASICReg
{
	public:
		static void 		SetHandler			(ReadFunction, WriteFunction, uint32 start, int count);

		static uint32		UnsupportedRead 	(emuptr address, int size);
		static void 		UnsupportedWrite	(emuptr address, int size, uint32 value);

		static uint32		StdRead 			(emuptr address, int size);
		static void 		StdWrite			(emuptr address, int size, uint32 value);
		static void 		NullWrite			(emuptr address, int size, uint32 value); // For read-only registers
};


struct HwrSymbolASICType
{
	uint8	S24IO[0x40];				// 0x00-0x3F

	uint8	__filler1[0x40];

	uint8	S24Attribute[0x20];			// 0x80-0x9F
	uint8	UART8251MacroSelect[0x04];	// 0xA0-0xA3

	uint8	__filler2;

	uint8	ScannerDecoderControl;		// 0xA5

	uint8	__filler3[0x51];

	uint8	Control;					// 0xF7

	uint8	__filler4[0x05];

	uint8	GPIOData;					// 0xFD

	uint8	__filler5;

	uint8	GPIODirection;				// 0xFF
};

const HwrSymbolASICType kInitialSymbolASICRegisterValues =
{
	{0},	// uint8	S24IO[0x40];				// 0x00-0x3F
	{0},	// uint8	__filler1[0x40];
	{0},	// uint8	S24Attribute[0x20];			// 0x80-0x9F
	{0},	// uint8	UART8251MacroSelect[0x04];	// 0xA0-0xA3
	0,		// uint8	__filler2;
	0,		// uint8	ScannerDecoderControl;		// 0xA5
	{0},	// uint8	__filler3[0x51];
	0,		// uint8	Control;					// 0xF7
	{0},	// uint8	__filler4[0x05];
	0,		// uint8	GPIOData;					// 0xFD
	0,		// uint8	__filler5;
	0		// uint8	GPIODirection;				// 0xFF
};


static AddressBank	gAddressBank =
{
	SymbolASICBank::GetLong,
	SymbolASICBank::GetWord,
	SymbolASICBank::GetByte,
	SymbolASICBank::SetLong,
	SymbolASICBank::SetWord,
	SymbolASICBank::SetByte,
	SymbolASICBank::GetRealAddress,
	SymbolASICBank::ValidAddress,
	NULL,
	NULL
};

static const uint32	kMemoryStart = hwrSymbolASICBase;

static HwrSymbolASICType	gSymbolASICRegs;

// These functions provide fetch and store access to the
// emulator's register memory.

static ReadFunction 	gReadFunctions[sizeof(HwrSymbolASICType)];
static WriteFunction	gWriteFunctions[sizeof(HwrSymbolASICType)];


#pragma mark -


/***********************************************************************
 *
 * FUNCTION:	SymbolASICBank::Initialize
 *
 * DESCRIPTION: Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void SymbolASICBank::Initialize (void)
{
	// Install the handlers for each register.

	SymbolASICBank::SetSubHandlers ();
}


/***********************************************************************
 *
 * FUNCTION:	SymbolASICBank::Reset
 *
 * DESCRIPTION: Standard reset function.  Sets the sub-system to a
 *				default state.	This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void SymbolASICBank::Reset (void)
{
	gSymbolASICRegs = kInitialSymbolASICRegisterValues;

//	Canonical (gSymbolASICRegs);
	ByteswapWords (&gSymbolASICRegs, sizeof(gSymbolASICRegs));
}


/***********************************************************************
 *
 * FUNCTION:	SymbolASICBank::Save
 *
 * DESCRIPTION: Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void SymbolASICBank::Save (SessionFile&)
{
	StWordSwapper	swapper (&gSymbolASICRegs, sizeof(gSymbolASICRegs));
//	f.WriteSymbolASICRegsType (gSymbolASICRegs);
}


/***********************************************************************
 *
 * FUNCTION:	SymbolASICBank::Load
 *
 * DESCRIPTION: Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void SymbolASICBank::Load (SessionFile& f)
{
	// Read in the SymbolASIC registers, and then byteswap them.

//	if (f.ReadSymbolASICRegsType (gSymbolASICRegs))
//	{
//		ByteswapWords (&gSymbolASICRegs, sizeof(gSymbolASICRegs));
//	}
//	else
	{
		f.SetCanReload (false);
	}
}


/***********************************************************************
 *
 * FUNCTION:	SymbolASICBank::Dispose
 *
 * DESCRIPTION: Standard dispose function.	Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void SymbolASICBank::Dispose (void)
{
}


/***********************************************************************
 *
 * FUNCTION:    SymbolASICBank::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void SymbolASICBank::SetBankHandlers (void)
{
	// Memory banks starting at kMemoryStart are managed by the functions in SymbolASICBank.

	long	numBanks = (sizeof (HwrSymbolASICType) + 0x0FFFF) / 0x10000;
	Memory::InitializeBanks (gAddressBank, bankindex (kMemoryStart), numBanks);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::SetSubHandlers
// ---------------------------------------------------------------------------

void SymbolASICBank::SetSubHandlers (void)
{
	SymbolASICReg::SetHandler (SymbolASICReg::UnsupportedRead, SymbolASICReg::UnsupportedWrite,
						kMemoryStart, sizeof (HwrSymbolASICType));

	// Now add standard/specialized handers for the defined registers.

#define INSTALL_HANDLER(read, write, reg) \
	SymbolASICReg::SetHandler (SymbolASICReg::read, SymbolASICReg::write, \
							addressof (reg), sizeof (gSymbolASICRegs.reg))


	INSTALL_HANDLER (StdRead,			StdWrite,				S24IO);
	INSTALL_HANDLER (StdRead,			StdWrite,				S24Attribute);
	INSTALL_HANDLER (StdRead,			StdWrite,				UART8251MacroSelect);
	INSTALL_HANDLER (StdRead,			StdWrite,				ScannerDecoderControl);
	INSTALL_HANDLER (StdRead,			StdWrite,				Control);
	INSTALL_HANDLER (StdRead,			StdWrite,				GPIOData);
	INSTALL_HANDLER (StdRead,			StdWrite,				GPIODirection);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::GetLong
// ---------------------------------------------------------------------------

uint32 SymbolASICBank::GetLong (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICLongRead]++;
	if (iAddress & 2)
		gMemoryAccess[kSymbolASICLongRead2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 4, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 4))
	{
		InvalidAccess (iAddress, 4, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETLONG (WAITSTATES_SymbolASIC);
#endif

	ReadFunction	fn = SymbolASICReg::UnsupportedRead;

	assert (fn);
	return fn (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::GetWord
// ---------------------------------------------------------------------------

uint32 SymbolASICBank::GetWord (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICWordRead]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 2, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 2))
	{
		InvalidAccess (iAddress, 2, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETWORD (WAITSTATES_SymbolASIC);
#endif

	ReadFunction	fn = SymbolASICReg::UnsupportedRead;

	assert (fn);
	return fn (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::GetByte
// ---------------------------------------------------------------------------

uint32 SymbolASICBank::GetByte (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICByteRead]++;
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 1, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 1))
	{
		InvalidAccess (iAddress, 1, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETBYTE (WAITSTATES_SymbolASIC);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	ReadFunction	fn = SymbolASICReg::UnsupportedRead;

	if (offset < sizeof (HwrSymbolASICType))
	{
		fn = gReadFunctions[offset];
	}

	assert (fn);
	uint32	result = fn (iAddress, 1);

//	LogAppendMsg ("Reading %02X:%02X", (uint32) (uint8) offset, (uint32) (uint8) result);

	return result;
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::SetLong
// ---------------------------------------------------------------------------

void SymbolASICBank::SetLong (emuptr iAddress, uint32 iLongValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICLongWrite]++;
	if (iAddress & 2)
		gMemoryAccess[kSymbolASICLongWrite2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 4, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 4))
	{
		InvalidAccess (iAddress, 4, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTLONG (WAITSTATES_SymbolASIC);
#endif

	WriteFunction	fn = SymbolASICReg::UnsupportedWrite;

	assert (fn);
	fn (iAddress, 4, iLongValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::SetWord
// ---------------------------------------------------------------------------

void SymbolASICBank::SetWord (emuptr iAddress, uint32 iWordValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICWordWrite]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 2, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 2))
	{
		InvalidAccess (iAddress, 2, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTWORD (WAITSTATES_SymbolASIC);
#endif

	WriteFunction	fn = SymbolASICReg::UnsupportedWrite;

	assert (fn);
	fn (iAddress, 2, iWordValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::SetByte
// ---------------------------------------------------------------------------

void SymbolASICBank::SetByte (emuptr iAddress, uint32 iByteValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSymbolASICByteWrite]++;
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 1, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 1))
	{
		InvalidAccess (iAddress, 1, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTBYTE (WAITSTATES_SED1375);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = SymbolASICReg::UnsupportedWrite;

	if (offset < sizeof (HwrSymbolASICType))
	{
		fn = gWriteFunctions[offset];
	}

//	LogAppendMsg ("Writing %02X:%02X", (uint32) (uint8) offset, (uint32) (uint8) iByteValue);

	assert (fn);
	fn (iAddress, 1, iByteValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 1);
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::ValidAddress
// ---------------------------------------------------------------------------

int SymbolASICBank::ValidAddress (emuptr iAddress, uint32 iSize)
{
	UNUSED_PARAM(iSize)

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = SymbolASICReg::UnsupportedWrite;

	if (offset < sizeof (HwrSymbolASICType))
	{
		fn = gWriteFunctions[offset];
	}

	int result = (fn != SymbolASICReg::UnsupportedWrite);

	return result;
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::GetRealAddress
// ---------------------------------------------------------------------------

uint8* SymbolASICBank::GetRealAddress (emuptr iAddress)
{
	uint8*	loc = ((uint8*) &gSymbolASICRegs) + (iAddress - kMemoryStart);

	return loc;
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::InvalidAccess
// ---------------------------------------------------------------------------

void SymbolASICBank::InvalidAccess (emuptr iAddress, long size, Bool forRead)
{
	UNUSED_PARAM(iAddress)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 SymbolASICBank::PreventedAccess
// ---------------------------------------------------------------------------

void SymbolASICBank::PreventedAccess (emuptr iAddress, long size, Bool forRead)
{
	Memory::PreventedAccess (iAddress, size, forRead, Errors::kRegisterAccess);
}


#pragma mark -

// ===========================================================================
//		 Register Accessors
// ===========================================================================

// ---------------------------------------------------------------------------
//		 SymbolASICReg::SetHandler
// ---------------------------------------------------------------------------

void SymbolASICReg::SetHandler (ReadFunction read, WriteFunction write,
							uint32 start, int count)
{
	int index = start - kMemoryStart;
	for (int ii = 0; ii < count; ++ii, ++index)
	{
		gReadFunctions[index] = read;
		gWriteFunctions[index] = write;
	}
}


// ---------------------------------------------------------------------------
//		 SymbolASICReg::UnsupportedRead
// ---------------------------------------------------------------------------

uint32 SymbolASICReg::UnsupportedRead (emuptr address, int size)
{
	if (!CEnableFullAccess::AccessOK ())
	{
		SymbolASICBank::InvalidAccess (address, size, true);
	}

	return ~0;
}


// ---------------------------------------------------------------------------
//		 SymbolASICReg::UnsupportedWrite
// ---------------------------------------------------------------------------

void SymbolASICReg::UnsupportedWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(value)

	if (!CEnableFullAccess::AccessOK ())
	{
		SymbolASICBank::InvalidAccess (address, size, false);
	}
}


// ---------------------------------------------------------------------------
//		 SymbolASICReg::StdRead
// ---------------------------------------------------------------------------

uint32 SymbolASICReg::StdRead (emuptr address, int size)
{
	if (size == 1)
		return do_get_mem_byte (SymbolASICBank::GetRealAddress (address));

	if (size == 2)
		return do_get_mem_word (SymbolASICBank::GetRealAddress (address));

	return do_get_mem_long (SymbolASICBank::GetRealAddress (address));
}


// ---------------------------------------------------------------------------
//		 SymbolASICReg::StdWrite
// ---------------------------------------------------------------------------

void SymbolASICReg::StdWrite (emuptr address, int size, uint32 value)
{
	if (size == 1)
		do_put_mem_byte (SymbolASICBank::GetRealAddress (address), value);

	else if (size == 2)
		do_put_mem_word (SymbolASICBank::GetRealAddress (address), value);

	else
		do_put_mem_long (SymbolASICBank::GetRealAddress (address), value);
}


// ---------------------------------------------------------------------------
//		 SymbolASICReg::NullWrite
// ---------------------------------------------------------------------------

void SymbolASICReg::NullWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(address)
	UNUSED_PARAM(size)
	UNUSED_PARAM(value)

	// Do nothing (for read-only registers).
}
