/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Bank_USB.h"

#include "Bank_Regs.h"			// RegsBank::GetROMSize
#include "Bank_ROM.h"			// ROMBank::IsPCInRAM
#include "Byteswapping.h"		// ByteswapWords
#include "CPU_REG.h"			// Software::BusError
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "Profiling.h"			// WAITSTATES_USB
#include "RAM_ROM.h"			// gMemoryAccess
#include "SessionFile.h"		// WriteUSBRegsType, ReadUSBRegsType, etc.

#define WAITSTATES_USB 0

typedef uint32 	(*ReadFunction)(emuptr address, int size);
typedef void	(*WriteFunction)(emuptr address, int size, uint32 value);

// Macro to return the Dragonball address of the specified register

#define addressof(x)	(kMemoryStart + offsetof(UsbHwrType, x))


// Macros for reading/writing USB registers.

#define READ_REGISTER(reg)	\
	((sizeof(gUSBRegs.reg) == 1) ? do_get_mem_byte(&gUSBRegs.reg) :	\
	 (sizeof(gUSBRegs.reg) == 2) ? do_get_mem_word(&gUSBRegs.reg) :	\
								  do_get_mem_long(&gUSBRegs.reg))

#define WRITE_REGISTER(reg, value)	\
	(sizeof(gUSBRegs.reg) == 1) ? do_put_mem_byte(&gUSBRegs.reg, value) : \
	(sizeof(gUSBRegs.reg) == 2) ? do_put_mem_word(&gUSBRegs.reg, value) : \
								 do_put_mem_long(&gUSBRegs.reg, value)

class USBReg
{
	public:
		static void 		SetHandler			(ReadFunction, WriteFunction, uint32 start, int count);

		static uint32		UnsupportedRead 	(emuptr address, int size);
		static void 		UnsupportedWrite	(emuptr address, int size, uint32 value);

		static uint32		StdRead 			(emuptr address, int size);
		static void 		StdWrite			(emuptr address, int size, uint32 value);
		static void 		NullWrite			(emuptr address, int size, uint32 value); // For read-only registers
};


typedef struct
{
	UInt8		data;		// $000: Data Register
	UInt8		___filler0;
	UInt8		command;	// $002: Command Register
	UInt8		___filler1;
} UsbHwrType;

const UsbHwrType kInitialUSBRegisterValues =
{
	0x00,		// UInt8		data;		// $000: Data Register
	0x00,		// UInt8		___filler0;
	0x00,		// UInt8		command;	// $002: Command Register
	0x00,		// UInt8		___filler1;
};


static AddressBank	gAddressBank =
{
	USBBank::GetLong,
	USBBank::GetWord,
	USBBank::GetByte,
	USBBank::SetLong,
	USBBank::SetWord,
	USBBank::SetByte,
	USBBank::GetRealAddress,
	USBBank::ValidAddress,
	NULL,
	NULL
};

static uint32	kMemoryStart = 0;

static UsbHwrType	gUSBRegs;

// These functions provide fetch and store access to the
// emulator's register memory.

static ReadFunction 	gReadFunctions[sizeof(UsbHwrType)];
static WriteFunction	gWriteFunctions[sizeof(UsbHwrType)];


#pragma mark -


/***********************************************************************
 *
 * FUNCTION:	USBBank::Initialize
 *
 * DESCRIPTION: Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void USBBank::Initialize (void)
{
	// Install the handlers for each register.

	USBBank::SetSubHandlers ();
}


/***********************************************************************
 *
 * FUNCTION:	USBBank::Reset
 *
 * DESCRIPTION: Standard reset function.  Sets the sub-system to a
 *				default state.	This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void USBBank::Reset (void)
{
	gUSBRegs = kInitialUSBRegisterValues;

//	Canonical (gUSBRegs);
	ByteswapWords (&gUSBRegs, sizeof(gUSBRegs));
}


/***********************************************************************
 *
 * FUNCTION:	USBBank::Save
 *
 * DESCRIPTION: Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void USBBank::Save (SessionFile&)
{
#if 0
	StWordSwapper	swapper (&gUSBRegs, sizeof(gUSBRegs));
	f.WriteUSBRegsType (gUSBRegs);
#endif
}


/***********************************************************************
 *
 * FUNCTION:	USBBank::Load
 *
 * DESCRIPTION: Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void USBBank::Load (SessionFile&)
{
#if 0
	// Read in the USB registers, and then byteswap them.

	if (f.ReadUSBRegsType (gUSBRegs))
	{
		ByteswapWords (&gUSBRegs, sizeof(gUSBRegs));
	}
	else
	{
		f.SetCanReload (false);
	}
#endif
}


/***********************************************************************
 *
 * FUNCTION:	USBBank::Dispose
 *
 * DESCRIPTION: Standard dispose function.	Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void USBBank::Dispose (void)
{
}


/***********************************************************************
 *
 * FUNCTION:    USBBank::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void USBBank::SetBankHandlers (void)
{
	/* 
		Auto-detect USB hardware base address.  The USB hardware is at CSA1,
		which resides just after the ROM (CSA0) in memory.  
	*/

	kMemoryStart = ROMBank::GetMemoryStart() + RegsBank::GetROMSize (); 

	// Memory banks starting at kMemoryStart are managed by the functions in USBBank.

	long	numBanks = (sizeof (UsbHwrType) + 0x0FFFF) / 0x10000;
	Memory::InitializeBanks (gAddressBank, bankindex (kMemoryStart), numBanks);
}


// ---------------------------------------------------------------------------
//		 USBBank::SetSubHandlers
// ---------------------------------------------------------------------------

void USBBank::SetSubHandlers (void)
{
	USBReg::SetHandler (USBReg::UnsupportedRead, USBReg::UnsupportedWrite,
						kMemoryStart, sizeof (UsbHwrType));

	// Now add standard/specialized handers for the defined registers.

#define INSTALL_HANDLER(read, write, reg) \
	USBReg::SetHandler (USBReg::read, USBReg::write, \
							addressof (reg), sizeof (gUSBRegs.reg))


	INSTALL_HANDLER (StdRead,			StdWrite,				data);
	INSTALL_HANDLER (StdRead,			NullWrite,				___filler0);
	INSTALL_HANDLER (StdRead,			StdWrite,				command);
	INSTALL_HANDLER (StdRead,			NullWrite,				___filler1);
}


// ---------------------------------------------------------------------------
//		 USBBank::GetLong
// ---------------------------------------------------------------------------

uint32 USBBank::GetLong (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBLongRead]++;
	if (iAddress & 2)
		gMemoryAccess[kUSBLongRead2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 4, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 4))
	{
		InvalidAccess (iAddress, 4, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETLONG (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	ReadFunction	fn = USBReg::UnsupportedRead;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gReadFunctions[offset];
	}

	assert (fn);
	return fn (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 USBBank::GetWord
// ---------------------------------------------------------------------------

uint32 USBBank::GetWord (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBWordRead]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 2, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 2))
	{
		InvalidAccess (iAddress, 2, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETWORD (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	ReadFunction	fn = USBReg::UnsupportedRead;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gReadFunctions[offset];
	}

	assert (fn);
	return fn (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 USBBank::GetByte
// ---------------------------------------------------------------------------

uint32 USBBank::GetByte (emuptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBByteRead]++;
#endif

#if (PREVENT_USER_REGISTER_GET)
	if (gMemAccessFlags.fProtect_RegisterGet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 1, true);
	}
#endif

#if (VALIDATE_REGISTER_GET)
	if (gMemAccessFlags.fValidate_RegisterGet && !ValidAddress (iAddress, 1))
	{
		InvalidAccess (iAddress, 1, true);
	}
#endif

#if HAS_PROFILING
	CYCLE_GETBYTE (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	ReadFunction	fn = USBReg::UnsupportedRead;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gReadFunctions[offset];
	}

	assert (fn);
	return fn (iAddress, 1);
}


// ---------------------------------------------------------------------------
//		 USBBank::SetLong
// ---------------------------------------------------------------------------

void USBBank::SetLong (emuptr iAddress, uint32 iLongValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBLongWrite]++;
	if (iAddress & 2)
		gMemoryAccess[kUSBLongWrite2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 4, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 4))
	{
		InvalidAccess (iAddress, 4, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTLONG (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = USBReg::UnsupportedWrite;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gWriteFunctions[offset];
	}

	assert (fn);
	fn (iAddress, 4, iLongValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 USBBank::SetWord
// ---------------------------------------------------------------------------

void USBBank::SetWord (emuptr iAddress, uint32 iWordValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBWordWrite]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 2, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 2))
	{
		InvalidAccess (iAddress, 2, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTWORD (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = USBReg::UnsupportedWrite;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gWriteFunctions[offset];
	}

	assert (fn);
	fn (iAddress, 2, iWordValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 USBBank::SetByte
// ---------------------------------------------------------------------------

void USBBank::SetByte (emuptr iAddress, uint32 iByteValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kUSBByteWrite]++;
#endif

#if (PREVENT_USER_REGISTER_SET)
	if (gMemAccessFlags.fProtect_RegisterSet && ROMBank::IsPCInRAM ())
	{
		PreventedAccess (iAddress, 1, false);
	}
#endif

#if (VALIDATE_REGISTER_SET)
	if (gMemAccessFlags.fValidate_RegisterSet && !ValidAddress (iAddress, 1))
	{
		InvalidAccess (iAddress, 1, false);
	}
#endif

#if HAS_PROFILING
	CYCLE_PUTBYTE (WAITSTATES_USB);
#endif

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = USBReg::UnsupportedWrite;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gWriteFunctions[offset];
	}

	assert (fn);
	fn (iAddress, 1, iByteValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 1);
}


// ---------------------------------------------------------------------------
//		 USBBank::ValidAddress
// ---------------------------------------------------------------------------

int USBBank::ValidAddress (emuptr iAddress, uint32 iSize)
{
	UNUSED_PARAM(iSize)

	uint32 		offset = iAddress - kMemoryStart;
	WriteFunction	fn = USBReg::UnsupportedWrite;

	if (offset < sizeof (UsbHwrType))
	{
		fn = gWriteFunctions[offset];
	}

	int result = (fn != USBReg::UnsupportedWrite);

	return result;
}


// ---------------------------------------------------------------------------
//		 USBBank::GetRealAddress
// ---------------------------------------------------------------------------

uint8* USBBank::GetRealAddress (emuptr iAddress)
{
	uint8*	loc = ((uint8*) &gUSBRegs) + (iAddress - kMemoryStart);

	return loc;
}



// ---------------------------------------------------------------------------
//		 USBBank::InvalidAccess
// ---------------------------------------------------------------------------

void USBBank::InvalidAccess (emuptr iAddress, long size, Bool forRead)
{
	UNUSED_PARAM(iAddress)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 USBBank::PreventedAccess
// ---------------------------------------------------------------------------

void USBBank::PreventedAccess (emuptr iAddress, long size, Bool forRead)
{
	Memory::PreventedAccess (iAddress, size, forRead, Errors::kRegisterAccess);
}


#pragma mark -

// ===========================================================================
//		 Register Accessors
// ===========================================================================

// ---------------------------------------------------------------------------
//		 USBReg::SetHandler
// ---------------------------------------------------------------------------

void USBReg::SetHandler (ReadFunction read, WriteFunction write,
							uint32 start, int count)
{
	int index = start - kMemoryStart;
	for (int ii = 0; ii < count; ++ii, ++index)
	{
		gReadFunctions[index] = read;
		gWriteFunctions[index] = write;
	}
}


// ---------------------------------------------------------------------------
//		 USBReg::UnsupportedRead
// ---------------------------------------------------------------------------

uint32 USBReg::UnsupportedRead (emuptr address, int size)
{
	if (!CEnableFullAccess::AccessOK ())
	{
		USBBank::InvalidAccess (address, size, true);
	}

	return ~0;
}


// ---------------------------------------------------------------------------
//		 USBReg::UnsupportedWrite
// ---------------------------------------------------------------------------

void USBReg::UnsupportedWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(value)

	if (!CEnableFullAccess::AccessOK ())
	{
		USBBank::InvalidAccess (address, size, false);
	}
}


// ---------------------------------------------------------------------------
//		 USBReg::StdRead
// ---------------------------------------------------------------------------

uint32 USBReg::StdRead (emuptr address, int size)
{
	if (size == 1)
		return do_get_mem_byte (USBBank::GetRealAddress (address));

	if (size == 2)
		return do_get_mem_word (USBBank::GetRealAddress (address));

	return do_get_mem_long (USBBank::GetRealAddress (address));
}


// ---------------------------------------------------------------------------
//		 USBReg::StdWrite
// ---------------------------------------------------------------------------

void USBReg::StdWrite (emuptr address, int size, uint32 value)
{
	if (size == 1)
		do_put_mem_byte (USBBank::GetRealAddress (address), value);

	else if (size == 2)
		do_put_mem_word (USBBank::GetRealAddress (address), value);

	else
		do_put_mem_long (USBBank::GetRealAddress (address), value);
}


// ---------------------------------------------------------------------------
//		 USBReg::NullWrite
// ---------------------------------------------------------------------------

void USBReg::NullWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(address)
	UNUSED_PARAM(size)
	UNUSED_PARAM(value)

	// Do nothing (for read-only registers).
}
