/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#ifndef EmPalmStructs_h
#define EmPalmStructs_h

#include "UAE_Utils.h"			// uae_memcpy

/*
	The Palm OS Emulator interacts closely with legacy Palm OS data
	structures.  These data structures are laid out by the compilers
	designed for the Palm OS.  However, the Palm OS Emulator -- which needs
	to access these structures at times -- is NOT necessarily compiled with
	a compiler that lays out the structures in a compatible fashion.  Thus,
	if the emulator needs to muck with, say, a variable of EventType, it
	can't merely create one and have at it.  Nor can it get a pointer to
	one, cast it to an EventType*, and have at it. There's just no guarantee
	that the compiler used to build Poser will put the fields in the same
	places that the Palm OS expects it.

	To solve this problem, for each Palm OS struct Poser wants to
	manipulate, it creates two parallel "Rep" classes: an "Alias" class and
	a "Proxy" class.  These Rep classes appear and act as much as the
	original struct as possible.  However, behind the scenes C++ magic
	allows Poser to access the data in a manner that is guaranteed to be
	binary-compatible with the Palm OS structs, regardless of what platform
	it's running on and what compiler it was built with.

	Aliases and Proxies are very similar to each other in the way the user
	interacts with them.  The have fields which, when used as an lvalue,
	will store the proferred data in the correct memory location and with
	the correct endianness and, when used as an rvalue, will produce the
	necessary value, fetching the data from the correct memory location and
	byteswapping it if necessary.

	Aliases and Proxies differ in their backing store.  An Alias class is
	initialized with a pointer to some buffer that already exists.  This
	buffer may exist as a Poser variable or as a buffer in the emulated Palm
	OS space.  All data accesses are made against that provided buffer. 
	When an Alias is created from another Alias, they both end up pointing
	to the same space.  If one Alias is used to modify the buffer, the
	second Alias will also reflect that change.

	Proxies, on the other hand, are more stand-alone.  They provide their
	own buffers as a backing store.  Copying one Proxy to another or
	creating one Proxy from another results in a new independent object. 
	After the copy has been made, modifying one object does not affect the
	second.

	Aliases are not only handy for wrapping a chunk of data that was
	generated elsewhere, but they are also essential for providing an "array
	operator" (that is, "operator[]") for both kinds of Rep classes.  That
	operator returns an Alias that points to the right location in memory. 
	If we didn't have Aliases -- for instance, we only had Proxies -- then
	the Proxy returned by operator[] could not be used to modify the backing
	store, since it's independent of the backing store used by the object
	that created it.

	Proxies, on the other hand, are handy when you need to treat an object
	more like a real-live struct.  Proxies can be copied and put into
	STL-like collections without having to worry about the lifetime of the
	object they were created from.

	Let's look at the Alias class for EventType:

		template <class A>
		class EmAliasEventType
		{
			public:
				typedef typename A::ptr_type ptr_type;

				EmAliasEventType (ptr_type);
				EmAliasEventType (const EmAliasEventType&);

				ptr_type GetPtr (void) const;
				static size_t GetSize (void);
				
				EmAliasEventType<A>& operator= (const EmAliasEventType<A>&);
				EmAliasEventType<A> operator[] (int);

			private:
				EmAliasEventType (void);

				EmAliasEventType<A>& operator= (const bool);
				EmAliasEventType<A>& operator= (const char);
				EmAliasEventType<A>& operator= (const unsigned char);
				EmAliasEventType<A>& operator= (const unsigned short);
				EmAliasEventType<A>& operator= (const unsigned int);
				EmAliasEventType<A>& operator= (const unsigned long);
				EmAliasEventType<A>& operator= (const signed char);
				EmAliasEventType<A>& operator= (const signed short);
				EmAliasEventType<A>& operator= (const signed int);
				EmAliasEventType<A>& operator= (const signed long);
				EmAliasEventType<A>& operator= (const void*);
				
				ptr_type fPtr;

			public:
				EmAliasInt16<A> eType;
				EmAliasBoolean<A> penDown;
				EmAliasUInt8<A> tapCount;
				EmAliasInt16<A> screenX;
				EmAliasInt16<A> screenY;
				EmAliasEventTypeData<A> data;
		};

	This is a template class parameterized with a data accessor.  This data
	accessor has static methods for reading and writing 1-, 2-, and 4-byte
	values.  In this module, we define two such accessors. One, called LAS
	for Local Address Space, accesses Big Endian values in Poser's address
	space.  The second, called PAS for Palm Address Space, accesses Big
	Endian values in the emulated address space.

	There are two constructors.  The first constructor takes a pointer to an
	appropriately-sized buffer that already exists -- all reads and write
	take place against that buffer.  The second constructor is the copy
	constructor.  It copies the pointer used by the original object.  After
	the copy, if the original object modifies the data in the backing store,
	the copy with reflect that change, too.

	GetPtr returns the pointer to the buffer backing the struct.  The return
	type of this function is based on the class the proxy class is
	instantiated with.  It is void* if used with LAS, and emuptr if used
	with PAS.

	GetSize returns the size of the represented struct, in this case
	EventType, which is 24 bytes long.

	An assignment operator is provided to perform a memberwise copy
	operation.

	An array operator is provided to return Alias objects that point to the
	appropriately index location in memory.

	Following these standard methods are the fields.  These fields are also
	Alias classes, initialized with the base pointer passed into the
	EmAliasEventType constructor plus an appropriate offset.

	Now let's take a look at a Proxy class:

		class EmProxyEventType
		{
			public:
				typedef void ptr_type;

				EmProxyEventType (void);
				EmProxyEventType (const EmProxyEventType&);

				ptr_type GetPtr (void) const;
				static size_t GetSize (void);

				EmProxyEventType& operator= (const EmProxyEventType&);
				EmAliasEventType<LAS> operator[] (int);

			private:
				EmProxyEventType& operator= (const bool);
				EmProxyEventType& operator= (const char);
				EmProxyEventType& operator= (const unsigned char);
				EmProxyEventType& operator= (const unsigned short);
				EmProxyEventType& operator= (const unsigned int);
				EmProxyEventType& operator= (const unsigned long);
				EmProxyEventType& operator= (const signed char);
				EmProxyEventType& operator= (const signed short);
				EmProxyEventType& operator= (const signed int);
				EmProxyEventType& operator= (const signed long);
				EmProxyEventType<& operator= (const void*);
				
				char fLocalData[24];

			public:
				EmAliasInt16<LAS> eType;
				EmAliasBoolean<LAS> penDown;
				EmAliasUInt8<LAS> tapCount;
				EmAliasInt16<LAS> screenX;
				EmAliasInt16<LAS> screenY;
				EmAliasEventTypeData<LAS> data;
		};

	As you can see, this Proxy class is almost the same as an Alias class.
	However:

	*	Instead of having an fPtr field to hold the pointer to the data
		buffer, it has an fLocalData field that holds the data itself.

	*	There is no constructor that accepts a pointer.  Instead, there
		is a public default constructor.

	*	It is not a template class.  Since the data is local, it is
		always accessed in a manner consistant with the LAS accessor.

	*	The copy constructor copies the data instead of the pointer.

	There are two sorts of Rep classes: primitive and aggregate. Primitive
	Rep classes are responsible for the management of simple scalar types. 
	Aggregate Rep classes represent structs, and are composed of other
	primitive and aggregate Alias classes. The biggest difference between
	primitive and aggregate Rep classes is the set of assignment operators
	clients are allowed to call.  For instance, you can assign integers to
	primitive Rep classes, but not to aggregate Rep classes

	EXAMPLE:

		Say you wanted to call void EvtAddEventToQueue (EventType*).
		You would use EmEventType as follows:

		EmProxyEventType	event;	// Use local buffer

		event.eType		= penDownEvent;
		event.penDown	= true;
		event.screenX	= 100;
		event.screenY	= 100;

		EvtAddEventToQueue((EventType*) event.GetPtr ());

	EFFICIENCY:

		There is some overhead using these classes instead of using structs
		directly.  At the very least, there's an additional level of
		redirection, as the fields accessed contain pointers to the actual
		data as opposed to the actual data itself. Additionally, on Little
		Endian systems, there is the cost of byteswapping on every access. 
		Finally, there is the cost of calling and executing the proxy class
		constructor.

		In EmPalmStructs.cpp, there are two functions: Test1 and Test2. The
		first accesses the struct SysPktRPC2Type using the proxy class, the
		second using the struct directly.

		On the Mac, using CW Pro 5.2 with optimizations on, Test1 is 88
		bytes, calls an out-of-line constructor, and each field access takes
		one additional level of indirection.  Test2 is 64 bytes. The 24-byte
		difference is taken up by an additional 4 bytes for each of the
		assignments, and an additional 12 bytes to call the proxy class's
		constructor.

	LIMITATIONS:

	*	Can't take the adress of an EmFoo.  Will get *its* address
		instead of what you want (the address of the storage). Instead,
		call GetPtr.

	*	Need explicit casts to extract the value in cases where the
		type is ambigous.  Classic case is "printf("%d", my_EmInt)". The
		object will be pushed onto the stack, not the wrapped integer.

	*	Can't take the size of the object.  Instead, call GetSize.

	*	Can't call offsetof to get field offsets.  Instead, call
		GetOffset.  Note that this returns the offset of the most
		immediate struct.  That is, struct1.struct2.field.GetOffset()
		returns the offset from the beginning of struct2 to field.
		Calling offsetof (StructType1, struct2.field) returns the offset
		from the beginning of struct1 to the field. *** Note this method
		is not yet implemented. ***

	*	Need a way to copy between classes instantiated with different
		address space objects.
*/

int		BadGetter (void);
void	BadSetter (void);

#include "EmPalmStructs.i"


// PUT_BE_LONG stores the given value at the given location as a 4-byte Big endian value

template <class T>
inline
void PUT_BE_LONG(T *q, unsigned long newValue)
{
#if PLATFORM_MAC
	unsigned long*	p = (unsigned long*) q;
	*p = newValue;
#else
	unsigned char*	p = (unsigned char*) q;
	p[0] = (unsigned char) (newValue >> 24);
	p[1] = (unsigned char) (newValue >> 16);
	p[2] = (unsigned char) (newValue >>  8);
	p[3] = (unsigned char) (newValue >>  0);
#endif
}


// PUT_BE_SHORT stores the given value at the given location as a 2-byte Big endian value

template <class T>
inline
void PUT_BE_SHORT(T *q, unsigned short newValue)
{
#if PLATFORM_MAC
	unsigned short*	p = (unsigned short*) q;
	*p = newValue;
#else
	unsigned char*	p = (unsigned char*) q;
	p[0] = (unsigned char) (newValue >> 8);
	p[1] = (unsigned char) (newValue >> 0);
#endif
}


// PUT_BE_BYTE stores the given value at the given location as a 1-byte Big endian value

template <class T>
inline
void PUT_BE_BYTE(T *q, unsigned char newValue)
{
	unsigned char*	p = (unsigned char*) q;
	p[0] = (unsigned char) (newValue);
}


// GET_BE_LONG fetches a 4-byte Big Endian value from the given location

template <class T>
inline
unsigned long GET_BE_LONG(const T *q)
{
#if PLATFORM_MAC
	unsigned long*	p = (unsigned long*) q;
	return *p;
#else
	const unsigned char*	p = (const unsigned char*) q;
	unsigned long			result;

	result =	(((unsigned long) p[0]) << 24) |
				(((unsigned long) p[1]) << 16) |
				(((unsigned long) p[2]) <<  8) |
				(((unsigned long) p[3]) <<  0);

	return result;
#endif
}


// GET_BE_SHORT fetches a 2-byte Big Endian value from the given location

template <class T>
inline
unsigned short GET_BE_SHORT(const T *q)
{
#if PLATFORM_MAC
	unsigned short*	p = (unsigned short*) q;
	return *p;
#else
	const unsigned char*	p = (const unsigned char*) q;
	unsigned short 			result;

	result =	(((unsigned short) p[0]) << 8) |
				(((unsigned short) p[1]) << 0);

	return result;
#endif
}


// GET_BE_BYTE fetches a 1-byte Big Endian value from the given location

template <class T>
inline
unsigned char GET_BE_BYTE(const T *q)
{
	const unsigned char*	p = (const unsigned char*) q;
	return p[0];
}


// Provide methods for getting and setting Big Endian data in Local Address Space

class LAS
{
	public:
		typedef void* ptr_type;

		inline static uint8		GetByte (ptr_type p)					{ return (uint8) GET_BE_BYTE (p); }
		inline static uint16	GetWord (ptr_type p)					{ return (uint16) GET_BE_SHORT (p); }
		inline static uint32	GetLong (ptr_type p)					{ return (uint32) GET_BE_LONG (p); }
		inline static void		PutByte (ptr_type p, const uint8 v)		{ PUT_BE_BYTE (p, v); }
		inline static void		PutWord (ptr_type p, const uint16 v)	{ PUT_BE_SHORT (p, v); }
		inline static void		PutLong (ptr_type p, const uint32 v)	{ PUT_BE_LONG (p, v); }

		inline static void		BlockCopy (ptr_type d, ptr_type s, size_t len)
																		{ memcpy (d, s, len); }
};


// Provide methods for getting and setting Big Endian data in Palm Addres Space

class PAS
{
	public:
		typedef emuptr ptr_type;

		inline static uint8		GetByte (ptr_type p)					{ return (uint8) get_byte (p); }
		inline static uint16	GetWord (ptr_type p)					{ return (uint16) get_word (p); }
		inline static uint32	GetLong (ptr_type p)					{ return (uint32) get_long (p); }
		inline static void		PutByte (ptr_type p, const uint8 v)		{ put_byte (p, v); }
		inline static void		PutWord (ptr_type p, const uint16 v)	{ put_word (p, v); }
		inline static void		PutLong (ptr_type p, const uint32 v)	{ put_long (p, v); }

		inline static void		BlockCopy (ptr_type d, ptr_type s, size_t len)
																		{ uae_memcpy (d, s, len); }
};


// ======================================================================
// Macro for creating the class that wraps up simple types.
// ======================================================================

#define MAKE_SCALAR_CLASSES(type)													\
																					\
	/* ===== Scalar Alias class interface ===== */									\
																					\
	template <class A>																\
	class EmAlias##type																\
	{																				\
		public:																		\
			typedef typename A::ptr_type ptr_type;									\
																					\
								EmAlias##type (ptr_type);							\
								EmAlias##type (const EmAlias##type<A>&);			\
																					\
			ptr_type			GetPtr	(void) const	{ return fPtr; }			\
			static size_t		GetSize (void)			{ return sizeof (type); }	\
																					\
			EmAlias##type<A>&	operator= (const bool);								\
			EmAlias##type<A>&	operator= (const char);								\
																					\
			EmAlias##type<A>&	operator= (const unsigned char);					\
			EmAlias##type<A>&	operator= (const unsigned short);					\
			EmAlias##type<A>&	operator= (const unsigned int);						\
			EmAlias##type<A>&	operator= (const unsigned long);					\
																					\
			EmAlias##type<A>&	operator= (const signed char);						\
			EmAlias##type<A>&	operator= (const signed short);						\
			EmAlias##type<A>&	operator= (const signed int);						\
			EmAlias##type<A>&	operator= (const signed long);						\
																					\
			EmAlias##type<A>&	operator= (const void*);							\
			EmAlias##type<A>&	operator= (const EmAlias##type<A>&);				\
																					\
								operator type (void) const;							\
																					\
			EmAlias##type<A>	operator[] (int);									\
																					\
		private:																	\
								EmAlias##type (void);								\
																					\
			ptr_type			fPtr;												\
	};																				\
																					\
	/* ===== Scalar Proxy class interface ===== */									\
																					\
	class EmProxy##type																\
	{																				\
		public:																		\
			typedef void* ptr_type;													\
																					\
								EmProxy##type (void);								\
								EmProxy##type (const EmProxy##type&);				\
																					\
			ptr_type			GetPtr	(void) const	{ return (ptr_type) &fLocalData; }	\
			static size_t		GetSize (void)			{ return sizeof (type); }	\
																					\
			EmProxy##type&		operator= (const bool);								\
			EmProxy##type&		operator= (const char);								\
																					\
			EmProxy##type&		operator= (const unsigned char);					\
			EmProxy##type&		operator= (const unsigned short);					\
			EmProxy##type&		operator= (const unsigned int);						\
			EmProxy##type&		operator= (const unsigned long);					\
																					\
			EmProxy##type&		operator= (const signed char);						\
			EmProxy##type&		operator= (const signed short);						\
			EmProxy##type&		operator= (const signed int);						\
			EmProxy##type&		operator= (const signed long);						\
																					\
			EmProxy##type&		operator= (const void*);							\
			EmProxy##type&		operator= (const EmProxy##type&);					\
																					\
								operator type (void) const;							\
																					\
			EmAlias##type<LAS>	operator[] (int);									\
																					\
		private:																	\
			struct {																\
				char			_bytes[sizeof (type)];								\
			} fLocalData;															\
	};																				\
																					\
	/* ===== Scalar Alias class implementation ===== */								\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (void) :									\
		fPtr ((ptr_type) NULL)														\
	{																				\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (ptr_type p) :							\
		fPtr (p)																	\
	{																				\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (const EmAlias##type<A>& that) :			\
		fPtr (that.GetPtr ())														\
	{																				\
	}																				\
																					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, bool)							\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, char)							\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, signed char)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, signed short)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, signed int)						\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, signed long)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned char)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned short)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned int)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned long)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, const void*)					\
	MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, const EmAlias##type<A>&)		\
																					\
	template <class A>																\
	inline EmAlias##type<A>::operator type (void) const								\
	{																				\
		if (sizeof (type) == 1)														\
			return (type) A::GetByte(this->GetPtr ());								\
																					\
		if (sizeof (type) == 2)														\
			return (type) A::GetWord(this->GetPtr ());								\
																					\
		if (sizeof (type) == 4)														\
			return (type) A::GetLong(this->GetPtr ());								\
																					\
		return (type) BadGetter ();													\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A> EmAlias##type<A>::operator [] (int index)				\
	{																				\
		return EmAlias##type<A> ((ptr_type) (((char*) this->GetPtr ()) + index * this->GetSize ()));	\
	}																				\
																					\
																					\
	/* ===== Scalar Proxy class implementation ===== */								\
																					\
	inline EmProxy##type::EmProxy##type (void)										\
	{																				\
	}																				\
																					\
	inline EmProxy##type::EmProxy##type (const EmProxy##type& that)					\
	{																				\
		*this = (type) that;														\
	}																				\
																					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, bool)							\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, char)							\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, signed char)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, signed short)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, signed int)						\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, signed long)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, unsigned char)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, unsigned short)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, unsigned int)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, unsigned long)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, const void*)					\
	MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, const EmProxy##type&)			\
																					\
	inline EmProxy##type::operator type (void) const								\
	{																				\
		if (sizeof (type) == 1)														\
			return (type) LAS::GetByte(this->GetPtr ());							\
																					\
		if (sizeof (type) == 2)														\
			return (type) LAS::GetWord(this->GetPtr ());							\
																					\
		if (sizeof (type) == 4)														\
			return (type) LAS::GetLong(this->GetPtr ());							\
																					\
		return (type) BadGetter ();													\
	}																				\
																					\
	inline EmAlias##type<LAS> EmProxy##type::operator [] (int index)				\
	{																				\
		return EmAlias##type<LAS> ((ptr_type) (((char*) this->GetPtr ()) + index * this->GetSize ()));	\
	}


// Macro that creates the implementation for an assignment operator for simple types.

#define MAKE_ONE_SCALAR_ALIAS_ASSIGNMENT_OPERATOR(type, rhs_type)					\
																					\
	template <class A>																\
	inline EmAlias##type<A>& EmAlias##type<A>::operator= (rhs_type val)				\
	{																				\
		if (sizeof (type) == 1)			A::PutByte (this->GetPtr (), (unsigned char) (type) val);	\
		else if (sizeof (type) == 2)	A::PutWord (this->GetPtr (), (unsigned short) (type) val);	\
		else if (sizeof (type) == 4)	A::PutLong (this->GetPtr (), (unsigned long) (type) val);	\
		else BadSetter ();															\
																					\
		return *this;																\
	}

#define MAKE_ONE_SCALAR_PROXY_ASSIGNMENT_OPERATOR(type, rhs_type)					\
																					\
	inline EmProxy##type& EmProxy##type::operator= (rhs_type val)					\
	{																				\
		if (sizeof (type) == 1)			LAS::PutByte (this->GetPtr (), (unsigned char) (type) val);	\
		else if (sizeof (type) == 2)	LAS::PutWord (this->GetPtr (), (unsigned short) (type) val);\
		else if (sizeof (type) == 4)	LAS::PutLong (this->GetPtr (), (unsigned long) (type) val);	\
		else BadSetter ();															\
																					\
		return *this;																\
	}





// ======================================================================
// Macro for creating the class that wraps up aggregate types.
// ======================================================================

#define MAKE_STRUCT_CLASSES(type, size, FOR_EACH_FIELD)								\
																					\
	/* ===== Struct Alias class interface ===== */									\
																					\
	template <class A>																\
	class EmAlias##type																\
	{																				\
		public:																		\
			typedef typename A::ptr_type ptr_type;									\
																					\
								EmAlias##type	(ptr_type);							\
								EmAlias##type	(const EmAlias##type&);				\
																					\
			ptr_type			GetPtr	(void) const	{ return fPtr; }			\
			static size_t		GetSize	(void)			{ return size; }			\
																					\
			EmAlias##type<A>&	operator= (const EmAlias##type<A>&);				\
																					\
			EmAlias##type<A>	operator[] (int);									\
																					\
		private:																	\
								EmAlias##type	(void);								\
																					\
			EmAlias##type<A>&	operator= (const bool);								\
			EmAlias##type<A>&	operator= (const char);								\
																					\
			EmAlias##type<A>&	operator= (const unsigned char);					\
			EmAlias##type<A>&	operator= (const unsigned short);					\
			EmAlias##type<A>&	operator= (const unsigned int);						\
			EmAlias##type<A>&	operator= (const unsigned long);					\
																					\
			EmAlias##type<A>&	operator= (const signed char);						\
			EmAlias##type<A>&	operator= (const signed short);						\
			EmAlias##type<A>&	operator= (const signed int);						\
			EmAlias##type<A>&	operator= (const signed long);						\
																					\
			EmAlias##type<A>&	operator= (const void*);							\
																					\
			ptr_type			fPtr;												\
																					\
		public:																		\
			FOR_EACH_FIELD(DECLARE_FIELD_ALIAS)										\
	};																				\
																					\
	/* ===== Struct Proxy class interface ===== */									\
																					\
	class EmProxy##type																\
	{																				\
		public:																		\
			typedef void* ptr_type;													\
																					\
								EmProxy##type	(void);								\
								EmProxy##type	(const EmProxy##type&);				\
																					\
			ptr_type			GetPtr	(void) const	{ return (ptr_type) &fLocalData; }	\
			static size_t		GetSize	(void)			{ return size; }			\
																					\
			EmProxy##type&		operator= (const EmProxy##type&);					\
																					\
			EmAlias##type<LAS>	operator[] (int);									\
																					\
		private:																	\
			EmProxy##type&		operator= (const bool);								\
			EmProxy##type&		operator= (const char);								\
																					\
			EmProxy##type&		operator= (const unsigned char);					\
			EmProxy##type&		operator= (const unsigned short);					\
			EmProxy##type&		operator= (const unsigned int);						\
			EmProxy##type&		operator= (const unsigned long);					\
																					\
			EmProxy##type&		operator= (const signed char);						\
			EmProxy##type&		operator= (const signed short);						\
			EmProxy##type&		operator= (const signed int);						\
			EmProxy##type&		operator= (const signed long);						\
																					\
			EmProxy##type&		operator= (const void*);							\
																					\
			struct {																\
				char			_bytes[size];										\
			} fLocalData;															\
																					\
		public:																		\
			FOR_EACH_FIELD(DECLARE_FIELD_PROXY)										\
	};																				\
																					\
	/* ===== Struct Alias class implementation ===== */								\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (void) :									\
		fPtr ((ptr_type) NULL)														\
		FOR_EACH_FIELD(MAKE_MEMBER_INITIALIZER_ALIAS)								\
	{																				\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (ptr_type p) :							\
		fPtr (p)																	\
		FOR_EACH_FIELD(MAKE_MEMBER_INITIALIZER_ALIAS)								\
	{																				\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A>::EmAlias##type (const EmAlias##type<A>& that) :			\
		fPtr (that.fPtr)															\
		FOR_EACH_FIELD(MAKE_MEMBER_INITIALIZER_ALIAS)								\
	{																				\
		*this = that;																\
	}																				\
																					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, bool)							\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, char)							\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, signed char)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, signed short)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, signed int)						\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, signed long)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned char)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned short)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned int)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, unsigned long)					\
	MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, const void*)					\
																					\
	template <class A>																\
	inline EmAlias##type<A>& EmAlias##type<A>::operator= (const EmAlias##type<A>& val)	\
	{																				\
		if (this != &val)															\
		{																			\
			A::BlockCopy (this->GetPtr (), val.GetPtr (), this->GetSize ());		\
		}																			\
																					\
		return *this;																\
	}																				\
																					\
	template <class A>																\
	inline EmAlias##type<A> EmAlias##type<A>::operator [] (int index)				\
	{																				\
		return EmAlias##type<A> ((ptr_type) (((char*) this->GetPtr ()) + index * this->GetSize ()));	\
	}																				\
																					\
	/* ===== Struct Proxy class implementation ===== */								\
																					\
	inline EmProxy##type::EmProxy##type (void) :									\
		fLocalData ()																\
		FOR_EACH_FIELD(MAKE_MEMBER_INITIALIZER_PROXY)								\
	{																				\
	}																				\
																					\
	inline EmProxy##type::EmProxy##type (const EmProxy##type& that) :				\
		fLocalData ()																\
		FOR_EACH_FIELD(MAKE_MEMBER_INITIALIZER_PROXY)								\
	{																				\
		*this = that;																\
	}																				\
																					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, bool)							\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, char)							\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, signed char)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, signed short)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, signed int)						\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, signed long)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, unsigned char)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, unsigned short)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, unsigned int)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, unsigned long)					\
	MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, const void*)					\
																					\
	inline EmProxy##type& EmProxy##type::operator= (const EmProxy##type& val)		\
	{																				\
		if (this != &val)															\
		{																			\
			LAS::BlockCopy (this->GetPtr (), val.GetPtr (), this->GetSize ());		\
		}																			\
																					\
		return *this;																\
	}																				\
																					\
	inline EmAlias##type<LAS> EmProxy##type::operator [] (int index)				\
	{																				\
		return EmAlias##type<LAS> ((ptr_type) (((char*) this->GetPtr ()) + index * this->GetSize ()));	\
	}


// Macro for creating the data member that acts as a proxy for the data
// we're wrapping up.

#define DECLARE_FIELD_ALIAS(x0, x1, field_type, x3, field_name2)					\
																					\
	EmAlias##field_type<A>	field_name2;


#define DECLARE_FIELD_PROXY(x0, x1, field_type, x3, field_name2)					\
																					\
	EmAlias##field_type<LAS>	field_name2;


// Macros for creating the data member initializers in the constructors

#define MAKE_MEMBER_INITIALIZER_ALIAS(offset, x1, x2, x3, field_name2)				\
																					\
	, field_name2((ptr_type) (((char*) fPtr) + offset))


#define MAKE_MEMBER_INITIALIZER_PROXY(offset, x1, x2, x3, field_name2)				\
																					\
	, field_name2((ptr_type) (((char*) this->GetPtr ()) + offset))


// Macro that creates the implementation for an assignment operator for aggregate types.
// Since it doesn't make sense to assign a scalar to a struct, such assignment operators
// are declared private and this macros defines functions with dummy contents.

#define MAKE_ONE_STRUCT_ALIAS_ASSIGNMENT_OPERATOR(type, rhs_type)					\
																					\
	template <class A>																\
	inline EmAlias##type<A>& EmAlias##type<A>::operator= (rhs_type)					\
	{																				\
		return *this;																\
	}

#define MAKE_ONE_STRUCT_PROXY_ASSIGNMENT_OPERATOR(type, rhs_type)					\
																					\
	inline EmProxy##type& EmProxy##type::operator= (rhs_type)						\
	{																				\
		return *this;																\
	}


typedef void*	voidPtr;
typedef UInt32*	UInt32Ptr;

MAKE_SCALAR_CLASSES(Boolean)
MAKE_SCALAR_CLASSES(char)
MAKE_SCALAR_CLASSES(Int8)
MAKE_SCALAR_CLASSES(UInt8)
MAKE_SCALAR_CLASSES(Int16)
MAKE_SCALAR_CLASSES(UInt16)
MAKE_SCALAR_CLASSES(Int32)
MAKE_SCALAR_CLASSES(UInt32)
MAKE_SCALAR_CLASSES(voidPtr)
MAKE_SCALAR_CLASSES(UInt32Ptr)
MAKE_SCALAR_CLASSES(MemPtr)
MAKE_SCALAR_CLASSES(MemHandle)
MAKE_SCALAR_CLASSES(SlkPktHeaderChecksum)

// MAKE_PTR_PROXY_INTERFACE(ControlType)
// MAKE_PTR_PROXY_IMPLEMENTATION(ControlType)

FOR_EACH_STRUCT1(MAKE_STRUCT_CLASSES)
FOR_EACH_STRUCT2(MAKE_STRUCT_CLASSES)
FOR_EACH_STRUCT3(MAKE_STRUCT_CLASSES)
FOR_EACH_STRUCT4(MAKE_STRUCT_CLASSES)

#endif	// EmPalmStructs_h
