/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmRegs.h"

#include "CPU_REG.h"			// Software::BusError
#include "RAM_ROM.h"			// CEnableFullAccess


// These functions provide fetch and store access to the
// emulator's register memory.

static ReadFunction		gReadFunctions[kMemorySize];
static WriteFunction	gWriteFunctions[kMemorySize];


// ---------------------------------------------------------------------------
//		 EmRegs::Initialize
// ---------------------------------------------------------------------------

void EmRegs::Initialize (void)
{
}


// ---------------------------------------------------------------------------
//		 EmRegs::Reset
// ---------------------------------------------------------------------------

void EmRegs::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 EmRegs::Save
// ---------------------------------------------------------------------------

void EmRegs::Save (SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 EmRegs::Load
// ---------------------------------------------------------------------------

void EmRegs::Load (SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 EmRegs::Dispose
// ---------------------------------------------------------------------------

void EmRegs::Dispose (void)
{
}


// ---------------------------------------------------------------------------
//		 EmRegs::GetLong
// ---------------------------------------------------------------------------

uint32 EmRegs::GetLong (emuptr address)
{
#if HAS_PROFILING
	CYCLE_GETLONG (WAITSTATES_REGISTERS);
#endif

	return GetReadHandler (address) (address, 4);
}


// ---------------------------------------------------------------------------
//		 EmRegs::GetWord
// ---------------------------------------------------------------------------

uint32 EmRegs::GetWord (emuptr address)
{
#if HAS_PROFILING
	CYCLE_GETWORD (WAITSTATES_REGISTERS);
#endif

	return GetReadHandler (address) (address, 2);
}


// ---------------------------------------------------------------------------
//		 EmRegs::GetByte
// ---------------------------------------------------------------------------

uint32 EmRegs::GetByte (emuptr address)
{
#if HAS_PROFILING
	CYCLE_GETBYTE (WAITSTATES_REGISTERS);
#endif

	return GetReadHandler (address) (address, 1);
}


// ---------------------------------------------------------------------------
//		 EmRegs::SetLong
// ---------------------------------------------------------------------------

void EmRegs::SetLong (emuptr address, uint32 value)
{
#if HAS_PROFILING
	CYCLE_PUTLONG (WAITSTATES_REGISTERS);
#endif

	GetWriteHandler (address) (address, 4, value);
}


// ---------------------------------------------------------------------------
//		 EmRegs::SetWord
// ---------------------------------------------------------------------------

void EmRegs::SetWord (emuptr address, uint32 value)
{
#if HAS_PROFILING
	CYCLE_PUTWORD (WAITSTATES_REGISTERS);
#endif

	GetWriteHandler (address) (address, 2, value);
}


// ---------------------------------------------------------------------------
//		 EmRegs::SetByte
// ---------------------------------------------------------------------------

void EmRegs::SetByte (emuptr address, uint32 value)
{
#if HAS_PROFILING
	CYCLE_PUTBYTE (WAITSTATES_REGISTERS);
#endif

	GetWriteHandler (address) (address, 1, value);
}


// ---------------------------------------------------------------------------
//		 EmRegs::ValidAddress
// ---------------------------------------------------------------------------

int EmRegs::ValidAddress (emuptr address, uint32 size)
{
	UNUSED_PARAM(size)

	WriteFunction	fn = GetWriteHandler (address);
	int result = (fn != NULL);

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegs::SetSubHandlers
// ---------------------------------------------------------------------------

void EmRegs::SetSubHandlers (void)
{
	// The register bank is further sub-divided into separate handlers for
	// each memory address.  First, install the default handlers for reading
	// and writing (which basically report the access as invalid).

	SetHandlers (EmRegs::UnsupportedRead, EmRegs::UnsupportedWrite,
						kMemoryStart, kMemorySize);
}


// ---------------------------------------------------------------------------
//		 EmRegs::SetHandlers
// ---------------------------------------------------------------------------

void EmRegs::SetHandlers (ReadFunction read, WriteFunction write,
							emuptr start, int32 count)
{
	int	index = start - kMemoryStart;
	for (int ii = 0; ii < count; ++ii, ++index)
	{
		gReadFunctions[index] = read;
		gWriteFunctions[index] = write;
	}
}


// ---------------------------------------------------------------------------
//		 EmRegs::GetReadHandler
// ---------------------------------------------------------------------------

ReadFunction EmRegs::GetReadHandler (emuptr address) const
{
	uint32			offset = address - kMemoryStart;
	ReadFunction	fn = EmRegs::UnsupportedRead;

	if (offset < kMemorySize)
	{
		fn = gReadFunctions[offset];
	}

	return fn;
}


// ---------------------------------------------------------------------------
//		 EmRegs::GetWriteHandler
// ---------------------------------------------------------------------------

WriteFunction EmRegs::GetWriteHandler (emuptr address) const
{
	uint32			offset = address - kMemoryStart;
	WriteFunction	fn = EmRegs::UnsupportedWrite;

	if (offset < kMemorySize)
	{
		fn = gWriteFunctions[offset];
	}

	return fn;
}


// ---------------------------------------------------------------------------
//		 EmRegs::UnsupportedRead
// ---------------------------------------------------------------------------

uint32 EmRegs::UnsupportedRead (emuptr address, int size)
{
	if (!CEnableFullAccess::AccessOK ())
	{
		InvalidAccess (address, size, true);
	}

	return ~0;
}


// ---------------------------------------------------------------------------
//		 EmRegs::UnsupportedWrite
// ---------------------------------------------------------------------------

void EmRegs::UnsupportedWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(value)

	if (!CEnableFullAccess::AccessOK ())
	{
		InvalidAccess (address, size, false);
	}
}


// ---------------------------------------------------------------------------
//		 EmRegs::NullWrite
// ---------------------------------------------------------------------------

void EmRegs::NullWrite (emuptr address, int size, uint32 value)
{
	UNUSED_PARAM(address)
	UNUSED_PARAM(size)
	UNUSED_PARAM(value)

	// Do nothing (for read-only registers).
}


// ---------------------------------------------------------------------------
//		 EmRegs::InvalidAccess
// ---------------------------------------------------------------------------

void EmRegs::InvalidAccess (emuptr iAddress, int size, Bool forRead)
{
	UNUSED_PARAM(iAddress)
	UNUSED_PARAM(size)
	UNUSED_PARAM(forRead)

	Software::BusError ();
}
