/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmRegsEZVisor.h"
#include "EmRegsEZPrv.h"

#include "CPU_REG.h"			// Emulator::GetDevice
#include "EmDevice.h"			// HardwareID


// Visor pin definitions from Handspring.  Mostly the same as a Palm V,
// but with some pins changed and moved around.


/************************************************************************
 * Port B Bit settings
 ************************************************************************/
#define hwrLegoPortBLCDVccOff		0x08	// (L) LCD Vcc
#define hwrLegoPortBLCDAdjOn		0x40	// (H) LCD Contrast Voltage


/************************************************************************
 * Port C Bit settings
 ************************************************************************/
#define hwrLegoPortCLCDEnableOn		0x80 	// (H) LCD Enable


/************************************************************************
 * Port D Bit settings
 ************************************************************************/
#define hwrLegoPortDKbdCol0On		0x01 	// (H) Keyboard Column 0
#define hwrLegoPortDKbdCol1On		0x02 	// (H) Keyboard Column 1
#define hwrLegoPortDKbdCol2On		0x04 	// (H) Keyboard Column 2

#define hwrLegoPortDKbdRow2			0x08 	// (H) Keyboard Row 2

#define hwrLegoPortDDock1IrqOff		0x10 	// (L) Dock1 interrupt
#define hwrLegoPortDSlotIrqOff		0x20 	// (L) Slot Interrupt
#define hwrLegoPortDUsbIrqOff		0x40 	// (L) USB Interrupt
#define hwrLegoPortDCardInstIrqOff	0x80 	// (L) Card Installed Interrupt

#define hwrLegoPortDKeyBits			0x07 	// (H) All Keyboard Columns



/************************************************************************
 * Port E Bit settings
 ************************************************************************/
#define hwrLegoPortESerialTxD		0x20	// ( ) Serial transmit line
#define hwrLegoPortESlotResetOff	0x40 	// (L) Slot reset
#define hwrLegoPortEDock2Off		0x80 	// (L) Dock2 input


/************************************************************************
 * Port F Bit settings
 ************************************************************************/
#define hwrLegoPortFContrast		0x01	// ( ) LCD Contrast PWM output
#define hwrLegoPortFIREnableOff		0x04	// (L) Shutdown IR
#define hwrLegoPortFUsbCsOff		0x80	// (L) USB chip select


/************************************************************************
 * Port G Bit settings
 ************************************************************************/
#define hwrLegoPortGBacklightOff	0x01	// (L) Backlight on/off
#define hwrLegoPortGUSBSuspend		0x02	// (X) USB Suspend bit - Output low for active
											//                     - Input for suspend
#define hwrLegoPortGPowerFailIrqOff	0x04	// (L) Power Fail IRQ
#define hwrLegoPortGKbdRow0 		0x08	// (H) Keyboard Row 0
#define hwrLegoPortGKbdRow1 		0x10	// (H) Keyboard Row 1
#define hwrLegoPortGAdcCsOff		0x20	// (L) A/D Select


const int		kNumButtonRows = 3;
const int		kNumButtonCols = 3;

const uint16	kButtonMap[kNumButtonRows][kNumButtonCols] =
{
	{ keyBitHard1,	keyBitHard2,	keyBitHard3 },
	{ keyBitPageUp,	keyBitPageDown,	0 },
	{ keyBitPower,	keyBitHard4,	0 }
};


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetLCDOn
// ---------------------------------------------------------------------------

Bool EmRegsEZVisor::GetLCDOn (void)
{
	return (READ_REGISTER (portCData) & hwrLegoPortCLCDEnableOn) != 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetBacklightOn
// ---------------------------------------------------------------------------

Bool EmRegsEZVisor::GetBacklightOn (void)
{
	return (READ_REGISTER (portGData) & hwrLegoPortGBacklightOff) == 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetSerialPortOn
// ---------------------------------------------------------------------------

Bool EmRegsEZVisor::GetSerialPortOn (void)
{
	uint16	uControl = READ_REGISTER (uControl);
	return (uControl & hwrEZ328UControlUARTEnable) != 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetPortInputValue
// ---------------------------------------------------------------------------

uint8 EmRegsEZVisor::GetPortInputValue (int port)
{
	uint8	result = EmRegsEZ::GetPortInputValue (port);

	if (port == 'E')
	{
		/*
			Support the Visor model ID.  From Bob Petersen at Handspring:

			The Visor ROM only works with 1 device so far. But, the Visor does have
			model detect pins so that future ROMs can work on multiple platforms. The
			model input pins are connected to port E, bits 0-2 (E[0-2]). To read them,
			you have to drive port G bit 2 low and enable the pullups on port E.

			// ==================================================================
			// Try and identify which hardware we are on since that may determine
			//  how we setup for using RAM
			// The hardware ID bits are read from the SPI pins while driving
			//  the IRQ7 pin as an output low
			// ==================================================================
			#define   modelInPins	  0x07	  // SPI pins on Visor
			#define   modelOutPin	  0x04	  // EMUIRQ (IRQ7)

			// Make the SPI pins inputs with pullups enabled
			dbP->portEDir &= ~modelInPins;
			dbP->portEPullupEn |= modelInPins;
			dbP->portESelect |= modelInPins;

			// Drive the IRQ7 pin output low
			dbP->portGData &= ~modelOutPin;
			dbP->portGDir |= modelOutPin;
			dbP->portGSelect |= modelOutPin;

			// Delay a bit
			PrvTemp = 100;
			while (--PrvTemp)
				;

			// Read the inputs now and save in D6 for later
			PrvModelID = (~dbP->portEData) & modelInPins;


			// Drive the IRQ7 output high before we switch it back to
			//  an interrupt input in case the pullup is weak.
			dbP->portGData |= modelOutPin;
			dbP->portGSelect &= ~modelOutPin;

			// Clear any pending IRQ7 sources
			dbP->emuStatus = dbP->emuStatus;
		*/

#define modelOutPin		0x04		// EMUIRQ (IRQ7)

		UInt8	portGData		= READ_REGISTER (portGData);
		UInt8	portGDir		= READ_REGISTER (portGDir);
		UInt8	portGSelect		= READ_REGISTER (portGSelect);

		if (((portGData & modelOutPin) == 0) &&
			((portGDir & modelOutPin) == modelOutPin) &&
			((portGSelect & modelOutPin) == modelOutPin))
		{
			EmDevice	device	= Emulator::GetDevice ();
			result |= device.HardwareID ();
		}
		else
		{
			// The pin for detecting "in cradle" is hwrLegoPortEDock2Off.  It should be low
			// if in the serial cradle.  If it is high, we assume USB.
			//			-- Bob Petersen

			result &= ~hwrLegoPortEDock2Off;
		}
	}
	else if (port == 'G')
	{
		// Make sure that hwrLegoPortGPowerFailIrqOff is set.  If it's clear,
		// the battery code will make the device go to sleep immediately.

		result |= hwrLegoPortGPowerFailIrqOff;
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetPortInternalValue
// ---------------------------------------------------------------------------

uint8 EmRegsEZVisor::GetPortInternalValue (int port)
{
	uint8	result = EmRegsEZ::GetPortInternalValue (port);

	if (port == 'D')
	{
		// Always set the high bit, indicating that there's no memory module
		// installed.  From Bob Petersen at Handspring:
		//
		//	This is a probe into the module memory, to check for a replacement ROM on a
		//	card that is plugged in.  The card is mapped to 0x20000000 and the card's
		//	ROM is at 0x28000000.  If it finds the card header (FEEDBEEF) and the card
		//	is plugged in that has a non-zero reset vector, the chip selects are changed
		//	so B0 addresses 10c00000, and then we jump to the reset vector and boot off
		//	the card.
		//
		//	This part of the boot code should only be called if it is detected that a
		//	card is currently plugged in (bit 7 of port D) and we aren't currently
		//	booting from the card.

		result |= hwrLegoPortDCardInstIrqOff;

		// Ensure that bit hwrLegoPortDDock1IrqOff is set.  If it's clear, HotSync
		// will sync via the modem instead of the serial port.

		result |= hwrLegoPortDDock1IrqOff;
	}
	else if (port == 'G')
	{
		// Make sure that hwrLegoPortGPowerFailIrqOff is set.  If it's clear,
		// the battery code will make the device go to sleep immediately.

		result |= hwrLegoPortGPowerFailIrqOff;
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegsEZVisor::GetKeyInfo
// ---------------------------------------------------------------------------

void EmRegsEZVisor::GetKeyInfo (int* numRows, int* numCols,
								uint16* keyMap, Bool* rows)
{
	*numRows = kNumButtonRows;
	*numCols = kNumButtonCols;

	memcpy (keyMap, kButtonMap, sizeof (kButtonMap));

	// Determine what row is being asked for.

	UInt8	portDDir	= READ_REGISTER (portDDir);
	UInt8	portDData	= READ_REGISTER (portDData);

	UInt8	portGDir	= READ_REGISTER (portGDir);
	UInt8	portGData	= READ_REGISTER (portGData);

	rows[0]	= (portGDir & hwrLegoPortGKbdRow0) != 0 && (portGData & hwrLegoPortGKbdRow0) == 0;
	rows[1]	= (portGDir & hwrLegoPortGKbdRow1) != 0 && (portGData & hwrLegoPortGKbdRow1) == 0;
	rows[2]	= (portDDir & hwrLegoPortDKbdRow2) != 0 && (portDData & hwrLegoPortDKbdRow2) == 0;
}
