/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#ifndef _EMULATORTYPES_H_
#define _EMULATORTYPES_H_

#include "EmFileRef.h"			// EmFileRef
#include "EmDevice.h"			// DeviceType

#include <deque>
#include <map>
#include <string>
#include <vector>


// ======================================
// ========== ENUMERATED TYPES ==========
// ======================================

// ---------- <Less of a type, more of a constant...> ----------

enum
{
	MRU_COUNT = 9
};

// ---------- CloseActionType ----------

enum CloseActionType
{
	kSaveAlways,
	kSaveAsk,
	kSaveNever
};
typedef vector<CloseActionType>	CloseActionList;


// ---------- LoggingType ----------

enum LoggingType
{
	kNormalLogging = 0x01,
	kGremlinLogging = 0x02,
	kAllLogging = kNormalLogging | kGremlinLogging
};
typedef vector<LoggingType>	LoggingList;


// =====================================
// ========== SIMPLE TYPEDEFS ==========
// =====================================

typedef uint8	ScaleType;
typedef int32	RAMSizeType;

typedef vector<ScaleType>		ScaleList;
typedef vector<RAMSizeType>		RAMSizeList;

typedef vector<unsigned char>	ByteList;
typedef vector<string>			StringList;
typedef vector<RectangleType>	RectangleList;
typedef vector<PointType>		PointLess;

typedef map<string, string>		StringStringMap;


// ================================
// ========== STRUCTURES ==========
// ================================

// ---------- RGBType ----------

struct RGBType
{
	RGBType (void) {}
	RGBType (uint8 red, uint8 green, uint8 blue) :
		fRed (red),
		fGreen (green),
		fBlue (blue)
	{}

	bool operator== (const RGBType& rhs) const
	{
		return fRed == rhs.fRed && fGreen == rhs.fGreen && fBlue == rhs.fBlue;
	}

	bool operator!= (const RGBType& rhs) const
	{
		return !(*this == rhs);
	}

	uint8		fRed;
	uint8		fGreen;
	uint8		fBlue;
};
typedef vector<RGBType>	RGBList;


// ---------- SystemCallContext ----------

struct SystemCallContext
{
	emuptr		fPC;			// PC at which the system call was made.
	emuptr		fNextPC;		// PC of instruction after system call.
	emuptr		fDestPC;		// Address of system function.
	uint16		fTrapWord;		// Trapword, even for SYSTRAP_FAST calls.
	uint16		fTrapIndex;		// Trapword with the high 4 bits removed.
	emuptr		fExtra;			// RefNum for library calls, D2 for "dispatch" calls.
	Bool		fViaTrap;		// True if called via TRAP $F.
	Bool		fViaJsrA1;		// True if called via SYSTRAP_FASTER.

	long		fError;			// If an error occurred getting the context, error is here
	long		fLibIndex;		// Resource base number for function name
	long		fMaxRefNum;		// If the refNum was too big, this is the max value it could be.
};
typedef vector<SystemCallContext>	SystemCallContextList;


// ---------- DatabaseInfo ----------

struct DatabaseInfo
{
	UInt32		creator;
	UInt32		type;
	UInt16		version;
	LocalID		dbID;
	UInt16 		cardNo;
	UInt32		modDate;
	UInt16		dbAttrs;

	char		name[dmDBNameLength];
	char		dbName[dmDBNameLength];

	Bool operator < (const DatabaseInfo& other) const
		{ if (creator < other.creator) return true;
		  if (creator > other.creator) return false;
		  if (type < other.type) return true;
		  return false; }

	Bool operator == (const DatabaseInfo& other) const
		{ return (creator == other.creator) &&
				 (type == other.type); }
};
typedef vector<DatabaseInfo>	DatabaseInfoList;


// ---------- Configuration ----------

struct Configuration
{
	Configuration () :
		fDeviceType (kDeviceUnspecified),
		fRAMSize (1024),
		fROMFile ()
	{
	}

	Configuration (DeviceType dt, long size, const EmFileRef& rom) :
		fDeviceType (dt),
		fRAMSize (size),
		fROMFile (rom)
	{
	}

	DeviceType				fDeviceType;
	RAMSizeType				fRAMSize;
	EmFileRef				fROMFile;
};
typedef vector<Configuration>	ConfigurationList;


// ---------- GremlinInfo ----------

struct GremlinInfo
{
	GremlinInfo () :
		fNumber (0),
		fSteps (-1),
		fSaveFrequency (10000),
		fAppList ()
	{
	}

	long				fNumber;
	long				fSteps;
	long				fSaveFrequency;
	DatabaseInfoList	fAppList;
};
typedef vector<GremlinInfo>	GremlinInfoList;


// ---------- HordeInfo ----------

struct HordeInfo
{
	HordeInfo () :
		fStartNumber (0),
		fStopNumber (0),
		fDepthSwitch (30000),
		fDepthSave (10000),
		fDepthStop (1000000),
		fCanSwitch (false),
		fCanSave (false),
		fCanStop (false),
		fAppList ()

	{
		NewToOld ();
	}

	long				fStartNumber;
	long				fStopNumber;

	long				fDepthSwitch;
	long				fDepthSave;
	long				fDepthStop;

	Bool				fCanSwitch;
	Bool				fCanSave;
	Bool				fCanStop;

	DatabaseInfoList	fAppList;

	// Old fields that I want to get rid of, but
	// I need to update Gremlins and Hordes first.

	long				fSaveFrequency;
	long				fSwitchDepth;
	long				fMaxDepth;

	void NewToOld (void)
	{
		fSwitchDepth	= fCanSwitch ? fDepthSwitch : -1;
		fSaveFrequency	= fCanSave ? fDepthSave : 0;
		fMaxDepth		= fCanStop ? fDepthStop : -1;
	}

	void OldToNew (void)
	{
		fCanSwitch = fSwitchDepth > 0;
		fDepthSwitch = fCanSwitch ? fSwitchDepth : 30000;

		fCanSave = fSaveFrequency > 0;
		fDepthSave = fCanSave ? fSaveFrequency : 10000;

		fCanStop = fMaxDepth > 0;
		fDepthStop = fCanStop ? fMaxDepth : 1000000;
	}
};
typedef vector<HordeInfo>	HordeInfoList;


// ---------- HordeInfo ----------

typedef struct SysLibTblEntryTypeV10 {
	MemPtr*			dispatchTblP;					// pointer to library dispatch table
	void*			globalsP;						// Library globals
	} SysLibTblEntryTypeV10;
typedef SysLibTblEntryTypeV10*	SysLibTblEntryV10Ptr;



// ----------------------------------------------------------------------
// For reasons of "convenience", the official type declarations for these
// packets omit their bodies.  For the emulator, we'd really like to have
// the bodies, so we define these parallel types.
// ----------------------------------------------------------------------

#include "PalmPack.h"

struct SysPktReadMemRspType2
{
	_sysPktBodyCommon;
	UInt8			data[sysPktMaxBodySize-2];
};

struct SysPktWriteMemCmdType2
{
	_sysPktBodyCommon;
	void*			address;					// Address to write
	UInt16			numBytes;					// # of bytes to write
	UInt8			data[sysPktMaxBodySize-8];
};

struct SysPktReadRegsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	M68KRegsType	reg;							// <- return registers
};

struct SysPktGetBreakpointsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktSetBreakpointsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktGetTrapBreaksRspType2
{
	_sysPktBodyCommon;								// Common Body header
	UInt16			trapBP[dbgTotalTrapBreaks];
};

struct SysPktSetTrapBreaksCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	UInt16			trapBP[dbgTotalTrapBreaks];
};

struct SysPktGetTrapConditionsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	UInt16			trapParam[dbgTotalTrapBreaks];
};

struct SysPktSetTrapConditionsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	UInt16			trapParam[dbgTotalTrapBreaks];
};

struct SysPktFindCmdType2
{
	_sysPktBodyCommon;								// Common Body header

	UInt32	firstAddr;								// first address to search
	UInt32	lastAddr;								// last address to begin searching
	UInt16	numBytes;								// number of data bytes to match
	Boolean	caseInsensitive;						// if true, perform a case-insensitive search
	UInt8 	padding;
	char	data[sysPktMaxBodySize - 14];
};

struct SysPktRPCType2
{
	_sysPktBodyCommon;					// Common Body header
	UInt16	trapWord;					// which trap to execute
	UInt32	resultD0;					// result from D0 placed here
	UInt32	resultA0;					// result from A0 placed here
	UInt16	numParams;					// how many parameters follow
	// Following is a variable length array of SlkRPCParamInfo's
//	SysPktRPCParamType	param[1];
	char	param[sysPktMaxBodySize - 14];
};

struct SysPktEmptyRspType
{
	_sysPktBodyCommon;			// Common Body header
};



struct SysPktRPC2Type
{
	_sysPktBodyCommon;				// Common Body header

	UInt16		trapWord;			// which trap to execute
	UInt32		resultD0;			// result from D0 placed here
	UInt32		resultA0;			// result from A0 placed here
	UInt16		resultException;

	UInt8		DRegMask;
	UInt8		ARegMask;
	UInt32		Regs[1];

	UInt16		numParams;		   // how many parameters follow
//	SysPktRPCParamType	param[1];
	char	param[sysPktMaxBodySize - 22];
};

struct SysPktRPCParamType2
{
	UInt8	byRef;					// true if param is by reference
	UInt8	size;					// # of Bytes of paramData	(must be even)
	union
	{
		UInt8	asByte;
		UInt16	asShort;
		UInt32	asLong;
	};
};

struct SysPktRPCParamByte
{
	UInt8	byRef;					// true if param is by reference
	UInt8	size;					// # of Bytes of paramData	(must be even)			
	UInt8	data;					// variable length array of paramData
	UInt8	padding;
};

struct SysPktRPCParamShort
{
	UInt8	byRef;					// true if param is by reference
	UInt8	size;					// # of Bytes of paramData	(must be even)			
	UInt16	data;					// variable length array of paramData
};

struct SysPktRPCParamLong
{
	UInt8	byRef;					// true if param is by reference
	UInt8	size;					// # of Bytes of paramData	(must be even)			
	UInt32	data;					// variable length array of paramData
};


#include "PalmPackPop.h"


// ============================================
// ========== Windows socket mapping ==========
// ============================================

#if !PLATFORM_WINDOWS
	typedef int				SOCKET;

	#define WSAEINVAL		EINVAL
	#define INVALID_SOCKET	((SOCKET) (~0))
	#define SOCKET_ERROR	(-1)
	#define closesocket		close
	#define ioctlsocket		ioctl

	#ifndef INADDR_LOOPBACK
	#define INADDR_LOOPBACK	0x7f000001
	#endif
#endif

#if PLATFORM_UNIX || PLATFORM_MAC
	#include <sys/types.h>
	#include <sys/socket.h>			// socklen_t

	#if (defined(__svr4__) && defined(sun)) || defined (__FreeBSD__)
		typedef int			socklen_t;	// not defined in <sys/socket.h> on solaris or FreeBSD
	#endif
#else
	typedef int			socklen_t;
#endif


#endif	// _EMULATORTYPES_H_
