/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "LoadApplication.h"

#include "Byteswapping.h"
#include "EmStreamFile.h"		// EmStreamFile
#include "ErrorHandling.h"		// Errors::ThrowIfPalmError, kError_OutOfMemory
#include "Miscellaneous.h"		// StMemory
#include "Platform.h"			// Platform::AllocateMemory, Platform::LFPW...
#include "PreferenceMgr.h"		// gEmuPrefs->UpdatePRCMRU
#include "Strings.r.h"			// kStr_ values
#include "TrapPatches.h"		// Patches::UIInitialized
#include "UAE_Utils.h"			// uae_memcpy

#include <algorithm>			// find

Bool PrvLauncherActive (void);
Bool PrvSwitchToApp (UInt32 type, UInt32 creator);

// Define ResTypeIDPair explicitly instead of typedeffing
// it to the right kind of pair<xxx,xxx>.  The latter has
// a constructor that takes references to its parameters.
// If the referenced objects are on a non-longword boundary
// on a Sparc system, the application will crash.  So instead
// we have our own class that takes the parameters as values
// instead of references.

//typedef pair<UInt32, UInt16>	ResTypeIDPair;
struct ResTypeIDPair
{
	ResTypeIDPair(UInt32 a, UInt16 b) : Type(a), ID(b) {}

	UInt32 Type;
	UInt16 ID;
};

inline bool operator==(const ResTypeIDPair& x, const ResTypeIDPair& y)
{
	return x.Type == y.Type && x.ID == y.ID;
}


static Bool PrvIsResources(UInt16 attributes)
{
	return (attributes & dmHdrAttrResDB);
}

static Bool PrvIsResources(DatabaseHdrPtr hdrP)
{
	return PrvIsResources(hdrP->attributes);
}

static UInt32 PrvGetEntrySize(DatabaseHdrPtr hdrP)
{
	if (PrvIsResources(hdrP))
		return sizeof (RsrcEntryType);
	
	return sizeof (RecordEntryType);
}

/************************************************************
 *
 *	FUNCTION: PrvValidateDatabaseHeader
 *
 *	DESCRIPTION: Validates database header (to be used before
 *		any of the rest of the database is byteswapped or
 *		otherwise processed)
 *
 *	PARAMETERS: 
 *			hdrP - pointer to database header
 *			bufferSize - size of the buffer holding the database
 *
 *	RETURNS: Err - error code
 *
 *************************************************************/

static ErrCode PrvValidateDatabaseHeader(DatabaseHdrPtr hdrP, UInt32 bufferSize)
{
	// Make sure the database name is not too long.
	size_t	len = strlen((char*) hdrP->name);
	if (len > dmDBNameLength)
	{
		Errors::SetParameter ("%database_name", (char*) hdrP->name);
		return kError_BadDB_NameNotNULLTerminated;
	}

	// Check to see that name consists of printable characters
	for (size_t ii = 0; ii < len; ++ii)
	{
		if (hdrP->name[ii] < ' ')
		{
			Errors::SetParameter ("%database_name", (char*) hdrP->name);
			return kError_BadDB_NameNotPrintable;
		}
	}

	// Check that the resource/database records headers fit within the space
	// of the buffer before we even byteswap so that we do not crash while
	// byteswapping them.
	UInt32	endOfEntryArray = offsetof (DatabaseHdrType, recordList.firstEntry) +
							PrvGetEntrySize(hdrP) * hdrP->recordList.numRecords;

	if (endOfEntryArray > bufferSize)
		return kError_BadDB_FileTooSmall;

	// If we got here, we're okay.
	return errNone;
}


static void PrvSetResourceTypeIDParameters (UInt32 resType, UInt16 resID)
{
	char	buffer[20];

	sprintf (buffer, "%c%c%c%c",
					(char) (resType >> 24),
					(char) (resType >> 16),
					(char) (resType >> 8),
					(char) (resType >> 0));
	Errors::SetParameter ("%res_type", buffer);

	Errors::SetParameter ("%res_id", resID);
}


/************************************************************
 *
 *	FUNCTION: PrvValidateEntries
 *
 *	DESCRIPTION: Validates entries in a database
 *
 *	PARAMETERS: 
 *			hdrP - pointer to database header
 *			bufferSize - size of the buffer holding the database
 *
 *	RETURNS: Err - error code
 *
 *************************************************************/

static ErrCode PrvValidateEntries(DatabaseHdrPtr hdrP, UInt32 bufferSize)
{
	// We use entryP to refer to both records OR resources; it is declared
	// as a UInt8* so that we can do byte-arithmetic to increment the appropriate
	// amount depending on whether we are dealing with records or resources
	UInt8* 	entryP	= (UInt8*) (&hdrP->recordList.firstEntry);

	// Establish the upper and lower ranges for localChunkIDs.	These IDs are
	// offsets from the beginning of the buffer/file, so they can't be larger
	// than the size of the buffer/file, nor can they be smaller than the
	// starting offset of the chunks.
	UInt32		upperRange = bufferSize;
	UInt32		lowerRange = offsetof (DatabaseHdrType, recordList.firstEntry) +
								PrvGetEntrySize (hdrP) * hdrP->recordList.numRecords;

	// Create a mirror buffer to check for overlapping records.
	StMemory	marks;
	marks.Adopt ((char*) Platform::AllocateMemoryClear (bufferSize));

	// Create an array to keep track of the discovered resource type/ids.
	// This is so we can find duplicate entries.
	vector<ResTypeIDPair>	resPairs;

	// nextRecordListID not supported for now.
	if (hdrP->recordList.nextRecordListID != 0)
		return kError_BadDB_nextRecordListIDNonZero;

	for (UInt16 recordNumber = 0; recordNumber < hdrP->recordList.numRecords; ++recordNumber)
	{
		LocalID 	id;
		long		itsSize;	// Use a signed value to detect non-increasing LocalIDs

		// Set the error parameters in case an error occurs.
		Errors::SetParameter ("%record_number", recordNumber);

		// Check for negative sizes and get the LocalID.
		// ---------------------------------------------
		// Though the code is similar for resources and for records, we have separate
		// blocks so that we can safely cast to different types to access the localChunkID
		// field.
		if (PrvIsResources(hdrP))
		{
			RsrcEntryPtr rsrcEntryP = (RsrcEntryPtr) entryP;

			// Set the error parameters in case an error occurs.
			PrvSetResourceTypeIDParameters (rsrcEntryP->type, rsrcEntryP->id);

			// Check for duplicate resources.
			ResTypeIDPair	thisPair (rsrcEntryP->type, rsrcEntryP->id);
			if (find (resPairs.begin(), resPairs.end(), thisPair) != resPairs.end())
				return kError_BadDB_DuplicateResource;
			resPairs.push_back (thisPair);

			// Get the size of the resource.
			if (recordNumber < hdrP->recordList.numRecords - 1)
			{
				itsSize = (UInt32) rsrcEntryP[1].localChunkID;
				itsSize -= (UInt32) rsrcEntryP[0].localChunkID;
			}
			else
				itsSize = bufferSize - (UInt32) rsrcEntryP[0].localChunkID;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%res_size", itsSize);

			// Check for negative sizes (zero is NOT okay for resources).
			if (itsSize <= 0)
				return kError_BadDB_ResourceTooSmall;

			id = rsrcEntryP->localChunkID;
		}
		else
		{
			RecordEntryPtr recordEntryP = (RecordEntryPtr) entryP;
			id = recordEntryP->localChunkID;

			if (recordNumber < hdrP->recordList.numRecords - 1)
			{
				itsSize = (UInt32) recordEntryP[1].localChunkID;
				itsSize -= (UInt32) recordEntryP[0].localChunkID;
			}
			else
				itsSize = bufferSize - (UInt32) recordEntryP[0].localChunkID;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%res_size", itsSize);

			// Check for negative sizes (zero is okay for records).
			if (itsSize < 0)
				return kError_BadDB_RecordTooSmall;

			id = recordEntryP->localChunkID;
		}

		// Test that the beginning and ending of the chunk are in range.
		// Note the special case for zero-length objects appearing at
		// the end of the file.
		if (itsSize > 0)
		{
			if (id >= upperRange || id + itsSize < lowerRange)
				return PrvIsResources (hdrP)
					? kError_BadDB_ResourceOutOfRange
					: kError_BadDB_RecordOutOfRange;
		}
		else if (id > upperRange || id + itsSize < lowerRange)
			return PrvIsResources (hdrP)
				? kError_BadDB_ResourceOutOfRange
				: kError_BadDB_RecordOutOfRange;

		// Check that the range of this record is not occupied by
		// another record.
		for (UInt32 ii = id; ii < id + itsSize; ++ii)
		{
			if (marks.Get () [ii] != 0)
				return PrvIsResources (hdrP)
					? kError_BadDB_OverlappingResource
					: kError_BadDB_OverlappingRecord;

			marks.Get () [ii] = 1;	// Mark this byte as occupied.
		}

		// next entry
		entryP += PrvGetEntrySize(hdrP);
	}

	// If we got here, we're okay
	return errNone;
}


/************************************************************
 *
 *	FUNCTION: PrvValidateDatabase
 *
 *	DESCRIPTION: 
 *
 *	PARAMETERS: 
 *			hdrP - pointer to database header
 *			bufferSize - size of the buffer holding the database
 *
 *	RETURNS: Err - error code
 *
 *************************************************************/

static ErrCode PrvValidateDatabase (DatabaseHdrPtr hdrP, UInt32 size)
{
	ErrCode err;

	if (size < sizeof (DatabaseHdrType))
	{
		err = kError_BadDB_FileTooSmall;
		goto Exit;
	}

	Canonical (*hdrP);	// Header now in little-endian format

	err = ::PrvValidateDatabaseHeader (hdrP, size);
	if (err != errNone)
		goto Exit;

	if (PrvIsResources(hdrP))
	{
		// Byteswap the resource entry headers

		RsrcEntryPtr	rsrcP  = (RsrcEntryPtr) (&hdrP->recordList.firstEntry);
		for (UInt16 ii = 0; ii < hdrP->recordList.numRecords; ++ii)
		{
			Canonical (*rsrcP);
			++rsrcP;
		}
	}
	else
	{
		// Byteswap the record entry headers

		RecordEntryPtr	recordP  = (RecordEntryPtr) (&hdrP->recordList.firstEntry);
		for (UInt16 ii = 0; ii < hdrP->recordList.numRecords; ++ii)
		{
			Canonical (*recordP);
			++recordP;
		}
	}

	// Check out the record entries; make sure they're okay
	err = ::PrvValidateEntries (hdrP, size);

Exit:
	if (!err)
		Errors::ClearAllParameters ();

	return err;
}


/************************************************************
 *
 *	FUNCTION: DmCreateDatabaseFromImage
 *
 *	DESCRIPTION: Can be called to create an entire database
 *			from a single resource that contains in image of the database. 
 *			Usually, this would be called from an app's reset action code during
 *			boot.
 *
 *	PARAMETERS: 
 *			bufferP - pointer to locked resource containing image of the database
 *
 *	RETURNS: void
 *
 *	CREATED: 3/13/95
 *
 *************************************************************/

static ErrCode	My_DmCreateDatabaseFromImage(MemPtr bufferP, UInt32 size, UInt16& cardNo, LocalID& dbID)
{
	DmOpenRef		dbP = 0;
	ErrCode 			err = 0;
	UInt8* 		srcP;
	UInt8* 		dstP;
	DatabaseHdrPtr	hdrP = (DatabaseHdrPtr)bufferP;
	UInt16			i;
	UInt32			uniqueIDSeed;
	DmAccessPtr 	dbP2;
	DmOpenInfoPtr	openP2;
	DatabaseHdrPtr	hdrP2;

	// Get the size and create it
	err = DmCreateDatabase(cardNo, (Char*)hdrP->name, hdrP->creator, hdrP->type, 
						PrvIsResources(hdrP));
	if (err)
	{
		err = ConvertFromPalmError (err);
		goto Exit;
	}

	// Set Database info
	dbID = DmFindDatabase(cardNo, (Char*)hdrP->name);
	if (!dbID)
	{
		err = ConvertFromPalmError (dmErrCantFind);
		goto Exit;
	}

	err = DmSetDatabaseInfo(cardNo, dbID, 0,
				&hdrP->attributes, &hdrP->version, &hdrP->creationDate,
				&hdrP->modificationDate, &hdrP->lastBackupDate,
				&hdrP->modificationNumber, 0,
				0, &hdrP->type,
				&hdrP->creator);

	if (err)
	{
		err = ConvertFromPalmError (err);
		goto Exit;
	}
	
	
	// Open the new database
	dbP = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	if (!dbP)
	{
		err = ConvertFromPalmError (dmErrCantOpen);
		goto Exit;
	}

	// The new database has been created with a random number in it.
	// We don't want the records copied into the new database to have
	// different unique id's between each hard reset or else the records
	// with the same contents but with different unique id's will appear
	// distinct to HotSync and so will accumulate with hard resets and syncs.
	// To stop this behavior we always set the seed to a constant number,
	// copy the records into the new database, and then we restore the
	// random seed.
	MemSemaphoreReserve(true);
//	uniqueIDSeed = ((DmAccessPtr)dbP)->openP->hdrP->uniqueIDSeed;
//	((DmAccessPtr)dbP)->openP->hdrP->uniqueIDSeed = dmDefaultRecordsID << 12;
	dbP2	= (DmAccessPtr) get_real_address((emuptr) dbP);
	openP2	= (DmOpenInfoPtr) get_real_address(do_get_mem_long(&dbP2->openP));
	hdrP2	= (DatabaseHdrPtr) get_real_address(do_get_mem_long(&openP2->hdrP));
	uniqueIDSeed = hdrP2->uniqueIDSeed;
	do_put_mem_long(&hdrP2->uniqueIDSeed, dmDefaultRecordsID << 12);
	MemSemaphoreRelease(true);

	
	//------------------------------------------------------------
	// If it's a resource database, Add the Resources
	//------------------------------------------------------------
	if (PrvIsResources(hdrP)) {
		RsrcEntryPtr	rsrcP;
		UInt32			resSize;
		MemHandle		newResH;

		Platform::LFPW_SetFileMax (hdrP->recordList.numRecords);

		// Pointer to first resource
		rsrcP  = (RsrcEntryPtr)(&hdrP->recordList.firstEntry);
		for (i = 0; i < hdrP->recordList.numRecords; i++)
		{
			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%record_number", i);
			PrvSetResourceTypeIDParameters (rsrcP->type, rsrcP->id);
		
			Platform::LFPW_SetFileValue (i);

			// Calculate the size of the resource
			if (i < hdrP->recordList.numRecords - 1)
			{
				resSize = (UInt32)rsrcP[1].localChunkID;
				resSize -= (UInt32)rsrcP[0].localChunkID;
			}
			else
				resSize = size - (UInt32)rsrcP[0].localChunkID;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%res_size", resSize);

			// (Should never get here, since the .prc file validity check
			//	above should filter out files with zero-sized resources.)
			if (resSize == 0)
			{
				rsrcP++;
				continue;
			}

			// Allocate the new resource
			newResH = (MemHandle) DmNewResource(dbP, rsrcP->type, rsrcP->id, resSize);
			if (!newResH)
			{
				err = kError_BadDB_ResourceMemError;
				goto Exit;
			}

			// Copy the data in
			srcP = (UInt8*)bufferP + (UInt32)rsrcP->localChunkID;
			dstP = (UInt8*)MemHandleLock((MemHandle) newResH);
			DmWrite(dstP, 0, srcP, resSize);
			MemHandleUnlock((MemHandle) newResH);
			// Done with it
			DmReleaseResource((MemHandle) newResH);

			// next one
			rsrcP++;
		}
	}

	//------------------------------------------------------------
	// If it's a Record database, Add the Records
	//------------------------------------------------------------
	else {
		MemHandle		newRecH;
		RecordEntryPtr	recordP;
		UInt32			recSize;
		LocalID 		appInfoID;

		// Pointer to first record
		recordP  = (RecordEntryPtr)(&hdrP->recordList.firstEntry);

		// If there's an appInfo, add that
		if (hdrP->appInfoID)
		{
			if (hdrP->sortInfoID)
				recSize = (UInt32)hdrP->sortInfoID - (UInt32)hdrP->appInfoID;
			else if (hdrP->recordList.numRecords > 0)
				recSize = (UInt32)recordP[0].localChunkID - (UInt32)hdrP->appInfoID;
			else
				recSize = size - (UInt32)hdrP->appInfoID;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%res_size", recSize);

			// Allocate the chunk
			newRecH = (MemHandle)DmNewHandle(dbP, recSize);
			if (!newRecH)
			{
				err = kError_BadDB_AppInfoMemError;
				goto Exit;
			}

			// Copy the data in
			srcP = (UInt8*)bufferP + (UInt32)hdrP->appInfoID;
			dstP = (UInt8*)MemHandleLock((MemHandle) newRecH);
			DmWrite(dstP, 0, srcP, recSize);
			MemHandleUnlock((MemHandle) newRecH);

			// Store it in the header
			appInfoID = MemHandleToLocalID((MemHandle) newRecH);
			DmSetDatabaseInfo(0, dbID, 0,
				0, 0, 0,
				0, 0,
				0, &appInfoID,
				0, 0, 
				0);
		}

		Platform::LFPW_SetFileMax (hdrP->recordList.numRecords);

		// Add Each of the records.
		for (i=0; i<hdrP->recordList.numRecords; i++)
		{
			UInt16		index;
			UInt16		attr;
			UInt32		id;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%record_number", i);

			Platform::LFPW_SetFileValue (i);

			// Calculate the size of the record
			if (i < hdrP->recordList.numRecords - 1)
			{
				recSize = (UInt32)recordP[1].localChunkID;
				recSize -= (UInt32)recordP[0].localChunkID;
			}
			else
				recSize = size - (UInt32)recordP[0].localChunkID;

			// Set the error parameters in case an error occurs.
			Errors::SetParameter ("%res_size", recSize);

			//	Only add non-nil records. When the default databases are created
			//	they sometimes have nil records if the user had to delete a
			//	record they didn't want.
			if (recSize)
			{
				// Allocate the new record
				index = 0xFFFF;
				newRecH = (MemHandle)DmNewRecord(dbP, &index, recSize);
				if (!newRecH)
				{
					err = kError_BadDB_RecordMemError;
					goto Exit;
				}

				// Copy the data in
				srcP = (UInt8*)bufferP + (UInt32)recordP->localChunkID;
				dstP = (UInt8*)MemHandleLock((MemHandle) newRecH);
				DmWrite(dstP, 0, srcP, recSize);
				MemHandleUnlock((MemHandle) newRecH);

				// Set the attributes
				attr = recordP->attributes;
				id = recordP->uniqueID[0];
				id = (id << 8) | recordP->uniqueID[1];
				id = (id << 8) | recordP->uniqueID[2];
				DmSetRecordInfo(dbP, index, &attr, &id);

				// Done with it
				DmReleaseRecord(dbP, index, true);
			}

			// next one
			recordP++;
		}
	}


		// Fixup the modification # to match what was in the image.
	DmSetDatabaseInfo(cardNo, dbID, 0,
				0, 0,0,
				0, 0,
				&hdrP->modificationNumber, 0,
				0, 0,
				0);


	// Restore the database's uniqueIDSeed to it's random value.
	MemSemaphoreReserve(true);
//	((DmAccessPtr)dbP)->openP->hdrP->uniqueIDSeed = uniqueIDSeed;
	// (Refresh the pointers; they may have changed since we initialized them.)
	dbP2	= (DmAccessPtr) get_real_address((emuptr) dbP);
	openP2	= (DmOpenInfoPtr) get_real_address(do_get_mem_long(&dbP2->openP));
	hdrP2	= (DatabaseHdrPtr) get_real_address(do_get_mem_long(&openP2->hdrP));
	hdrP2->uniqueIDSeed = uniqueIDSeed;
	MemSemaphoreRelease(true);

	// If we got to here, no error 
	::DmCloseDatabase(dbP);
	dbP = 0;


Exit:
	if (dbP)
	{
		::DmCloseDatabase(dbP);
		if (dbID)
			DmDeleteDatabase(cardNo, dbID);
	}

	return err;
}


// ---------------------------------------------------------------------------
//		 LoadPalmFile
// ---------------------------------------------------------------------------
// Loads a Palm OS program or database file into the emulator.

void LoadPalmFile (EmStreamFile& appFile)
{
	UInt16 cardNo = 0;
	LocalID dbID = 0;

	// Get the size of the file.

	int32 fileSize = appFile.GetLength ();

	// Create a buffer to hold the file data.

	StMemory	fileBuffer (fileSize);

	// Load the file into memory.

	appFile.GetBytes (fileBuffer.Get (), fileSize);

	// Check for corrupt databases.

	ErrCode errCode = ::PrvValidateDatabase ((DatabaseHdrPtr) fileBuffer.Get (), fileSize);
	Errors::ThrowIfError (errCode);

	// Let the emulated ROM access the memory in the file buffer.

//	StMemoryMapper	mapper (fileBuffer.Get (), fileSize);

	// See if there's already a database with this name.

	LocalID tempID = DmFindDatabase (cardNo, fileBuffer.Get ());

	// If the database already exists, delete it.

	if (tempID != 0)
	{
		Err	err = DmDeleteDatabase (cardNo, tempID);

		if (err != dmErrROMBased)
			Errors::ThrowIfPalmError (err);
	}

	// Create the new database.

	errCode = My_DmCreateDatabaseFromImage (fileBuffer.Get (), fileSize, cardNo, dbID);
	Errors::ThrowIfError (errCode);
}



// ---------------------------------------------------------------------------
//		 LoadPalmFileList
// ---------------------------------------------------------------------------

void LoadPalmFileList (const EmFileRefList& fileList)
{
	if (fileList.size () > 0)
	{
		Bool	restoreLauncher =	::PrvLauncherActive () &&
									::PrvSwitchToApp (sysFileTApplication, sysFileCCalculator);

		Platform::LFPW_Open ();

		Platform::LFPW_SetNumFiles (fileList.size ());

		EmFileRef	fileRef;

		try
		{
			EmFileRefList::const_iterator	iter = fileList.begin ();
			while (iter != fileList.end () && !Platform::LFPW_UserAborts())
			{
				fileRef = *iter;
				EmStreamFile	stream (fileRef, kOpenExistingForRead);

				Platform::LFPW_StartNewFile (fileRef.GetName ().c_str ());

				::LoadPalmFile (stream);

				gEmuPrefs->UpdatePRCMRU (fileRef);

				++iter;
			}
		}
		catch (ErrCode errCode)
		{
			if (fileList.size () > 1)
			{
				Errors::ReportIfError (kStr_CmdInstallMany, errCode, 0, false);
			}
			else
			{
				Errors::SetParameter ("%filename", fileRef.GetName ());
				Errors::ReportIfError (kStr_CmdInstall, errCode, 0, false);
			}
		}

		Platform::LFPW_Close ();

		if (restoreLauncher)
		{
			::PrvSwitchToApp (sysFileTApplication, sysFileCLauncher);
		}
	}
}


/************************************************************
 * Export a resource or record database as a normal data file in 
 *	Pilot format.
 *************************************************************/

static void My_ShlExportAsPilotFile(EmStreamFile& fh, UInt16 cardNo, const char* nameP)
{
	DmOpenRef		dbP = 0;
	Err 			err;
	UInt16				numRecords;
	int 			size;
	DatabaseHdrPtr	hdrP;
	long				offset;
	int 			i;
	
	
try
{
	//------------------------------------------------------------
	// Locate the Database and see what kind of database it is
	//------------------------------------------------------------

	LocalID dbID;
	dbID = ::DmFindDatabase(cardNo, (Char*) nameP);
	if (!dbID) Errors::ThrowIfPalmError(DmGetLastErr());
	
	UInt16		attributes, version;
	UInt32		crDate, modDate, bckUpDate;
	UInt32		type, creator;
	UInt32		modNum;
	LocalID appInfoID, sortInfoID;
	err = ::DmDatabaseInfo(cardNo, dbID, 0,
				&attributes, &version, &crDate,
				&modDate, &bckUpDate,
				&modNum, &appInfoID,
				&sortInfoID, &type,
				&creator);
	Errors::ThrowIfPalmError(err);
	
	dbP = ::DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
	if (!dbP) Errors::ThrowIfPalmError(DmGetLastErr());


	//------------------------------------------------------------
	// Write out a resource database
	//------------------------------------------------------------
	if (PrvIsResources(attributes)) {
		DmResType	resType;
		DmResID		resID;
		UInt32		resSize;

		numRecords = DmNumRecords(dbP);
		
		size = sizeof(DatabaseHdrType) + numRecords * sizeof(RsrcEntryType);
		StMemory	outP (size, true);
		
		// Fill in header
		hdrP = (DatabaseHdrPtr)outP.Get ();
		
		strcpy((char*)hdrP->name, nameP);
		hdrP->attributes = attributes;
		hdrP->version = version;
		hdrP->creationDate = crDate;
		hdrP->modificationDate = modDate;
		hdrP->lastBackupDate = bckUpDate;
		hdrP->modificationNumber = modNum;
		hdrP->appInfoID = 0;
		hdrP->sortInfoID = 0;
		hdrP->type = type;
		hdrP->creator = creator;
	
		hdrP->recordList.nextRecordListID = 0;
		hdrP->recordList.numRecords = numRecords;
	
	
		// Fill in the info on each resource into the header
		RsrcEntryPtr	entryP = (RsrcEntryPtr)(&hdrP->recordList.firstEntry);
		offset = size;
		for (i=0; i<numRecords; i++)  {
			
			DmResourceInfo(dbP, i, &resType, &resID, NULL);
			MemHandle	resH = DmGetResourceIndex(dbP, i);
			if (resH)  {
				resSize = MemHandleSize(resH);
				DmReleaseResource(resH);
				}
			else {
				resSize = 0;
				}
			entryP->type = resType;
			entryP->id =resID;
			entryP->localChunkID = offset;
			
			offset += resSize;
			entryP++;
			}

		// Byteswap the resource entry headers
		entryP	= (RsrcEntryPtr)(&hdrP->recordList.firstEntry);
		for (i=0; i<hdrP->recordList.numRecords; i++)
		{
			Canonical(*entryP);
			entryP++;
		}
		Canonical(*hdrP);	// Now in big-endian format
	
		// Write out  entry table
		fh.PutBytes (outP.Get (), size);

		// Write out each resource
		for (i=0; i<numRecords; i++) {
			MemHandle	srcResH;
			LocalID resChunkID;
			
			DmResourceInfo(dbP, i, &resType, &resID, &resChunkID);

			if (resChunkID) {
				UInt32	srcP;
				srcResH = DmGetResourceIndex(dbP, i);
				if (!srcResH) {
					Errors::ThrowIfPalmError(-1);
					}
				resSize = MemHandleSize(srcResH);

				StMemory	outP (resSize);
				srcP = (UInt32)MemHandleLock(srcResH);
				uae_memcpy((void*) outP.Get (), srcP, resSize);
				MemPtrUnlock((MemPtr)srcP);
				fh.PutBytes (outP.Get (), resSize);

				DmReleaseResource(srcResH);
				}
			}

		} // Resource database
		
		

	//------------------------------------------------------------
	// Write out a records database
	//------------------------------------------------------------
	else {
		UInt16		attr;
		UInt32		uniqueID;
		MemHandle	srcH;
		
		numRecords = DmNumRecords(dbP);
		
		size = sizeof(DatabaseHdrType) + numRecords * sizeof(RecordEntryType);
		StMemory	outP(size);
		
		// Fill in header
		hdrP = (DatabaseHdrPtr)outP.Get ();
		
		strcpy((char*)hdrP->name, nameP);
		hdrP->attributes = attributes;
		hdrP->version = version;
		hdrP->creationDate = crDate;
		hdrP->modificationDate = modDate;
		hdrP->lastBackupDate = bckUpDate;
		hdrP->modificationNumber = modNum;
		hdrP->appInfoID = 0;
		hdrP->sortInfoID = 0;
		hdrP->type = type;
		hdrP->creator = creator;
	
		hdrP->recordList.nextRecordListID = 0;
		hdrP->recordList.numRecords = numRecords;
		
		
		// Get the size of the appInfo and sort Info if they exist
		offset = size;
		MemHandle	appInfoH=0, sortInfoH=0;
		UInt32		appInfoSize=0, sortInfoSize=0;
		if (appInfoID) {
			hdrP->appInfoID = offset;
			appInfoH = (MemHandle)MemLocalIDToGlobal(appInfoID, cardNo);
			if (!appInfoH) {
				Errors::ThrowIfPalmError(-1);
				}
			appInfoSize = MemHandleSize(appInfoH);
			offset += appInfoSize;
			}
			
		if (sortInfoID) {
			hdrP->sortInfoID = offset;
			sortInfoH = (MemHandle)MemLocalIDToGlobal(sortInfoID, cardNo);
			if (!sortInfoH) {
				Errors::ThrowIfPalmError(-1);
				}
			sortInfoSize = MemHandleSize(sortInfoH);
			offset += sortInfoSize;
			}			
	
	
		// Fill in the info on each resource into the header
		RecordEntryPtr	entryP = (RecordEntryPtr)(&hdrP->recordList.firstEntry);
		for (i=0; i<numRecords; i++)  {
			
			err = DmRecordInfo(dbP, i, &attr, &uniqueID, 0);
			if (err) {
				Errors::ThrowIfPalmError(-1);
				}
			
			srcH = DmQueryRecord(dbP, i);

			entryP->localChunkID = offset;
			entryP->attributes = attr;
			entryP->uniqueID[0] = (uniqueID >> 16) & 0x00FF;	// vmk 10/16/95 fixed: && 0x00FF --> & 0x00FF
			entryP->uniqueID[1] = (uniqueID >> 8) & 0x00FF;
			entryP->uniqueID[2] = uniqueID	& 0x00FF;
			
			if (srcH)
				offset += MemHandleSize(srcH);
			entryP++;
			}

		// Byteswap the resource entry headers
		entryP	= (RecordEntryPtr)(&hdrP->recordList.firstEntry);
		for (i=0; i<hdrP->recordList.numRecords; i++)
		{
			Canonical(*entryP);
			entryP++;
		}
		Canonical(*hdrP);	// Now in big-endian format
	
		// Write out  entry table
		fh.PutBytes (outP.Get(), size);

		// Write out the appInfo followed by sortInfo, if they exist
		if (appInfoID && appInfoSize) {
			UInt32		srcP;
			StMemory	outP(appInfoSize);
			srcP = (UInt32)MemHandleLock(appInfoH);
			uae_memcpy((void*) outP.Get(), srcP, appInfoSize);
			MemPtrUnlock((MemPtr)srcP);
			fh.PutBytes (outP.Get(), appInfoSize);
			}
			
		if (sortInfoID && sortInfoSize) {
			UInt32		srcP;
			StMemory	outP(sortInfoSize);
			srcP = (UInt32)MemHandleLock(sortInfoH);
			uae_memcpy((void*) outP.Get(), srcP, sortInfoSize);
			MemPtrUnlock((MemPtr)srcP);
			fh.PutBytes (outP.Get(), sortInfoSize);
			}
	
	
	
		// Write out each record
		for (i=0; i<numRecords; i++) {
			UInt32		recSize;
			
			err = DmRecordInfo(dbP, i, &attr, &uniqueID, 0);
			if (err) {
				Errors::ThrowIfPalmError(-1);
				}
				
			srcH = DmQueryRecord(dbP, i);
				
			if (srcH) {
				UInt32	srcP;
				
				recSize = MemHandleSize(srcH);
				StMemory	outP(recSize);
				srcP = (UInt32)MemHandleLock(srcH);
				uae_memcpy((void*) outP.Get(), srcP, recSize);
				MemPtrUnlock((MemPtr)srcP);
				fh.PutBytes (outP.Get(), recSize);
				}
			}
	
			
		}
		
	// Clean up
	::DmCloseDatabase(dbP);
}
catch (...)
{
	if (dbP)
		::DmCloseDatabase (dbP);

	throw;
}
}


// ---------------------------------------------------------------------------
//		 SavePalmFile
// ---------------------------------------------------------------------------
// Saves a Palm OS program or database file to a file.

void SavePalmFile (EmStreamFile& appFile, UInt16 cardNo, const char* databaseName)
{
	My_ShlExportAsPilotFile (appFile, cardNo, databaseName);
}



Bool PrvLauncherActive (void)
{
	if (!Patches::UIInitialized ())
		return false;

	EmuAppInfo	info = Patches::GetRootAppInfo ();

	UInt32	type;
	UInt32	creator;
	/*Err 	err =*/ ::DmDatabaseInfo (info.fCardNo, info.fDBID,
									NULL,		// const Char* nameP,
									NULL,		// UIntPtr attributesP,
									NULL,		// UIntPtr versionP,
									NULL,		// ULongPtr crDateP,
									NULL,		// ULongPtr modDateP,
									NULL,		// ULongPtr bckUpDateP,
									NULL,		// ULongPtr modNumP,
									NULL,		// LocalID* appInfoIDP,
									NULL,		// LocalID* sortInfoIDP,
									&type,		// ULongPtr typeP,
									&creator);	// ULongPtr creatorP)

	if (type == sysFileTApplication && creator == sysFileCLauncher)
		return true;

	return false;
}


Bool PrvSwitchToApp (UInt32 /*type*/, UInt32 /*creator*/)
{
#if 0
	Err 				err;
	DmSearchStateType	searchInfo;
	UInt16				cardNo;
	LocalID 			dbID;

	err = ::DmGetNextDatabaseByTypeCreator (true, &searchInfo,
			 type, creator, true, &cardNo, &dbID);

	if (err == errNone)
	{
		err = ::SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
	}

	if (err == errNone)
	{
		Patches::StartWatchingAppLaunch ();

		do {
			Emulator::ExecuteUntilATrap ();
		} while (!Patches::AppFinishedLaunching ());
	}

	return err == errNone;
#else
	return false;
#endif
}
