/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Marshal.h"

#include "Byteswapping.h"		// Canonical

#if PLATFORM_UNIX || PLATFORM_MAC
#include <netinet/in.h>			// ntohl, ntohs
#endif

// -------------------------
// ----- Binary buffer -----
// -------------------------

void* Marshal::GetBuffer (emuptr p, long len)
{
	void*	result = NULL;

	if (p)
	{
		result = Platform::AllocateMemory (len);

		if (result)
		{
			uae_memcpy (result, p, len);
		}
	}

	return result;
}

#if 0	// inline
template <class T>
void Marshal::PutBuffer (emuptr p, T*& buf, long len)
{
	if (p)
	{
		uae_memcpy (p, (void*) buf, len);
		Platform::DisposeMemory (buf);
		buf = NULL;
	}
}
#endif


// ---------------------
// ----- EventType -----
// ---------------------

EventType Marshal::GetEventType (emuptr p)
{
	EventType	event;
	memset (&event, 0, sizeof (event));

	if (p)
	{
		event.eType		= (eventsEnum) get_word (p + offsetof (EventType, eType));
		event.penDown	= get_byte (p + offsetof (EventType, penDown));
		event.screenX	= get_word (p + offsetof (EventType, screenX));
		event.screenY	= get_word (p + offsetof (EventType, screenY));

		// Aargh! Need to do all the union structs!
	}

	return event;
}


void Marshal::PutEventType (emuptr p, EventType& event)
{
	if (p)
	{
		put_word (p + offsetof (EventType, eType),		event.eType);
		put_byte (p + offsetof (EventType, penDown),	event.penDown);
		put_word (p + offsetof (EventType, screenX),	event.screenX);
		put_word (p + offsetof (EventType, screenY),	event.screenY);

		// Aargh! Need to do all the union structs!
	}
}


// -----------------------------
// ----- NetSocketAddrType -----
// -----------------------------

NetSocketAddrType Marshal::GetNetSocketAddrType (emuptr p)
{
	NetSocketAddrType	netSocketAddr;
	memset (&netSocketAddr, 0, sizeof (netSocketAddr));

	if (p)
	{
		// We should be able to pretty much just block-copy this sucker without
		// having to worry too much about byte-swapping the contents.  A generic
		// address block is a 2-byte family in HBO followed by 14 bytes of data.
		// For Internet addresses, this 14 bytes of data consists of an IP address
		// and a socket in NBO, so no byteswapping is needed.  For Raw addresses,
		// Palm OS has defined the contents to be a 2-byte instance and a 4-byte
		// creator; there's no mention of their byte-ordering, so for now let's
		// assume NBO as well.
		//
		// By doing a blind memcpy of the contents, we relieve ourselves of the
		// responsibity of determining if they're actually initialized (if they're
		// not initialized, the family field may contain garbage, and we may go
		// off in the weeds trying to interpret the rest of the contents).

		switch (get_word (p))
		{
			case netSocketAddrRaw:
				uae_memcpy ((void*) &netSocketAddr, p, sizeof (NetSocketAddrRawType));
				break;

			case netSocketAddrINET:
				uae_memcpy ((void*) &netSocketAddr, p, sizeof (NetSocketAddrINType));
				break;

			default:
				// Do the best we can...
				uae_memcpy ((void*) &netSocketAddr, p, sizeof (NetSocketAddrType));
				break;
		}

		Canonical (netSocketAddr.family);
	}

	return netSocketAddr;
}


void Marshal::PutNetSocketAddrType (emuptr p, NetSocketAddrType& netSocketAddr)
{
	if (p)
	{
		Int16	family = netSocketAddr.family;

		Canonical (netSocketAddr.family);

		switch (family)
		{
			case netSocketAddrRaw:
				uae_memcpy (p, (void*) &netSocketAddr, sizeof (NetSocketAddrRawType));
				break;

			case netSocketAddrINET:
				uae_memcpy (p, (void*) &netSocketAddr, sizeof (NetSocketAddrINType));
				break;

			default:
				// Do the best we can...
				uae_memcpy (p, (void*) &netSocketAddr, sizeof (NetSocketAddrType));
				break;
		}

		Canonical (netSocketAddr.family);
	}
}


// --------------------------
// ----- NetIOParamType -----
// --------------------------

NetIOParamType Marshal::GetNetIOParamType (emuptr p)
{
	NetIOParamType	netIOParam;
	memset (&netIOParam, 0, sizeof (netIOParam));

	if (p)
	{
		netIOParam.addrLen			= get_word (p + offsetof (NetIOParamType, addrLen));
		netIOParam.iovLen			= get_word (p + offsetof (NetIOParamType, iovLen));
		netIOParam.accessRightsLen	= get_word (p + offsetof (NetIOParamType, accessRightsLen));

		netIOParam.addrP			= (UInt8*) GetBuffer (p + offsetof (NetIOParamType, addrP), netIOParam.addrLen);
		if (netIOParam.addrP)
			Canonical (((NetSocketAddrType*) (netIOParam.addrP))->family);

		netIOParam.accessRights		= (UInt8*) GetBuffer (p + offsetof (NetIOParamType, accessRights), netIOParam.accessRightsLen);

		netIOParam.iov				= (NetIOVecPtr) Platform::AllocateMemory (netIOParam.iovLen * sizeof (NetIOVecType));
		for (UInt16 ii = 0; ii < netIOParam.iovLen; ++ii)
		{
			netIOParam.iov[ii].bufLen	= get_word (p + offsetof (NetIOParamType, iovLen) + ii * sizeof (NetIOVecType) + offsetof (NetIOVecType, bufLen));
			netIOParam.iov[ii].bufP		= (UInt8*) GetBuffer (p + offsetof (NetIOParamType, iovLen) + ii * sizeof (NetIOVecType) + offsetof (NetIOVecType, bufP), netIOParam.iov[ii].bufLen);
		}
	}

	return netIOParam;
}


void Marshal::PutNetIOParamType (emuptr p, NetIOParamType& netIOParam)
{
	if (p)
	{
		put_word (p + offsetof (NetIOParamType, addrLen),			netIOParam.addrLen);
		put_word (p + offsetof (NetIOParamType, iovLen),			netIOParam.iovLen);
		put_word (p + offsetof (NetIOParamType, accessRightsLen),	netIOParam.accessRightsLen);

		if (netIOParam.addrP)
			Canonical (((NetSocketAddrType*) (netIOParam.addrP))->family);
		PutBuffer (p + offsetof (NetIOParamType, addrP),			netIOParam.addrP,			netIOParam.addrLen);
		if (netIOParam.addrP)
			Canonical (((NetSocketAddrType*) (netIOParam.addrP))->family);

//		PutBuffer (p + offsetof (NetIOParamType, iov),				netIOParam.iov,				netIOParam.iovLen);
		PutBuffer (p + offsetof (NetIOParamType, accessRights),		netIOParam.accessRights,	netIOParam.accessRightsLen);

		for (UInt16 ii = 0; ii < netIOParam.iovLen; ++ii)
		{
			put_word (p + offsetof (NetIOParamType, iovLen) + ii * sizeof (NetIOVecType) + offsetof (NetIOVecType, bufLen), netIOParam.iov[ii].bufLen);
			PutBuffer (p + offsetof (NetIOParamType, iovLen) + ii * sizeof (NetIOVecType) + offsetof (NetIOVecType, bufP), netIOParam.iov[ii].bufP, netIOParam.iov[ii].bufLen);
		}
		Platform::DisposeMemory (netIOParam.iov);
	}
}


// ------------------------------
// ----- NetHostInfoBufType -----
// ------------------------------

/*
typedef struct {
	NetHostInfoType	hostInfo;
	{
		Char*			nameP; -------------------------+
		Char**			nameAliasesP;-------------------|---+
		UInt16			addrType;                       |   |
		UInt16			addrLen;                        |   |
		UInt8**		addrListP;----------------------|---|---+
	}                                                   |   |   |
										                |   |   |
	Char			name[netDNSMaxDomainName+1];   <----+   |   |
                                                            |   |
	Char*			aliasList[netDNSMaxAliases+1];   <------+   |
	Char			aliases[netDNSMaxAliases][netDNSMaxDomainName+1];
                                                                |
	NetIPAddr*		addressList[netDNSMaxAddresses];   <--------+
	NetIPAddr		address[netDNSMaxAddresses];

	} NetHostInfoBufType, *NetHostInfoBufPtr;
*/

NetHostInfoBufType	Marshal::GetNetHostInfoBufType (emuptr p)
{
	NetHostInfoBufType	netHostInfoBufType;
	memset (&netHostInfoBufType, 0, sizeof (netHostInfoBufType));

	if (p)
	{
	}

	return netHostInfoBufType;
}


void Marshal::PutNetHostInfoBufType (emuptr p, NetHostInfoBufType& netHostInfoBuf)
{
	if (p)
	{
		int	ii;	// Goddampieceoshit VC++ doesn't scope loop variables right!

		// First copy over the easy fields (the ones with real values in them).

			// NetHostInfoBufType.hostInfo.addrType
		put_word (p + offsetof (NetHostInfoBufType, hostInfo.addrType),
					netHostInfoBuf.hostInfo.addrType);

			// NetHostInfoBufType.hostInfo.addrLen
		put_word (p + offsetof (NetHostInfoBufType, hostInfo.addrLen),
					netHostInfoBuf.hostInfo.addrLen);

			// NetHostInfoBufType.name
		uae_strcpy (p + offsetof (NetHostInfoBufType, name), netHostInfoBuf.name);

			// NetHostInfoBufType.aliases
		uae_memcpy (p + offsetof (NetHostInfoBufType, aliases),
					(void*) netHostInfoBuf.aliases,
					sizeof (netHostInfoBuf.aliases));

			// NetHostInfoBufType.address
			// Copy them one at a time to sort out endian issues.  Just how this
			// is supposed to be done is not clear (NetLib documentation says that
			// the addresses are in HBO, while the WinSock header says that the
			// addresses are supplied in NBO but returned in HBO), but the following
			// seems to work.
		for (ii = 0; ii < netDNSMaxAddresses; ++ii)
		{
			NetIPAddr	addr = ntohl (netHostInfoBuf.address[ii]);
			put_long (p + offsetof (NetHostInfoBufType, address[ii]), NetHToNL (addr));
		}

		// Second, set up the pointers to those values.

			// NetHostInfoType.hostInfo.nameP
		put_long (p + offsetof (NetHostInfoBufType, hostInfo.nameP),
					p + offsetof (NetHostInfoBufType, name));

			// NetHostInfoType.hostInfo.nameAliasesP
		put_long (p + offsetof (NetHostInfoBufType, hostInfo.nameAliasesP),
					p + offsetof (NetHostInfoBufType, aliasList));

			// NetHostInfoType.hostInfo.addrListP
		put_long (p + offsetof (NetHostInfoBufType, hostInfo.addrListP),
					p + offsetof (NetHostInfoBufType, addressList));

			// NetHostInfoType.aliasList
		for (ii = 0; ii < netDNSMaxAliases+1; ++ii)
		{
			if (netHostInfoBuf.aliasList[ii] != NULL)
			{
				put_long (p + offsetof (NetHostInfoBufType, aliasList[ii]),
					p + netHostInfoBuf.aliasList[ii] - (Char*) &netHostInfoBuf);
			}
			else
			{
				put_long (p + offsetof (NetHostInfoBufType, aliasList[ii]), UAE_NULL);
				break;
			}
		}

			// NetHostInfoType.addressList
		for (ii = 0; ii < netDNSMaxAddresses; ++ii)
		{
			if (netHostInfoBuf.addressList[ii] != NULL)
			{
				put_long (p + offsetof (NetHostInfoBufType, addressList[ii]),
							p + offsetof (NetHostInfoBufType, address[ii]));
			}
			else
			{
				put_long (p + offsetof (NetHostInfoBufType, addressList[ii]),
							UAE_NULL);
				break;
			}
		}
	}
}


// ------------------------------
// ----- NetServInfoBufType -----
// ------------------------------

NetServInfoBufType	Marshal::GetNetServInfoBufType (emuptr p)
{
	NetServInfoBufType	netServInfoBuf;
	memset (&netServInfoBuf, 0, sizeof (netServInfoBuf));

	if (p)
	{
	}

	return netServInfoBuf;
}

void Marshal::PutNetServInfoBufType (emuptr p, NetServInfoBufType& netServInfoBuf)
{
	if (p)
	{
		// =======================================================
		// Convert NetHostInfoBufType.hostInfo
		// =======================================================

		// Get a handy pointer to the servInfo field.
		emuptr	p2 = p + offsetof (NetServInfoBufType, servInfo);

		// -------------------------------------------------------
		// NetServInfoBufType.servInfo.nameP = &NetServInfoBufType.name
		// -------------------------------------------------------

		put_long (p2 + offsetof (NetServInfoType, nameP), p + offsetof (NetServInfoBufType, name));

		// -------------------------------------------------------
		// NetServInfoBufType.servInfo.nameAliasesP = &NetServInfoBufType.aliasList
		// -------------------------------------------------------

		put_long (p2 + offsetof (NetServInfoType, nameAliasesP), p + offsetof (NetServInfoBufType, aliasList));

		UInt16	port = ntohs (netServInfoBuf.servInfo.port);
		put_word (p2 + offsetof (NetServInfoType, port), NetHToNS (port));

		// -------------------------------------------------------
		// NetServInfoBufType.servInfo.protoP = &NetServInfoBufType.protoName
		// -------------------------------------------------------

		put_long (p2 + offsetof (NetServInfoType, protoP), p + offsetof (NetServInfoBufType, protoName));


		// =======================================================
		// Convert remaining NetHostInfoBufType fields
		// =======================================================

		// -------------------------------------------------------
		// name, protoName
		// -------------------------------------------------------

		uae_memcpy (p + offsetof (NetServInfoBufType, name), (void*) &netServInfoBuf.name, netServMaxName+1);
		uae_memcpy (p + offsetof (NetServInfoBufType, protoName), (void*) &netServInfoBuf.protoName, netProtoMaxName+1);

		// -------------------------------------------------------
		// aliasList, addressList
		// -------------------------------------------------------

		long	offset;
		long	index = 0;
		while (netServInfoBuf.aliasList[index])
		{
			// aliasList contains a list of string addresses.  Calculate the offset
			// of that string from the beginning of the NetHostInfoBufType.
			offset = ((char*) &netServInfoBuf.aliasList[index]) - (char*) &netServInfoBuf;

			// Use that offset to calculate the address of the corresponding location
			// in the destination.
			put_long (p + offsetof (NetServInfoBufType, aliasList) + index * sizeof (Char*), p + offset);

			// Copy over the string.
			uae_strcpy (p + offset, netServInfoBuf.aliasList[index]);

			++index;
		}
		put_long (p + offsetof (NetServInfoBufType, aliasList) + index * sizeof (Char*), UAE_NULL);
	}
}

//TRG:

ffs_diskfree_t Marshal::GetDiskFreeBufType (emuptr p)
{
	ffs_diskfree_t	diskfree;
	memset (&diskfree, 0, sizeof (ffs_diskfree_t));

	if (p)
	{
		diskfree.avail_clusters			= get_long (p + offsetof (ffs_diskfree_t, avail_clusters));
		diskfree.total_clusters			= get_long (p + offsetof (ffs_diskfree_t, total_clusters));

		diskfree.bytes_per_sector		= get_word (p + offsetof (ffs_diskfree_t, bytes_per_sector));
		diskfree.sectors_per_cluster	= get_word (p + offsetof (ffs_diskfree_t, sectors_per_cluster));
	}

	return diskfree;
}

void Marshal::PutDiskFreeBufType (emuptr p, ffs_diskfree_t& diskfree)
{
	if (p)
	{
		put_long (p + offsetof (ffs_diskfree_t, avail_clusters), diskfree.avail_clusters);
		put_long (p + offsetof (ffs_diskfree_t, total_clusters), diskfree.total_clusters);

		put_word (p + offsetof (ffs_diskfree_t, bytes_per_sector), diskfree.bytes_per_sector);
		put_word (p + offsetof (ffs_diskfree_t, sectors_per_cluster), diskfree.sectors_per_cluster);
	}
}


ffs_ffblk Marshal::GetFfblkBufType (emuptr p)
{
	ffs_ffblk	ff_blk;
	memset (&ff_blk, 0, sizeof (ffs_ffblk));

	if (p)
	{
		uae_memcpy ((void*) &ff_blk.ff_reserved,	p + offsetof (ffs_ffblk, ff_reserved), 21);

		ff_blk.ff_attrib	= get_byte (p + offsetof (ffs_ffblk, ff_attrib));
		ff_blk.ff_ftime		= get_word (p + offsetof (ffs_ffblk, ff_ftime));
		ff_blk.ff_fdate		= get_word (p + offsetof (ffs_ffblk, ff_fdate));
		ff_blk.ff_fsize		= get_long (p + offsetof (ffs_ffblk, ff_fsize));

		uae_memcpy ((void*) &ff_blk.ff_name,		p + offsetof (ffs_ffblk, ff_name), 13);
		uae_memcpy ((void*) &ff_blk.ff_longname,	p + offsetof (ffs_ffblk, ff_longname), 21);
	}

	return ff_blk;
}

void Marshal::PutFfblkBufType (emuptr p, ffs_ffblk& ff_blk)
{
	if (p)
	{
		uae_memcpy	(p + offsetof (ffs_ffblk, ff_reserved),	(void*) &ff_blk.ff_reserved, 13);

		put_byte	(p + offsetof (ffs_ffblk, ff_attrib),	ff_blk.ff_attrib);
		put_word	(p + offsetof (ffs_ffblk, ff_ftime),	ff_blk.ff_ftime);
		put_word	(p + offsetof (ffs_ffblk, ff_fdate),	ff_blk.ff_fdate);
		put_long	(p + offsetof (ffs_ffblk, ff_fsize),	ff_blk.ff_fsize);

		uae_memcpy	(p + offsetof (ffs_ffblk, ff_name),		(void*) &ff_blk.ff_name, 21);
		uae_memcpy	(p + offsetof (ffs_ffblk, ff_longname),	(void*) &ff_blk.ff_longname, 256);
	}
}

ffs_stat Marshal::GetStatBufType (emuptr p)
{
	ffs_stat	pstat;
	memset (&pstat, 0, sizeof (ffs_stat));

	if (p)
	{
		pstat.st_dev		= get_word  (p + offsetof (ffs_stat, st_dev));
		pstat.st_ino		= get_word  (p + offsetof (ffs_stat, st_ino));
		pstat.st_mode		= get_long  (p + offsetof (ffs_stat, st_mode));
		pstat.st_nlink		= get_word  (p + offsetof (ffs_stat, st_nlink));
		pstat.st_uid		= get_word  (p + offsetof (ffs_stat, st_uid));
		pstat.st_gid		= get_word  (p + offsetof (ffs_stat, st_gid));
		pstat.st_rdev		= get_word  (p + offsetof (ffs_stat, st_rdev));
		pstat.st_size		= get_long  (p + offsetof (ffs_stat, st_size));
		pstat.st_atime.date	= get_word  (p + offsetof (ffs_stat, st_atime.date));
		pstat.st_mtime.date	= get_word  (p + offsetof (ffs_stat, st_mtime.date));
		pstat.st_ctime.date	= get_word  (p + offsetof (ffs_stat, st_ctime.date));
		pstat.st_atime.time	= get_word  (p + offsetof (ffs_stat, st_atime.time));
		pstat.st_mtime.time	= get_word  (p + offsetof (ffs_stat, st_mtime.time));
		pstat.st_ctime.time	= get_word  (p + offsetof (ffs_stat, st_ctime.time));
		pstat.st_attr		= get_byte  (p + offsetof (ffs_stat, st_attr));
	}

	return pstat;
}

void Marshal::PutStatBufType (emuptr p, ffs_stat& pstat)
{
	if (p)
	{
		put_word (p + offsetof (ffs_stat, st_dev),			pstat.st_dev);
		put_word (p + offsetof (ffs_stat, st_ino),			pstat.st_ino);
		put_long (p + offsetof (ffs_stat, st_mode),			pstat.st_mode);
		put_word (p + offsetof (ffs_stat, st_nlink),		pstat.st_nlink);
		put_word (p + offsetof (ffs_stat, st_uid),			pstat.st_uid);
		put_word (p + offsetof (ffs_stat, st_gid),			pstat.st_gid);
		put_word (p + offsetof (ffs_stat, st_rdev),			pstat.st_rdev);
		put_long (p + offsetof (ffs_stat, st_size),			pstat.st_size);
		put_word (p + offsetof (ffs_stat, st_atime.date),	pstat.st_atime.date);
		put_word (p + offsetof (ffs_stat, st_mtime.date),	pstat.st_mtime.date);
		put_word (p + offsetof (ffs_stat, st_ctime.date),	pstat.st_ctime.date);
		put_word (p + offsetof (ffs_stat, st_atime.time),	pstat.st_atime.time);
		put_word (p + offsetof (ffs_stat, st_mtime.time),	pstat.st_mtime.time);
		put_word (p + offsetof (ffs_stat, st_ctime.time),	pstat.st_ctime.time);
		put_byte (p + offsetof (ffs_stat, st_attr),			pstat.st_attr);
	}
}
