/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#ifndef _PREFERENCEMGR_H_
#define _PREFERENCEMGR_H_

#include "EmFileRef.h"		 	// EmFileRefList
#include "EmulatorTypes.h"	 	// DatabaseInfoList, DeviceType
#include "omnithread.h" 		// omni_mutex, omni_mutex_lock
#include "EmFileRef.h"		 	// EmFileRef
#include "Skins.h"				// SkinNameList

#include <map>
#include <vector>


/*
	This file contains the routines for loading, saving, and accessing a collection
	of preferences/settings.

	All settings are accessed via the Preference class.  This is a templatized class
	that is specialized on the type of the setting and parameterized on the name of
	the setting.  This sounds nasty, but in practice is simple. For example, say you
	needed to access whether or not NetLib redirection was turned on.  You would use
	the following:

		Preference<bool>	pref(kPrefKeyRedirectNetLib);
		
		if (*pref)
		{
			// It's on.
		}

	Changing a value works similarly:

		Preference<DeviceType>	pref(kPrefKeyDeviceType);
		*pref = kDevicePalmIII;

	When the Preference object is destructed, it's new value is written back out to
	the Preference Manager.

	The Preference Manager is the "container" for all the preferences used by the
	application.  You can think of it as a map/dictionary/associative array.  You
	directly access the Preference Manager in order to create it, have it load the
	collection of preferences from disk, have it save them to disk, and to destroy
	it.  While there are GetPref and SetPref methods on this class, those are for
	the use of the Preference objects.

	All preferences are saved as text in the form:

		<key>=<value>

	<key> is any unique string.  It can consist of any text characters except for
	'='.  Keys can have subparts separated by '.', and can indicate array indices
	with "[#]".  For example:

		Scale=2
		GremlinInfo.fNumber=100
		PRC_MRU[0]=MyApplication.prc
		GremlinInfo.fAppList[0].name=AddressBook

	With the Preference class, it's possible to pull in all or part of the
	hierarchical information.  For instance:

		Preference<GremlinInfo>			pref("GremlinInfo");
		Preference<DatabaseInfoList>	pref("GremlinInfo.fAppList");
		Preference<DatabaseInfo>		pref("GremlinInfo.fAppList[0]");
		Preference<string>				pref("GremlinInfo.fAppList[0].name");
*/

typedef const char* PrefKeyType;

class BasePreference
{
	public:
								BasePreference		(PrefKeyType name, bool = true);
								BasePreference		(long index, bool = true);
		virtual 				~BasePreference 	(void);

		bool					Loaded				(void)	{ return fLoaded; }

		void					Flush				(void)	{ this->Save (); }

	protected:
		void					Load				(void);
		void					Save				(void);

	protected:
		virtual bool			DoLoad				(void) = 0;
		virtual void			DoSave				(void) = 0;

	protected:
		string					fName;
		bool					fLoaded;
		bool					fChanged;
		bool					fAcquireLock;
};

template <class T>
class Preference : public BasePreference
{
	public:
								Preference			(PrefKeyType name, bool = true);
								Preference			(long index, bool = true);
		virtual 				~Preference 		(void);

	// I would *like* to have these operators.	That way, I could pass in a
	// "Preference<Foo>" any place that accepts a "const Foo&" as a parameter.
	// However, CodeWarrior seems to use the conversion operators here in place
	// that accept merely a "Foo&", which has lead to some problems.
//								operator const T&() const	{ return *GetValue (); }
//								operator const T*() const	{ return GetValue (); }
		const T&				operator*() const			{ return *GetValue (); }
		const T*				operator->() const			{ return GetValue (); }

		const T*				GetValue () const			{ assert (fLoaded); return &fValue; }

		const T&				operator=(const T& rhs) 	{ fValue = rhs;
															  fLoaded = true;
															  fChanged = true;
															  this->Flush ();
															  return fValue; }

	protected:
		virtual bool			DoLoad				(void);
		virtual void			DoSave				(void);

	private:
		T						fValue;
};

typedef void (*PrefNotifyFunc)(PrefKeyType);
typedef StringList	PrefKeyList;

class Preferences
{
	public:
								Preferences 		(void);
		virtual 				~Preferences		(void);

		virtual void			Load				(void);
		virtual void			Save				(void);

// This should work, but CW doesn't appear to like it.
//	private:
//		template <class T>	friend class Preference;
	public:
		bool					GetPref 			(const string& key, string& value);
		void					SetPref 			(const string& key, const string& value);
		void					DeletePref			(const string& key);

	public:
		void					PushPrefix			(const string& prefix);
		void					PushPrefix			(long index);
		void					PopPrefix			(void);

		string					ExpandKey			(const string& name);
		string					AppendName			(string key, const string& name);

	public:
		void					AddNotification		(PrefNotifyFunc, PrefKeyType);
		void					AddNotification		(PrefNotifyFunc, const PrefKeyList&);
		void					RemoveNotification	(PrefNotifyFunc);
		void					RemoveNotification	(PrefNotifyFunc, PrefKeyType);
		void					RemoveNotification	(PrefNotifyFunc, const PrefKeyList&);
		void					DoNotify			(const string& key);

	protected:
		virtual EmFileRef		GetPrefRef			(void);
		virtual void			WriteBanner 		(FILE*);
		virtual bool			ReadBanner			(FILE*);
		virtual void			StripUnused 		(void);

	protected:
		typedef StringStringMap			PrefList;
		typedef PrefList::value_type	PrefPairType;
		typedef PrefList::iterator		iterator;

		PrefList				fPreferences;

		typedef StringList		PrefixType;
		PrefixType				fPrefixes;

		typedef map<PrefNotifyFunc, PrefKeyList> PrefNotifyList;

		PrefNotifyList			fNotifications;

	public:
		static omni_mutex		fgPrefsMutex;
};

extern Preferences* gPrefs;


class EmulatorPreferences : public Preferences
{
	public:
								EmulatorPreferences (void);
		virtual 				~EmulatorPreferences(void);

		virtual void			Load				(void);

		EmFileRef				GetIndPRCMRU		(int);
		EmFileRef				GetIndRAMMRU		(int);
		EmFileRef				GetIndMRU			(const EmFileRefList&, int);

		void					UpdatePRCMRU		(const EmFileRef&);
		void					UpdateRAMMRU		(const EmFileRef&);
		void					UpdateMRU			(EmFileRefList&, const EmFileRef&);

		void					RemovePRCMRU		(const EmFileRef&);
		void					RemoveRAMMRU		(const EmFileRef&);
		void					RemoveMRU			(EmFileRefList&, const EmFileRef&);

	protected:
		virtual void			WriteBanner 		(FILE*);
		virtual bool			ReadBanner			(FILE*);
		virtual void			StripUnused 		(void);

		int 					fCurLoggingType;
};

extern EmulatorPreferences* gEmuPrefs;


/*
	The FOR_EACH_PREF macro is used to manage the preferences keys.  Preferences
	are accessed via keys passed to the Preference constructor.  These keys have
	the symbolic form:

		kPrefKey<Label>

	The macro below is used to declare the keys used to access the settings, define
	the keys, initialize the settings to default values.
*/

#define FOR_EACH_PREF(DO_TO_PREF)								\
	FOR_EACH_UNINIT_PREF(DO_TO_PREF)							\
	FOR_EACH_INIT_PREF(DO_TO_PREF)

#define FOR_EACH_UNINIT_PREF(DO_TO_PREF)						\
	DO_TO_PREF(WindowLocation, unused, unused)					\
	DO_TO_PREF(GCWLocation, unused, unused)						\
	DO_TO_PREF(BackgroundColor, unused, unused) 				\
	DO_TO_PREF(HighlightColor, unused, unused)

#define FOR_EACH_INIT_PREF(DO_TO_PREF)							\
	DO_TO_PREF(CloseAction, CloseActionType, (kSaveAsk))		\
	DO_TO_PREF(CommPort, string, (""))							\
	DO_TO_PREF(RedirectNetLib, bool, (false))					\
	DO_TO_PREF(EnableSounds, bool, (false))						\
																\
	DO_TO_PREF(ReportCorruptedHeap, bool, (true))				\
	DO_TO_PREF(ReportFreeChunkAccess, bool, (true)) 			\
	DO_TO_PREF(ReportHardwareRegisterAccess, bool, (true))		\
	DO_TO_PREF(ReportInvalidPC, bool, (true))					\
	DO_TO_PREF(ReportLowMemoryAccess, bool, (true)) 			\
	DO_TO_PREF(ReportLowStackAccess, bool, (true))				\
	DO_TO_PREF(ReportMemMgrDataAccess, bool, (true))			\
	DO_TO_PREF(ReportMemMgrSemaphore, bool, (true)) 			\
	DO_TO_PREF(ReportScreenAccess, bool, (true))				\
	DO_TO_PREF(ReportStackAlmostOverflow, bool, (true)) 		\
	DO_TO_PREF(ReportStackOverflow, bool, (true))				\
	DO_TO_PREF(ReportStorageHeapAccess, bool, (true))			\
	DO_TO_PREF(ReportSysFatalAlert, bool, (true))				\
	DO_TO_PREF(ReportSystemGlobalAccess, bool, (true))			\
	DO_TO_PREF(ReportUnhandledException, bool, (true))			\
	DO_TO_PREF(ReportUnimplementedTrap, bool, (true))			\
	DO_TO_PREF(ReportInvalidLibraryRefNum, bool, (true))		\
	DO_TO_PREF(ReportUninitializedChunkAccess, bool, (true))	\
	DO_TO_PREF(ReportUninitializedStackAccess, bool, (true))	\
	DO_TO_PREF(ReportUnlockedChunkAccess, bool, (true)) 		\
																\
	DO_TO_PREF(LogErrorMessages, uint8, (0))					\
	DO_TO_PREF(LogWarningMessages, uint8, (0)) 					\
	DO_TO_PREF(LogGremlins, uint8, (0))							\
	DO_TO_PREF(LogCPUOpcodes, uint8, (0))						\
	DO_TO_PREF(LogEnqueuedEvents, uint8, (0))					\
	DO_TO_PREF(LogDequeuedEvents, uint8, (0))					\
	DO_TO_PREF(LogSystemCalls, uint8, (0)) 						\
	DO_TO_PREF(LogApplicationCalls, uint8, (0))					\
	DO_TO_PREF(LogSerial, uint8, (0))							\
	DO_TO_PREF(LogSerialData, uint8, (0))						\
	DO_TO_PREF(LogNetLib, uint8, (0))							\
	DO_TO_PREF(LogNetLibData, uint8, (0))						\
	DO_TO_PREF(LogExgMgr, uint8, (0))							\
	DO_TO_PREF(LogExgMgrData, uint8, (0))						\
	DO_TO_PREF(LogHLDebugger, uint8, (0))						\
	DO_TO_PREF(LogHLDebuggerData, uint8, (0))					\
	DO_TO_PREF(LogLLDebugger, uint8, (0))						\
	DO_TO_PREF(LogLLDebuggerData, uint8, (0))					\
	DO_TO_PREF(LogRPC, uint8, (0)) 								\
	DO_TO_PREF(LogRPCData, uint8, (0)) 							\
																\
	DO_TO_PREF(ReportMemoryLeaks, bool, (false))				\
	DO_TO_PREF(ReportLockedRecords, bool, (false))				\
																\
	DO_TO_PREF(DebuggerSocketPort, long, (6414))				\
	DO_TO_PREF(RPCSocketPort, long, (6415)) 					\
																\
	DO_TO_PREF(InterceptSysFatalAlert, bool, (true))			\
																\
	DO_TO_PREF(AskAboutStartMenu, bool, (true)) 				\
	DO_TO_PREF(StartMenuItem, EmFileRef, ())					\
																\
	DO_TO_PREF(FillNewBlocks, bool, (false))					\
	DO_TO_PREF(FillResizedBlocks, bool, (false))				\
	DO_TO_PREF(FillDisposedBlocks, bool, (false))				\
	DO_TO_PREF(FillStack, bool, (false))						\
																\
	DO_TO_PREF(LastConfiguration, Configuration, (kDevicePalmIII, 1024, EmFileRef())) 		\
																\
	DO_TO_PREF(GremlinInfo, GremlinInfo, ())					\
	DO_TO_PREF(HordeInfo, HordeInfo, ())						\
																\
	DO_TO_PREF(LastPSF, EmFileRef, ())							\
																\
	DO_TO_PREF(PRC_MRU, EmFileRefList, ())						\
	DO_TO_PREF(PSF_MRU, EmFileRefList, ())						\
																\
	DO_TO_PREF(Skins, SkinNameList, ())							\
	DO_TO_PREF(Scale, ScaleType, (2))							\
																\
	DO_TO_PREF(LastTracerType, string, (""))					\
	DO_TO_PREF(TracerTypes, string, (""))						\
																\
	DO_TO_PREF(FfsHome, string, (""))							\
																\
	DO_TO_PREF(UserName, string, ("Palm OS Emulator"))

// Declare all the keys
#define DECLARE_PREF_KEYS(name, type, init) extern PrefKeyType kPrefKey##name;
FOR_EACH_PREF(DECLARE_PREF_KEYS)

#endif	// _PREFERENCEMGR_H_
