/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Quantizer.h"

static long PrvGetRowBytes (LPBITMAPINFO dib);

struct NODE
{
	Bool	bIsLeaf;		// TRUE if node has no children
	uint32 nPixelCount;	// Number of pixels represented by this leaf
	uint32 nRedSum;		// Sum of red components
	uint32 nGreenSum;		// Sum of green components
	uint32 nBlueSum;		// Sum of blue components
	NODE*	pChild[8];		// Pointers to child nodes
	NODE*	pNext;			// Pointer to next reducible node
};

CQuantizer::CQuantizer (uint32 nMaxColors, uint32 nColorBits)
{
	assert (nColorBits <= 8);

	m_pTree = NULL;
	m_nLeafCount = 0;
	for (int i=0; i<=(int) nColorBits; i++)
		m_pReducibleNodes[i] = NULL;
	m_nMaxColors = nMaxColors;
	m_nColorBits = nColorBits;
}

CQuantizer::~CQuantizer ()
{
	if (m_pTree != NULL)
		DeleteTree (&m_pTree);
}

Bool CQuantizer::ProcessImage (/*HANDLE hImage*/LPBITMAPINFO bm, void* bits)
{
	uint32 rmask, gmask, bmask;
	int rright, gright, bright;
	int rleft, gleft, bleft;
	uint8* pbBits;
	uint32* pwBits;
	uint32* pdwBits;
	uint8 r, g, b;
	uint32 wColor;
	uint32 dwColor;
	int i, j;
	HDC hdc;
	uint8* pBuffer;
	BITMAPINFO bmi;

#if 0
	DIBSECTION ds;
	::GetObject (hImage, sizeof (ds), &ds);
	int nPad = ds.dsBm.bmWidthBytes - (((ds.dsBmih.biWidth *
		ds.dsBmih.biBitCount) + 7) / 8);
#else
	int nPad = ::PrvGetRowBytes (bm) - (((bm->bmiHeader.biWidth *
		bm->bmiHeader.biBitCount) + 7) / 8);
#endif

#if 0
   switch (ds.dsBmih.biBitCount) {
#else
   switch (bm->bmiHeader.biBitCount) {
#endif

#if 0
	case 1: // 1-bit DIB
	case 4: // 4-bit DIB
	case 8: // 8-bit DIB
		//
		// The strategy here is to use ::GetDIBits to convert the
		// image into a 24-bit DIB one scan line at a time. A pleasant
		// side effect of using ::GetDIBits in this manner is that RLE-
		// encoded 4-bit and 8-bit DIBs will be uncompressed.
		//
		hdc = ::GetDC (NULL);
		pBuffer = new uint8[ds.dsBmih.biWidth * 3];

		::ZeroMemory (&bmi, sizeof (bmi));
		bmi.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
		bmi.bmiHeader.biWidth = ds.dsBmih.biWidth;
		bmi.bmiHeader.biHeight = ds.dsBmih.biHeight;
		bmi.bmiHeader.biPlanes = 1;
		bmi.bmiHeader.biBitCount = 24;
		bmi.bmiHeader.biCompression = BI_RGB;

		for (i=0; i<ds.dsBmih.biHeight; i++) {
			::GetDIBits (hdc, (HBITMAP) hImage, i, 1, pBuffer, &bmi,
				DIB_RGB_COLORS);
			pbBits = pBuffer;
			for (j=0; j<ds.dsBmih.biWidth; j++) {
				b = *pbBits++;
				g = *pbBits++;
				r = *pbBits++;
				AddColor (&m_pTree, r, g, b, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount > m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount,
						m_pReducibleNodes);
			}
		}

		delete pBuffer;
		::ReleaseDC (NULL, hdc);
		break;

	case 16: // 16-bit DIB
		if (ds.dsBmih.biCompression == BI_BITFIELDS) {
			rmask = ds.dsBitfields[0];
			gmask = ds.dsBitfields[1];
			bmask = ds.dsBitfields[2];
		}
		else {
			rmask = 0x7C00;
			gmask = 0x03E0;
			bmask = 0x001F;
		}

		rright = GetRightShiftCount (rmask);
		gright = GetRightShiftCount (gmask);
		bright = GetRightShiftCount (bmask);

		rleft = GetLeftShiftCount (rmask);
		gleft = GetLeftShiftCount (gmask);
		bleft = GetLeftShiftCount (bmask);

		pwBits = (uint32*) ds.dsBm.bmBits;
		for (i=0; i<ds.dsBmih.biHeight; i++) {
			for (j=0; j<ds.dsBmih.biWidth; j++) {
				wColor = *pwBits++;
				b = (uint8) (((wColor & (uint32) bmask) >> bright) << bleft);
				g = (uint8) (((wColor & (uint32) gmask) >> gright) << gleft);
				r = (uint8) (((wColor & (uint32) rmask) >> rright) << rleft);
				AddColor (&m_pTree, r, g, b, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount > m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount, m_pReducibleNodes);
			}
			pwBits = (uint32*) (((uint8*) pwBits) + nPad);
		}
		break;
#endif

	case 24: // 24-bit DIB
#if 0
		pbBits = (uint8*) ds.dsBm.bmBits;
#else
		pbBits = (uint8*) bits;
#endif
#if 0
		for (i=0; i<ds.dsBmih.biHeight; i++) {
			for (j=0; j<ds.dsBmih.biWidth; j++) {
#else
		for (i=0; i<bm->bmiHeader.biHeight; i++) {
			for (j=0; j<bm->bmiHeader.biWidth; j++) {
#endif
				b = *pbBits++;
				g = *pbBits++;
				r = *pbBits++;
				AddColor (&m_pTree, r, g, b, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount > m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount, m_pReducibleNodes);
			}
			pbBits += nPad;
		}
		break;

#if 0
	case 32: // 32-bit DIB
		if (ds.dsBmih.biCompression == BI_BITFIELDS) {
			rmask = ds.dsBitfields[0];
			gmask = ds.dsBitfields[1];
			bmask = ds.dsBitfields[2];
		}
		else {
			rmask = 0x00FF0000;
			gmask = 0x0000FF00;
			bmask = 0x000000FF;
		}

		rright = GetRightShiftCount (rmask);
		gright = GetRightShiftCount (gmask);
		bright = GetRightShiftCount (bmask);

		pdwBits = (uint32*) ds.dsBm.bmBits;
		for (i=0; i<ds.dsBmih.biHeight; i++) {
			for (j=0; j<ds.dsBmih.biWidth; j++) {
				dwColor = *pdwBits++;
				b = (uint8) ((dwColor & bmask) >> bright);
				g = (uint8) ((dwColor & gmask) >> gright);
				r = (uint8) ((dwColor & rmask) >> rright);
				AddColor (&m_pTree, r, g, b, m_nColorBits, 0, &m_nLeafCount,
					m_pReducibleNodes);
				while (m_nLeafCount > m_nMaxColors)
					ReduceTree (m_nColorBits, &m_nLeafCount, m_pReducibleNodes);
			}
			pdwBits = (uint32*) (((uint8*) pdwBits) + nPad);
		}
		break;
#endif

	default: // Unrecognized color format
		return FALSE;
	}
	return TRUE;
}

int CQuantizer::GetLeftShiftCount (uint32 dwVal)
{
	int nCount = 0;
	for (int i=0; i<sizeof (uint32) * 8; i++) {
		if (dwVal & 1)
			nCount++;
		dwVal >>= 1;
	}
	return (8 - nCount);
}

int CQuantizer::GetRightShiftCount (uint32 dwVal)
{
	for (int i=0; i<sizeof (uint32) * 8; i++) {
		if (dwVal & 1)
			return i;
		dwVal >>= 1;
	}
	return -1;
}

void CQuantizer::AddColor (NODE** ppNode, uint8 r, uint8 g, uint8 b,
	uint32 nColorBits, uint32 nLevel, uint32* pLeafCount, NODE** pReducibleNodes)
{
	static uint8 mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

	//
	// If the node doesn't exist, create it.
	//
	if (*ppNode == NULL)
		*ppNode = CreateNode (nLevel, nColorBits, pLeafCount,
			pReducibleNodes);

	//
	// Update color information if it's a leaf node.
	//
	if ((*ppNode)->bIsLeaf) {
		(*ppNode)->nPixelCount++;
		(*ppNode)->nRedSum += r;
		(*ppNode)->nGreenSum += g;
		(*ppNode)->nBlueSum += b;
	}

	//
	// Recurse a level deeper if the node is not a leaf.
	//
	else {
		int shift = 7 - nLevel;
		int nIndex = (((r & mask[nLevel]) >> shift) << 2) |
			(((g & mask[nLevel]) >> shift) << 1) |
			((b & mask[nLevel]) >> shift);
		AddColor (&((*ppNode)->pChild[nIndex]), r, g, b, nColorBits,
			nLevel + 1, pLeafCount, pReducibleNodes);
	}
}

NODE* CQuantizer::CreateNode (uint32 nLevel, uint32 nColorBits, uint32* pLeafCount,
	NODE** pReducibleNodes)
{
	NODE* pNode;

	if ((pNode = (NODE*) HeapAlloc (GetProcessHeap (), HEAP_ZERO_MEMORY,
		sizeof (NODE))) == NULL)
		return NULL;

	pNode->bIsLeaf = (nLevel == nColorBits) ? TRUE : FALSE;
	if (pNode->bIsLeaf)
		(*pLeafCount)++;
	else {
		pNode->pNext = pReducibleNodes[nLevel];
		pReducibleNodes[nLevel] = pNode;
	}
	return pNode;
}

void CQuantizer::ReduceTree (uint32 nColorBits, uint32* pLeafCount,
	NODE** pReducibleNodes)
{
	//
	// Find the deepest level containing at least one reducible node.
	//
	for (int i=nColorBits - 1; (i>0) && (pReducibleNodes[i] == NULL); i--);

	//
	// Reduce the node most recently added to the list at level i.
	//
	NODE* pNode = pReducibleNodes[i];
	pReducibleNodes[i] = pNode->pNext;

	uint32 nRedSum = 0;
	uint32 nGreenSum = 0;
	uint32 nBlueSum = 0;
	uint32 nChildren = 0;

	for (i=0; i<8; i++) {
		if (pNode->pChild[i] != NULL) {
			nRedSum += pNode->pChild[i]->nRedSum;
			nGreenSum += pNode->pChild[i]->nGreenSum;
			nBlueSum += pNode->pChild[i]->nBlueSum;
			pNode->nPixelCount += pNode->pChild[i]->nPixelCount;
			HeapFree (GetProcessHeap (), 0, pNode->pChild[i]);
			pNode->pChild[i] = NULL;
			nChildren++;
		}
	}

	pNode->bIsLeaf = TRUE;
	pNode->nRedSum = nRedSum;
	pNode->nGreenSum = nGreenSum;
	pNode->nBlueSum = nBlueSum;
	*pLeafCount -= (nChildren - 1);
}

void CQuantizer::DeleteTree (NODE** ppNode)
{
	for (int i=0; i<8; i++) {
		if ((*ppNode)->pChild[i] != NULL)
			DeleteTree (&((*ppNode)->pChild[i]));
	}
	HeapFree (GetProcessHeap (), 0, *ppNode);
	*ppNode = NULL;
}

void CQuantizer::GetPaletteColors (NODE* pTree, RGBQUAD* prgb, uint32* pIndex)
{
	if (pTree->bIsLeaf) {
		prgb[*pIndex].rgbRed =
			(uint8) ((pTree->nRedSum) / (pTree->nPixelCount));
		prgb[*pIndex].rgbGreen =
			(uint8) ((pTree->nGreenSum) / (pTree->nPixelCount));
		prgb[*pIndex].rgbBlue =
			(uint8) ((pTree->nBlueSum) / (pTree->nPixelCount));
		prgb[*pIndex].rgbReserved = 0;
		(*pIndex)++;
	}
	else {
		for (int i=0; i<8; i++) {
			if (pTree->pChild[i] != NULL)
				GetPaletteColors (pTree->pChild[i], prgb, pIndex);
		}
	}
}

uint32 CQuantizer::GetColorCount ()
{
	return m_nLeafCount;
}

void CQuantizer::GetColorTable (RGBQUAD* prgb)
{
	uint32 nIndex = 0;
	GetPaletteColors (m_pTree, prgb, &nIndex);
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetRowBytes
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

long PrvGetRowBytes (LPBITMAPINFO dib)
{
	int32	biWidth = dib->bmiHeader.biWidth;
	int16	biBitCount = dib->bmiHeader.biBitCount;

	return ((biWidth * biBitCount + 31) & ~31) / 8;
}


