/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Skins.h"

#include "ChunkFile.h"			// Chunk
#include "EmFileRef.h"			// EmFileRef
#include "EmMapFile.h"			// EmMapFile
#include "EmStreamFile.h"		// EmStreamFile, kOpenExistingForRead
#include "Miscellaneous.h"		// StartsWith
#include "Platform.h"			// _stricmp
#include "PreferenceMgr.h"		// Preference

#include <algorithm>			// find()

struct ButtonBounds
{
	SkinElementType	fButton;
	RectangleType	fBounds;
};
typedef vector<ButtonBounds>	ButtonBoundsList;
 
struct Skinfo
{
	Skinfo () :
		fSkinFile (),
		fName (),
		fImageName1x (),
		fImageName2x (),
		fDevices (),
		fButtons ()
		{}

	EmFileRef			fSkinFile;
	SkinName			fName;
	string				fImageName1x;
	string				fImageName2x;
	RGBType				fBackgroundColor;
	RGBType				fHighlightColor;
	DeviceList			fDevices;
	ButtonBoundsList	fButtons;
};
typedef vector<Skinfo>	SkinList;


static DeviceType		gCurrentDevice;
static Skinfo			gCurrentSkin;
static ScaleType		gCurrentScale;

static void				PrvBuildSkinList	(SkinList&);
static void				PrvGetSkins			(DeviceType type, SkinList& results);
static Bool				PrvGetNamedSkin		(DeviceType type, const SkinName& name, Skinfo& result);
static void				PrvGetGenericSkin	(Skinfo& skin);
static void				PrvGetDefaultSkin	(DeviceType type, Skinfo& skin);
static void				PrvSetSkin			(DeviceType type, const Skinfo&, ScaleType scale);
static RectangleType	PrvGetTouchscreen	(void);
static SkinElementType	PrvTestPoint		(const PointType&, int outset);
static Bool				PrvPtInRect			(const PointType&, const AbsRectType&);


static const char* kElementNames[] =
{
	"PowerButton",
	"UpButton",
	"DownButton",
	"App1Button",
	"App2Button",
	"App3Button",
	"App4Button",
	"CradleButton",
	"Antenna",
	"ContrastButton",

		// Symbol-specific
	"TriggerLeft",
	"TriggerCenter",
	"TriggerRight",
	"UpButtonLeft",
	"UpButtonRight",
	"DownButtonLeft",
	"DownButtonRight",

	"Touchscreen",
	"LCD"
};


static const char	kGenericSkinName[] = "Generic";

static ButtonBounds	kGenericButtons [] =
{
	{ kElement_PowerButton,		{ {   1, 274 }, {  16,  24 } } },
	{ kElement_UpButton,		{ {  96, 272 }, {  32,  16 } } },
	{ kElement_DownButton,		{ {  96, 293 }, {  32,  16 } } },
	{ kElement_App1Button,		{ {  23, 270 }, {  32,  32 } } },
	{ kElement_App2Button,		{ {  60, 270 }, {  32,  32 } } },
	{ kElement_App3Button,		{ { 131, 270 }, {  32,  32 } } },
	{ kElement_App4Button,		{ { 168, 270 }, {  32,  32 } } },
	{ kElement_CradleButton,	{ {   0,   0 }, {   0,   0 } } },
	{ kElement_Antenna,			{ {   0,   0 }, {   0,   0 } } },
	{ kElement_ContrastButton,	{ {   0,   0 }, {   0,   0 } } },
	{ kElement_Touchscreen,		{ {  32,  32 }, { 160, 220 } } },
	{ kElement_LCD,				{ {  32,  32 }, { 160, 160 } } }
};

static RGBType	kGenericBackgroundColor	(0x7B, 0x8C, 0x5A);
static RGBType	kGenericHighlightColor	(0x64, 0xF0, 0xDC);


/***********************************************************************
 *
 * FUNCTION:    SkinGetSkinName
 *
 * DESCRIPTION: Get the name of the user-chosen skin for the given
 *				device type.
 *
 * PARAMETERS:  type - the device type for which we need the name
 *					of the skin to use.
 *
 * RETURNED:    The skin name.  At the very least, this will be the
 *				name of some default skin if the user hasn't made a
 *				choice or if the chosen skin is invalid.
 *
 ***********************************************************************/

SkinName SkinGetSkinName (DeviceType type)
{
	// Get the chosen skin for this device.

	Preference<SkinNameList>	pref(kPrefKeySkins);
	SkinNameList				skinNames = *pref;
	SkinName					name;
	
	if ((int) type < (int) skinNames.size ())
		name = skinNames[(int) type];

	// If the name is empty or invalid, chose a default skin.

	if (!::SkinValidSkin (type, name))
	{
		name = ::SkinGetDefaultSkinName (type);
	}

	return name;
}


/***********************************************************************
 *
 * FUNCTION:    SkinGetDefaultSkinName
 *
 * DESCRIPTION: Get the name of the user-chosen skin for the given
 *				device type.
 *
 * PARAMETERS:  type - the device type for which we need the name
 *					of the skin to use.
 *
 * RETURNED:    The skin name.  At the very least, this will be the
 *				name of some default skin if the user hasn't made a
 *				choice or if the chosen skin is invalid.
 *
 ***********************************************************************/

SkinName SkinGetDefaultSkinName (DeviceType type)
{
	Skinfo	skin;
	::PrvGetDefaultSkin (type, skin);
	return skin.fName;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetSkinNames
 *
 * DESCRIPTION:	Get the list of names of available skins for the given
 *				device.
 *
 * PARAMETERS:	type - the device for which the list of skins should
 *					be returned.  If kDeviceUnspecified, return all
 *					skins for all devices.
 *
 *				results - receives the list of skin names.  Any skin
 *					names are *added* to this list; the list is not
 *					cleared out first.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void SkinGetSkinNames (DeviceType type, SkinNameList& results)
{
	// Always push on the name of the default skin first.

	results.push_back (kGenericSkinName);

	// Get the names of any custom skins for the device and
	// add those, too.

	SkinList	skins;
	::PrvGetSkins (type, skins);

	SkinList::iterator	iter = skins.begin ();
	while (iter != skins.end ())
	{
		results.push_back (iter->fName);

		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	SkinSetSkin
 *
 * DESCRIPTION:	Establish the skin to use, based on the current settings.
 *				All other funtions in this module will then work within
 *				the context of the specified skin.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void SkinSetSkin (void)
{
	Preference<Configuration>	cfgPref (kPrefKeyLastConfiguration);
	Preference<SkinNameList>	skinsPref (kPrefKeySkins);
	Preference<ScaleType>		scalePref (kPrefKeyScale);

	SkinNameList	skins	= *skinsPref;
	DeviceType		type	= cfgPref->fDeviceType;

	::SkinSetSkin (type, *scalePref, skins[type]);
}


void SkinSetSkin (DeviceType type, ScaleType scale, const SkinName& name)
{
	Skinfo		skin;

	if (!::PrvGetNamedSkin (type, name, skin))
	{
		::PrvGetDefaultSkin (type, skin);
	}

	::PrvSetSkin (type, skin, scale);
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetSkinfoFile
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmFileRef SkinGetSkinfoFile (void)
{
	return gCurrentSkin.fSkinFile;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetSkinFile
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmFileRef SkinGetSkinFile (void)
{
	return ::SkinGetSkinFile (gCurrentScale);
}


EmFileRef SkinGetSkinFile (ScaleType scale)
{
	EmFileRef		skinfoFile = ::SkinGetSkinfoFile ();

	if (!skinfoFile.IsSpecified ())
	{
		return EmFileRef ();
	}

	assert (!Platform::SkinfoResourcePresent ());
	assert (!Platform::Skin1xResourcePresent ());
	assert (!Platform::Skin2xResourcePresent ());

	string			name;

	if (scale == 1)
		name = gCurrentSkin.fImageName1x.c_str();
	else
		name = gCurrentSkin.fImageName2x.c_str();

	EmDirRef		skinDir = skinfoFile.GetParent ();
	EmFileRef		skinImage (skinDir, name);

	// If the skin file doesn't exist, try looking for it
	// in a Mac or Windows sub-directory.  This helps development
	// of the emulator, as images for different platforms are
	// stored in directories with those names.  This sub-directory
	// looking shouldn't be needed for the image archive made
	// available to developers.

	if (!skinImage.Exists ())
	{
#if PLATFORM_MAC
		EmDirRef	skinSubDir (skinDir, "Mac");
#else
		EmDirRef	skinSubDir (skinDir, "Windows");
#endif

		skinImage = EmFileRef (skinSubDir, name);
		if (!skinImage.Exists ())
		{
			EmFileRef temp;
			return (temp);
		}
	}

	return skinImage;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetSkinStream
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmStream* SkinGetSkinStream (void)
{
	return ::SkinGetSkinStream (gCurrentScale);
}


EmStream* SkinGetSkinStream (ScaleType scale)
{
	EmStream*	result = NULL;

	if (gCurrentSkin.fName == kGenericSkinName)
		return result;

	if (::IsBound ())
	{
		// If we're bound, open up a stream on the resource data.

		Bool	haveRes;
		Chunk*	chunk = new Chunk;

		if (scale == 1)
			haveRes = Platform::GetSkin1xResource (*chunk);
		else
			haveRes = Platform::GetSkin2xResource (*chunk);

		if (haveRes)
			result = new EmStreamChunk (chunk);
		else
			delete chunk;
	}
	else
	{
		EmFileRef	file = ::SkinGetSkinFile (scale);

		if (file.Exists ())
		{
			// If we're not bound, try opening a stream on the file.

			result = new EmStreamFile (file, kOpenExistingForRead);
		}
	}

	if (!result)
	{
		// If we can't get the image, revert to the default image.
		// !!! Should probably also remove the skin from the skin list.

		Skinfo	skin;
		::PrvGetDefaultSkin (gCurrentDevice, skin);
		::PrvSetSkin (gCurrentDevice, skin, scale);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SkinValidSkin
 *
 * DESCRIPTION:	Returns whether the given device has a skin with the
 *				given name.
 *
 * PARAMETERS:	type - the device type.
 *
 *				skinName - the skin name.
 *
 * RETURNED:	True if the given device has a skin with the given name.
 *				False otherwise.
 *
 ***********************************************************************/

Bool SkinValidSkin (DeviceType type, const SkinName& skinName)
{
	SkinNameList	skins;
	::SkinGetSkinNames (type, skins);

	SkinNameList::iterator	iter = skins.begin ();
	while (iter != skins.end ())
	{
		if (*iter == skinName)
		{
			return true;
		}

		++iter;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetBackgroundColor
 *
 * DESCRIPTION:	Return the default background color for the current
 *				skin.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	The default background color.
 *
 ***********************************************************************/

RGBType		SkinGetBackgroundColor	(void)
{
	return gCurrentSkin.fBackgroundColor;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetHighlightColor
 *
 * DESCRIPTION:	Return the default highlight color for the current
 *				skin.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	The default highlight color.
 *
 ***********************************************************************/

RGBType		SkinGetHighlightColor	(void)
{
	return gCurrentSkin.fHighlightColor;
}


/***********************************************************************
 *
 * FUNCTION:	SkinTestPoint
 *
 * DESCRIPTION:	Tests the given point to see what skin element it's over.
 *
 * PARAMETERS:	pt - location in the window to test.
 *
 * RETURNED:	If the point is within an element, returns the id for
 *				that element.  If an element was just missed, returns
 *				kElement_None.  Otherwise, returns kElement_Frame.
 *
 ***********************************************************************/

SkinElementType	SkinTestPoint		(const PointType& pt)
{	
	// See if we hit an element.  PrvTestPoint will return either the
	// element hit or kElement_Frame if none were hit.  If an element
	// was hit, return it.

	SkinElementType	result = ::PrvTestPoint (pt, 0);

	if (result != kElement_Frame)
		return result;

	// Test again, this time allowing for some slop around the
	// elements.  If an element was hit this time, then we hit the
	// small "dead area" we allow around the elements.  In that case,
	// we want to return kElement_None.  Otherwise, if no element was
	// hit, signal that the frame was hit.

	result = ::PrvTestPoint (pt, 5);
	if (result != kElement_Frame)
		return kElement_None;

	return kElement_Frame;
}


/***********************************************************************
 *
 * FUNCTION:	SkinWindowToTouchscreen
 *
 * DESCRIPTION:	Convert a point from window coordinates to LCD
 *				coordinates.
 *
 * PARAMETERS:	pt - point in window coordinates to convert.
 *
 * RETURNED:	The point in LCD coordinates (where the topleft corner
 *				of the LCD is 0,0 and the scale is 1x).
 *
 ***********************************************************************/

PointType		SkinWindowToTouchscreen		(const PointType& pt)
{
	PointType		result	= pt;
	RectangleType	r		= ::PrvGetTouchscreen ();

	result.x -= r.topLeft.x;
	result.y -= r.topLeft.y;

	if (result.x < 0)
		result.x = 0;
	
	if (result.y < 0)
		result.y = 0;
	
	if (result.x >= r.extent.x)
		result.x = r.extent.x - 1;
	
	if (result.y >= r.extent.y)
		result.y = r.extent.y - 1;

	result = ::SkinScaleDown (result);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SkinLCDToWindow
 *
 * DESCRIPTION:	Convert a point from LCD coordinates to window
 *				coordinates.
 *
 * PARAMETERS:	pt - point in LCD coordinates to convert.
 *
 * RETURNED:	The point in window coordinates (where the topleft
 *				corner of the window is 0,0 and the scale is the
 *				scale chosen by the user).
 *
 ***********************************************************************/

PointType		SkinTouchscreenToWindow		(const PointType& lcdPt)
{
	PointType		result	= lcdPt;
	RectangleType	r		= ::PrvGetTouchscreen ();

	result.x += r.topLeft.x;
	result.y += r.topLeft.y;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetElementInfo
 *
 * DESCRIPTION:	Get information on the given element.
 *
 * PARAMETERS:	index - index of the element we're querying.
 *
 *				type - type of the element we've queried
 *
 *				bounds - bounds of the element we've queried.  The
 *					is scaled up if necessary.
 *
 * RETURNED:	TRUE if there was such an element.  FALSE if index is
 *				out of range.
 *
 ***********************************************************************/

Bool	SkinGetElementInfo	(int index, SkinElementType& type, RectangleType& bounds)
{
	if (index < (int) gCurrentSkin.fButtons.size ())
	{
		type = gCurrentSkin.fButtons[index].fButton;
		bounds = gCurrentSkin.fButtons[index].fBounds;
		bounds = ::SkinScaleUp (bounds);

		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	SkinGetElementInfo
 *
 * DESCRIPTION:	Get information on the given element.
 *
 * PARAMETERS:	index - index of the element we're querying.
 *
 *				type - type of the element we've queried
 *
 *				bounds - bounds of the element we've queried.  The
 *					is scaled up if necessary.
 *
 * RETURNED:	TRUE if there was such an element.  FALSE if index is
 *				out of range.
 *
 ***********************************************************************/

Bool	SkinGetElementInfo	(int index, SkinElementType& type, AbsRectType& bounds)
{
	RectangleType	r;

	if (::SkinGetElementInfo (index, type, r))
	{
		bounds.left		= r.topLeft.x;
		bounds.top		= r.topLeft.y;
		bounds.right	= r.topLeft.x + r.extent.x;
		bounds.bottom	= r.topLeft.y + r.extent.y;

		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	SkinScaleDown
 *
 * DESCRIPTION:	Convert a point from 1x or 2x to just 1x.
 *
 * PARAMETERS:	pt - point to change.
 *
 * RETURNED:	Normalized point.
 *
 ***********************************************************************/

PointType	SkinScaleDown	(const PointType& pt)
{
	PointType	result =
	{
		pt.x / gCurrentScale,
		pt.y / gCurrentScale
	};

#if 0
	// Note: I used to do it this way, but this runs afoul of a code
	// generation bug when using egcs 1.1.2 for x86 -- the y coordinate
	// gets trashed.

	result.x			/= gCurrentScale;
	result.y			/= gCurrentScale;
#endif

	return result;
}


RectangleType	SkinScaleDown	(const RectangleType& r)
{
	RectangleType	result = r;

	result.topLeft.x	/= gCurrentScale;
	result.topLeft.y	/= gCurrentScale;
	result.extent.x		/= gCurrentScale;
	result.extent.y		/= gCurrentScale;

	return result;
}


AbsRectType	SkinScaleDown	(const AbsRectType& r)
{
	AbsRectType	result = r;

	result.left			/= gCurrentScale;
	result.top			/= gCurrentScale;
	result.right		/= gCurrentScale;
	result.bottom		/= gCurrentScale;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SkinScaleUp
 *
 * DESCRIPTION:	Convert a point to 1x or 2x, depending on the scaling
 *				factor.
 *
 * PARAMETERS:	pt - point to change.
 *
 * RETURNED:	Denormalized point.
 *
 ***********************************************************************/

PointType	SkinScaleUp	(const PointType& pt)
{
	PointType	result =
	{
		pt.x * gCurrentScale,
		pt.y * gCurrentScale
	};

	return result;
}


RectangleType	SkinScaleUp	(const RectangleType& r)
{
	RectangleType	result = r;

	result.topLeft.x	*= gCurrentScale;
	result.topLeft.y	*= gCurrentScale;
	result.extent.x		*= gCurrentScale;
	result.extent.y		*= gCurrentScale;

	return result;
}


AbsRectType	SkinScaleUp	(const AbsRectType& r)
{
	AbsRectType	result = r;

	result.left			*= gCurrentScale;
	result.top			*= gCurrentScale;
	result.right		*= gCurrentScale;
	result.bottom		*= gCurrentScale;

	return result;
}


#pragma mark -


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static int PrvGetNumber (const string& s)
{
	// Get the (stringized) component to convert.

	string	numString = ::Strip (s, " \t", true, true);

	// Try converting it from hex to numeric format.

	int		result;
	int		scanned = sscanf (numString.c_str (), "0x%X", &result);

	// If that failed, try converting it from decimal to numeric format.
	// We have to be careful about scanning "0".  That text will match
	// the "0" in "0x%X", causing it to return EOF, but not filling
	// in "result".

	if (scanned == 0 || scanned == EOF)
		scanned = sscanf (numString.c_str (), "%d", &result);

	// If that failed, throw an exception.
	
	if (scanned == 0 || scanned == EOF)
		throw 1;

	return result;
}

/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static uint8 PrvGetOneRGB (const StringList& rgbs, int component)
{
	// Convert the string to a number.

	int	result = ::PrvGetNumber (rgbs[component]);

	// If that failed, or if the resulting number is out of range,
	// throw an exception indicating an invalid input.

	if (result < 0 || result > 255)
		throw 1;

	return (uint8) result;
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static RGBType PrvGetRGB (const StringStringMap& entries, const char* key)
{
	RGBType	result;

	// Get the string containing the devices.

	string		rgbString = entries.find (key)->second;

	// Break it up into parts.

	StringList	rgbs;
	::SeperateList (rgbs, rgbString, ',');

	// Make sure we have the right number
	
	if (rgbs.size () != 3)
		throw 1;

	result.fRed		= ::PrvGetOneRGB (rgbs, 0);
	result.fGreen	= ::PrvGetOneRGB (rgbs, 1);
	result.fBlue	= ::PrvGetOneRGB (rgbs, 2);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static DeviceList PrvGetDeviceList (const StringStringMap& entries)
{
	DeviceList	result;

	// Get the string containing the devices.

	string		devicesString = entries.find ("Devices")->second;

	// Break it up into parts.

	StringList	deviceNames;
	::SeperateList (deviceNames, devicesString, ',');

	// Get the list of device strings that we recognize.

	EmDeviceStartupItemList	knownDevices = EmDevice::GetStartupItemList ();

	// Iterate over the given device names, and see if they're
	// in the list of names we recognize.

	StringList::iterator	iter = deviceNames.begin ();
	while (iter != deviceNames.end ())
	{
		// Get a device name, and remove any leading or trailing whitespace.

		string	deviceName = ::Strip (*iter, " \t", true, true);

		// Iterate over the list of supported devices, and see if the provided
		// device name is on it.

		EmDeviceStartupItemList::iterator	iter2 = knownDevices.begin ();
		while (iter2 != knownDevices.end ())
		{
			// A valid device name was provided, so add it to the list of OK devices.

			if (deviceName == iter2->second)
			{
				result.push_back (iter2->first);
				break;
			}

			++iter2;
		}

		++iter;
	}

	// If no valid devices were provided, throw an exception to signal
	// the caller that a completely invalid list was provided.
	
	if (result.size () == 0)
		throw 1;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static SkinElementType PrvGetElementType (const string& s)
{
	// Get the element name and try to ID it.

	string	eltName = ::Strip (s, " \t", true, true);

	for (SkinElementType ii = kElement_First; ii < kElement_NumElements; ++ii)
	{
		if (kElementNames[ii] != NULL &&
			_stricmp (eltName.c_str (), kElementNames[ii]) == 0)
		{
			return ii;
		}
	}
	
	// If we couldn't determine the name, error out.

	throw 1;

	return kElement_None;	// (For the compiler)
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static ButtonBounds PrvGetOneElement (const string& s)
{
	ButtonBounds	result;

	// Break up the string into its parts, which should be an
	// element name followed by its bounds.

	StringList		parts;
	::SeperateList (parts, s, ',');

	// Make sure we have the right number of pieces.
	
	if (parts.size () != 5)
		throw 1;

	result.fButton = ::PrvGetElementType (parts[0]);

	// Get the bounds
	
	result.fBounds.topLeft.x = ::PrvGetNumber (parts[1]);
	result.fBounds.topLeft.y = ::PrvGetNumber (parts[2]);
	result.fBounds.extent.x = ::PrvGetNumber (parts[3]);
	result.fBounds.extent.y = ::PrvGetNumber (parts[4]);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static ButtonBoundsList PrvGetElementList (const StringStringMap& entries)
{
	ButtonBoundsList	result;
	Bool				haveLCD = false;
	Bool				haveTouchScreen = false;

	// Iterate over all of the entries in "entries", looking for ones
	// that start with "Element".

	StringStringMap::const_iterator	iter = entries.begin ();
	while (iter != entries.end ())
	{
		if (::StartsWith (iter->first.c_str (), "Element"))
		{
			ButtonBounds	elt = ::PrvGetOneElement (iter->second);
			result.push_back (elt);

			if (elt.fButton == kElement_LCD)
				haveLCD = true;

			if (elt.fButton == kElement_Touchscreen)
				haveTouchScreen = true;
		}

		++iter;
	}

	if (!haveLCD || !haveTouchScreen)
		throw 1;

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static void PrvAddSkin (SkinList& skins, StringStringMap& entries,
						const EmFileRef* file = NULL)
{
	// Check for required fields.

	if (entries.find ("Name")				== entries.end ()	||
		entries.find ("File1x")				== entries.end ()	||
		entries.find ("File2x")				== entries.end ()	||
		entries.find ("BackgroundColor")	== entries.end ()	||
		entries.find ("HighlightColor")		== entries.end ()	||
		entries.find ("Devices")			== entries.end ())
	{
		return;
	}

	// Now extract the values.

	try
	{
		Skinfo	skin;

		skin.fSkinFile			= file ? *file : EmFileRef();
		skin.fName				= entries["Name"];
		skin.fImageName1x		= entries["File1x"];
		skin.fImageName2x		= entries["File2x"];
		skin.fBackgroundColor	= ::PrvGetRGB (entries, "BackgroundColor");
		skin.fHighlightColor	= ::PrvGetRGB (entries, "HighlightColor");
		skin.fDevices			= ::PrvGetDeviceList (entries);
		skin.fButtons			= ::PrvGetElementList (entries);

		skins.push_back (skin);
	}
	catch (...)
	{
	}
}

static void PrvAddSkin (SkinList& skins, EmStream& skinStream)
{
	StringStringMap	entries;
	EmMapFile::Read (skinStream, entries);

	PrvAddSkin (skins, entries);
}

static void PrvAddSkin (SkinList& skins, const EmFileRef& skinFile)
{
	StringStringMap	entries;
	EmMapFile::Read (skinFile, entries);

	PrvAddSkin (skins, entries, &skinFile);
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static void PrvScanForSkinFiles (SkinList& skins, const EmDirRef& skinDir)
{
	EmFileRefList	files;
	skinDir.GetChildren (&files, NULL);

	EmFileRefList::iterator	iter = files.begin ();
	while (iter != files.end ())
	{
		if (iter->IsType (kFileTypeSkin))
		{
			::PrvAddSkin (skins, *iter);
		}

		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	.
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

static void PrvScanForSkinDirs (SkinList& skins, const EmDirRef& scanDir)
{
	EmDirRefList	dirs;
	scanDir.GetChildren (NULL, &dirs);

	EmDirRefList::iterator	iter = dirs.begin ();
	while (iter != dirs.end ())
	{
		string	dirName = iter->GetName ();
		if (::StartsWith (dirName.c_str (), "skins"))
		{
			::PrvScanForSkinFiles (skins, *iter);
		}

		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvBuildSkinList
 *
 * DESCRIPTION:	Create the full list of skins known to the emulator.
 *				This list of skins includes both the built-in ones and
 *				any found on disk.
 *
 * PARAMETERS:	skins - receives the list of skins.  The list of skins
 *					is *added* to this collection; it is not cleared
 *					out first.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvBuildSkinList (SkinList& skins)
{
	// If we're a "bound" Poser, the skin list consists of the
	// skin we're bound to.

	if (::IsBound ())
	{
		Chunk	chunk;
		if (Platform::GetSkinfoResource (chunk))
		{
			EmStreamChunk	stream (chunk);
			::PrvAddSkin (skins, stream);
		}

		return;
	}

	// Scan the current directory for sub-directories starting with "Skins"
	// Also check our parent and our grandparent.  That way, we can have
	// Skins directories above us that are shared by different Poser builds.

	EmDirRef	scanDir = EmDirRef::GetEmulatorDirectory ();

	for (int ii = 0; ii < 3; ++ii)
	{
		if (!scanDir.Exists ())
			break;

		::PrvScanForSkinDirs (skins, scanDir);
		scanDir = scanDir.GetParent ();
	}
}


#pragma mark -


/***********************************************************************
 *
 * FUNCTION:	PrvSetSkin
 *
 * DESCRIPTION:	Common low-level routine to setting the current skin.
 *
 * PARAMETERS:	skin - new skin.
 *
 *				scale - new scale.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void PrvSetSkin (DeviceType type, const Skinfo& skin, ScaleType scale)
{
	gCurrentDevice	= type;
	gCurrentSkin	= skin;
	gCurrentScale	= scale;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetSkins
 *
 * DESCRIPTION:	Get the list of available skins for the given device.
 *				Does not include the generic skin.
 *
 * PARAMETERS:	type - the device for which the list of skins should
 *					be returned.  If kDeviceUnspecified, return all
 *					skins for all devices.
 *
 *				results - receives the list of skins.  Any skins are
 *					*added* to this list; the list is not cleared out
 *					first.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void PrvGetSkins (DeviceType type, SkinList& results)
{
	static SkinList	gFullSkinList;
	static Bool		gInitialized;

	if (!gInitialized)
	{
		gInitialized = true;
		::PrvBuildSkinList (gFullSkinList);
	}

	SkinList::iterator	iter = gFullSkinList.begin ();
	while (iter != gFullSkinList.end ())
	{
		if (type == kDeviceUnspecified ||
			find (iter->fDevices.begin (), iter->fDevices.end (), type) != iter->fDevices.end ())
		{
			results.push_back (*iter);
		}

		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:    PrvGetNamedSkin
 *
 * DESCRIPTION: Find the Skinfo for the given device and that has the
 *				given name.
 *
 * PARAMETERS:  type - the device whose skin we're looking for.
 *
 *				name - the name of the skin to find.
 *
 *				result - reference to the Skinfo in which to place the
 *					found skin information, if any.
 *
 * RETURNED:    True if the skin could be found, false othewise.
 *
 ***********************************************************************/

Bool PrvGetNamedSkin (DeviceType type, const SkinName& name, Skinfo& result)
{
	if (name == kGenericSkinName)
	{
		::PrvGetGenericSkin (result);
		return true;
	}

	SkinList	skins;
	::PrvGetSkins (type, skins);

	SkinList::iterator	iter = skins.begin();
	while (iter != skins.end ())
	{
		if (iter->fName == name)
		{
			result = *iter;
			return true;
		}

		++iter;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    PrvGetGenericSkin
 *
 * DESCRIPTION: Return skin information for the "generic" skin, the one
 *				that is built-in and can be used for any device if a
 *				custom one cannot be found.
 *
 * PARAMETERS:  skin - reference to the Skinfo in which to place the
 *					default skin information.
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvGetGenericSkin (Skinfo& skin)
{
	skin.fDevices.clear ();
	skin.fButtons.clear ();

	skin.fName				= kGenericSkinName;
	skin.fImageName1x		= "";
	skin.fImageName2x		= "";
	skin.fBackgroundColor	= kGenericBackgroundColor;
	skin.fHighlightColor	= kGenericHighlightColor;

	for (size_t ii = 0; ii < countof (kGenericButtons); ++ii)
		skin.fButtons.push_back (kGenericButtons[ii]);
}


/***********************************************************************
 *
 * FUNCTION:    PrvGetDefaultSkin
 *
 * DESCRIPTION: Return default skin information for the given device.
 *				This function returns information in the case that
 *				a skin with a desired name could not be found.
 *
 * PARAMETERS:  type - the device whose default skin information we want.
 *
 *				skin - reference to the Skinfo in which to place the
 *					default skin information.
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void PrvGetDefaultSkin (DeviceType type, Skinfo& skin)
{
	SkinList	skins;
	::PrvGetSkins (type, skins);

	if (skins.size () == 0)
	{
		::PrvGetGenericSkin (skin);
	}
	else
	{
		skin = skins[0];
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetTouchscreen
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

RectangleType	PrvGetTouchscreen	(void)
{
	int	index = 0;
	SkinElementType	type;
	RectangleType	bounds;

	while (::SkinGetElementInfo (index, type, bounds))
	{
		if (type == kElement_Touchscreen)
		{
			return bounds;
		}

		++index;
	}

	assert (false);

	// Shut up the compiler
	RectangleType	dummy = { {0, 0}, {0, 0} };
	return dummy;
}


/***********************************************************************
 *
 * FUNCTION:	PrvTestPoint
 *
 * DESCRIPTION:	Test the given point against all of the skin elements.
 *				An optional outset value can be provided which is
 *				used to modify the element bounds before they are
 *				tested.
 *
 * PARAMETERS:	pt - window location to test.
 *
 *				outset - outset value to apply to the bounds of all
 *					the skin elements.
 *
 * RETURNED:	If one contains the given point, return that skin
 *				element.  Otherwise, return kElement_Frame.
 *
 ***********************************************************************/

SkinElementType	PrvTestPoint (const PointType& pt, int outset)
{
	ButtonBoundsList::iterator	iter = gCurrentSkin.fButtons.begin ();
	while (iter != gCurrentSkin.fButtons.end ())
	{
		RectangleType	bounds = iter->fBounds;
		bounds = ::SkinScaleUp (bounds);

		AbsRectType		r;

		r.left = bounds.topLeft.x - outset;
		r.top = bounds.topLeft.y - outset;
		r.right = bounds.topLeft.x + bounds.extent.x + outset;
		r.bottom = bounds.topLeft.y + bounds.extent.y + outset;

		if (::PrvPtInRect (pt, r))
			return iter->fButton;

		++iter;
	}

	return kElement_Frame;
}


/***********************************************************************
 *
 * FUNCTION:	PrvPtInRect
 *
 * DESCRIPTION:	Test if the given point is within the given rectangle.
 *
 * PARAMETERS:	pt - point to test.
 *
 *				r - rectangle to test against.
 *
 * RETURNED:	True if the point is within the rectangle, using a
 *				half-open interval.  False otherwise.
 *
 ***********************************************************************/

Bool	PrvPtInRect		(const PointType& pt, const AbsRectType& r)
{
	return
		pt.x >= r.left &&
		pt.y >= r.top &&
		pt.x < r.right &&
		pt.y < r.bottom;
}
