/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "UART.h"

#include "Bank_Regs.h"			// RegsBank
#include "EmTransportSerial.h"	// EmTransportSerial
#include "Logging.h"
#include "Platform.h"
#include "PreferenceMgr.h"		// Preference, kPrefKeyCommPort

/*
	This module contains the routines for handling serial I/O.  It
	is responsible for responding to changes in state enacted by
	software (from either the OS or user), and for dealing with
	the actual transmission and reception of serial data.

	There are four ways in which serial activity could occur: something
	could write to a UART register, something could read from a UART
	register, a byte could be received from the host serial port, or
	a byte could be sent out the host serial port.  Here is what
	happens on each of those events.

	Something reads a UART register:
		- If the register is the RX_DATA register, clear the DATA_READY bit
		- Make sure the state is up-to-date (including interrupts)
		- Return the register contents

	Something writes to a UART register:
		- Update the writable parts of the register
		- React to any changes
		- Make sure the state is up-to-date (including interrupts)

	Data appears at the host serial port:
		- Post the byte to the RX FIFO (if there is room)
		- Make sure the state is up-to-date (including interrupts)

	Data needs to be sent to the host serial port:
		- Send the first byte in the TX FIFO
		- Make sure the state is up-to-date (including interrupts)
*/


// ======================================================================
//	Globals and constants
// ======================================================================

static UART::State	gCurrentState(UART::kUART_Dragonball);
static TByteQueue	gRxFIFO;
static TByteQueue	gTxFIFO;
static EmTransportSerial*	gSerial;
static Bool			gNeedPortChange;


// ======================================================================
//	Private functions
// ======================================================================

static Bool			PrvPinBaud		(EmTransportSerial::Baud& newBaud);
static Bool			PrvPinBaud		(EmTransportSerial::Baud& newBaud, EmTransportSerial::Baud testBaud);
static void			PrvEmptyQueue	(TByteQueue&);
static void			PrvPrefsChanged (PrefKeyType key);

#define PRINTF	if (!LogSerial ()) ; else LogAppendMsg


/***********************************************************************
 *
 * FUNCTION:	UART::Initialize
 *
 * DESCRIPTION:	Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  May also be called from the Load function
 *				to share common functionality.
 *
 * PARAMETERS:	type - the type of UART to emulate.  The Dragonball
 *					and DragonballEZ UARTs are similar enough that we
 *					can handle them both here with just a few tests in
 *					the places where they differ.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::Initialize (UART_Type type)
{
	gCurrentState.UART_TYPE = type;

	assert (!gSerial);

	Preference<EmTransportSerial::PortName>		pref (kPrefKeyCommPort);
	EmTransportSerial::ConfigSerial				config;
	config.fPort = *pref;

	gSerial = new EmTransportSerial (config);

	gNeedPortChange = false;
	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeyCommPort);
}


/***********************************************************************
 *
 * FUNCTION:	UART::Reset
 *
 * DESCRIPTION:	Standard reset function.  Sets the sub-system to a
 *				default state.  This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::Reset (void)
{
	PrvEmptyQueue (gRxFIFO);
	PrvEmptyQueue (gTxFIFO);

	// UART control register bits

	gCurrentState.UART_ENABLE		= 0;
	gCurrentState.RX_ENABLE			= 0;
	gCurrentState.TX_ENABLE			= 0;
	gCurrentState.RX_CLK_CONT		= 0;
	gCurrentState.PARITY_EN			= 0;
	gCurrentState.ODD_EVEN			= 0;
	gCurrentState.STOP_BITS			= 0;
	gCurrentState.CHAR8_7			= 0;
	gCurrentState.GPIO_DELTA_ENABLE	= 0;	// 68328 only
	gCurrentState.OLD_ENABLE		= 0;	// 68EZ328 only
	gCurrentState.CTS_DELTA_ENABLE	= 0;
	gCurrentState.RX_FULL_ENABLE	= 0;
	gCurrentState.RX_HALF_ENABLE	= 0;
	gCurrentState.RX_RDY_ENABLE		= 0;
	gCurrentState.TX_EMPTY_ENABLE	= 0;
	gCurrentState.TX_HALF_ENABLE	= 0;
	gCurrentState.TX_AVAIL_ENABLE	= 0;

	// Baud control register bits

	gCurrentState.GPIO_DELTA		= 0;	// 68328 only
	gCurrentState.GPIO				= 0;	// 68328 only
	gCurrentState.GPIO_DIR			= 0;	// 68328 only
	gCurrentState.GPIO_SRC			= 0;	// 68328 only
	gCurrentState.UCLK_DIR			= 0;	// 68EZ328 only
	gCurrentState.BAUD_SRC			= 0;
	gCurrentState.DIVIDE			= 0;
	gCurrentState.PRESCALER			= 0x3F;

	// Receive register bits

	gCurrentState.RX_FIFO_FULL		= 0;
	gCurrentState.RX_FIFO_HALF		= 0;
	gCurrentState.DATA_READY		= 0;
	gCurrentState.OLD_DATA			= 0;	// 68EZ328 only
	gCurrentState.OVRUN				= 0;
	gCurrentState.FRAME_ERROR		= 0;
	gCurrentState.BREAK				= 0;
	gCurrentState.PARITY_ERROR		= 0;
	gCurrentState.RX_DATA			= 0;

	// Transmitter register bits

	gCurrentState.TX_FIFO_EMPTY		= 0;
	gCurrentState.TX_FIFO_HALF		= 0;
	gCurrentState.TX_AVAIL			= 0;
	gCurrentState.SEND_BREAK		= 0;
	gCurrentState.IGNORE_CTS		= 0;
	gCurrentState.BUSY				= 0;	// 68EZ328 only
	gCurrentState.CTS_STATUS		= 0;
	gCurrentState.CTS_DELTA			= 0;
	gCurrentState.TX_DATA			= 0;

	// Misc register bits

	gCurrentState.BAUD_TEST			= 0;	// 68EZ328 only
	gCurrentState.CLK_SRC			= 0;
	gCurrentState.FORCE_PERR		= 0;
	gCurrentState.LOOP				= 0;
	gCurrentState.BAUD_RESET		= 0;	// 68EZ328 only
	gCurrentState.IR_TEST			= 0;	// 68EZ328 only
	gCurrentState.RTS_CONT			= 0;
	gCurrentState.RTS				= 0;
	gCurrentState.IRDA_ENABLE		= 0;
	gCurrentState.IRDA_LOOP			= 0;
	gCurrentState.RX_POL			= 0;	// 68EZ328 only
	gCurrentState.TX_POL			= 0;	// 68EZ328 only
}


/***********************************************************************
 *
 * FUNCTION:	UART::Save
 *
 * DESCRIPTION:	Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::Save (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	UART::Load
 *
 * DESCRIPTION:	Standard load function.  Loads any sub-system state
 *				from the given session file.
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::Load (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	UART::Dispose
 *
 * DESCRIPTION:	Standard dispose function.  Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::Dispose (void)
{
	gPrefs->RemoveNotification (PrvPrefsChanged);

	if (gSerial)
	{
		gSerial->Close ();
		delete gSerial;
		gSerial = NULL;
	}
}


/***********************************************************************
 *
 * FUNCTION:	UART::LineDriverChanged
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::LineDriverChanged (void)
{
	if (RegsBank::GetSerialPortOn ())
	{
		gSerial->Open ();
	}
	else
	{
		gSerial->Close ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	UART::StateChanged
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::StateChanged (State& newState, Bool sendTxData)
{
	assert (gCurrentState.UART_TYPE == newState.UART_TYPE);

	// The following registers are not referenced by the ROM and
	// are currently not supported.
	//
	//		RX_CLK_CONT
	//		GPIO_DELTA_ENABLE
	//		OLD_ENABLE
	//		CTS_DELTA_ENABLE
	//		RX_FULL_ENABLE
	//		RX_HALF_ENABLE
	//		TX_EMPTY_ENABLE
	//		TX_HALF_ENABLE
	//		TX_AVAIL_ENABLE
	//		GPIO_DELTA
	//		GPIO
	//		GPIO_DIR
	//		GPIO_SRC
	//		UCLK_DIR
	//		BAUD_SRC
	//		BAUD_TEST
	//		CLK_SRC
	//		FORCE_PERR
	//		BAUD_RESET
	//		IR_TEST
	//		IRDA_LOOP
	//		RX_POL
	//		TX_POL
	//		RX_CLK_CONT

	assert (gSerial);

	EmTransportSerial::ConfigSerial	config;
	gSerial->GetConfig (config);


	// ========== RX_ENABLE ==========
	//
	// This bit enables the receiver block. While this bit is low, the receiver is disabled and the
	// receive FIFO is flushed. This bit resets to 0.

	if (gCurrentState.RX_ENABLE != newState.RX_ENABLE)
	{
		if (newState.RX_ENABLE == 0)
		{
			PrvEmptyQueue (gRxFIFO);
		}
	}


	// ========== TX_ENABLE ==========
	//
	// This bit enables the transmitter block. While this bit is low, the transmitter is disabled and
	// the transmit FIFO is flushed. This bit resets to 0.

	if (gCurrentState.TX_ENABLE != newState.TX_ENABLE)
	{
		if (newState.TX_ENABLE == 0)
		{
			PrvEmptyQueue (gTxFIFO);
		}
	}


	// ========== PARITY_EN ==========
	//
	// This bit controls the parity generator in the transmitter and parity checker in the receiver.
	// When this bit is high, they are enabled. When it is low, they are disabled.
	//
	// ========== ODD_EVEN ==========
	//
	// This bit controls the sense of the parity generator and checker. When this bit is high, odd
	// parity is generated and expected. When this bit is low, even parity is generated and
	// expected. This bit has no function if PARITY EN is low.

	if (newState.PARITY_EN == 0)
	{
		config.fParity = EmTransportSerial::kNoParity;
	}
	else if (newState.ODD_EVEN)
	{
		config.fParity = EmTransportSerial::kOddParity;
	}
	else
	{
		config.fParity = EmTransportSerial::kEvenParity;
	}


	// ========== STOP_BITS =========
	//
	// This bit controls the number of stop bits transmitted after a character. When this bit is high,
	// two stop bits are sent. When this bit is low, one stop bit is sent. This bit has no effect on the
	// receiver, which expects one or more stop bits.

	config.fStopBits = newState.STOP_BITS ? 2 : 1;


	// ========== CHAR8_7 ==========
	//
	// This bit controls the character length. While high, the transmitter and receiver are in 8-bit
	// mode. While low, they are in 7-bit mode. The transmitter then ignores B7 and the receiver
	// sets B7 to 0.

	config.fDataBits = newState.CHAR8_7 ? 8 : 7;


	// ========== RX_RDY_ENABLE ==========
	//
	// When this bit is high, it enables an interrupt when the receiver has at least one data byte in
	// the FIFO. When it is low, this interrupt is disabled.

	if (gCurrentState.RX_RDY_ENABLE != newState.RX_RDY_ENABLE)
	{
		// Nothing to do here.  Interrupt generated elsewhere.
	}


	// ========== DIVIDE ==========
	//
	// These bits control the clock frequency produced by the baud rate generator.
	//
	//		000 = Divide by 1.
	//		001 = Divide by 2.
	//		010 = Divide by 4.
	//		011 = Divide by 8.
	//		100 = Divide by 16.
	//		101 = Divide by 32.
	//		110 = Divide by 64.
	//		111 = Divide by 128.
	//
	// ========== PRESCALER ==========
	//
	// These bits control the division value of the baud generator prescaler. The division value is
	// determined by the following formula:
	//
	//		Prescaler division value = 65 (decimal) - PRESCALER

	// Baud rate is sysClockFreq / preScaler / divider / 16
	//
	// (Using sysClockFreq / (preScaler * divider * 16) might get closer to the
	//  intended baud (as would using floating point), but does it mirror what
	//  the hardware actually does?)

	int32	sysClockFreq = RegsBank::GetSystemClockFrequency ();
	config.fBaud = sysClockFreq / (65 - newState.PRESCALER) / (1 << newState.DIVIDE) / 16;

	// "newRate" is only approximate to within 0.1%. Pin the value to an
	// exact value.
	//
	// !!! What to do if we can't pin to a valid value?

	(void) PrvPinBaud (config.fBaud);


	// ========== IGNORE_CTS ==========
	//
	// When this bit is high, it forces the CTS signal that is presented to the transmitter to always
	// be asserted, which effectively ignores the external pin.

	config.fHwrHandshake = newState.IGNORE_CTS == 0;


	// ========== RTS_CONT ==========
	//
	// This bit selects the function of the RTS pin.
	//
	//		0 = RTS pin is controlled by the RTS bit.
	//		1 = RTS pin is controlled by the receiver FIFO. When the FIFO is full (one slot is
	//			remaining), RTS is negated.
	//
	// ========== RTS ==========
	//
	// This bit controls the RTS pin while the RTS CONT bit is 0.
	//
	//		0 = RTS pin is 1.
	//		1 = RTS pin is 0.

	if (gCurrentState.RTS_CONT != newState.RTS_CONT ||
		gCurrentState.RTS != newState.RTS)
	{
		// Nothing to do here.  These settings are looked at in the cycle code
		// that generates interrupts. !!! TBD
	}


	// ========== UART_ENABLE ==========
	//
	// This bit enables the UART module. When this bit is low, the UART module is disabled and
	// in low-power mode. While this bit is high, the UART module is active. This bit resets to 0.

	// ========== IRDA_ENABLE ==========
	//
	// This bit enables the IrDA interface.
	//
	//		0 = Normal NRZ operation.
	//		1 = IRDA operation.

	if (gCurrentState.UART_ENABLE != newState.UART_ENABLE ||
		gCurrentState.IRDA_ENABLE != newState.IRDA_ENABLE)
	{
		// Nothing to do here.
	}

	// Prepare to update the host serial port.  First, if we need to change
	// host serial ports (because the user made a preferences change, then
	// close the serial port first.  That's the only way to get it to change
	// serial ports.

	Bool	portIsOpen = gSerial->CanRead ();
	if (gNeedPortChange || !RegsBank::GetSerialPortOn ())
	{
		gSerial->Close ();
		portIsOpen = false;
	}

	// Establish the new settings.

	gSerial->SetConfig (config);

	// Now see if the serial port needs to be turned back on.

	if (RegsBank::GetSerialPortOn () && !portIsOpen)
	{
		gSerial->Open ();
	}

	// ========== SEND_BREAK ==========
	//
	// This bit forces the transmitter to immediately send continuous zeros creating a break
	// character.

	if (gCurrentState.SEND_BREAK != newState.SEND_BREAK)
	{
		// Nothing to do right now.  External cycling will send out the breaks.
		// !!! TBD.
	}


	// ========== TX_DATA ==========
	//
	// These bits are the parallel transmit-data inputs. In 7-bit mode, D7 is ignored and in 8-bit
	// mode, all of the bits are used. Data is transmitted LSB first. A new character is transmitted
	// when these bits are written and have passed through the FIFO.

	uint8	charToSend = newState.TX_DATA;

	if (newState.CHAR8_7)
		charToSend &= 0x07F;

#if 1
	if (sendTxData && newState.UART_ENABLE && newState.TX_ENABLE)
	{
		if (newState.LOOP == 0)
		{
			if (gSerial->CanWrite ())		// The host serial port is open
			{
				// With or without hardware handshaking, we'll put data
				// in the FIFO, and let the host's handshaking take care
				// of when the data is removed from the FIFO.

				if (gTxFIFO.GetFree () > 0)		// There's room in the FIFO
				{
					gTxFIFO.Put (newState.TX_DATA);	// so add the data

					// Call TransmitTxFIFO here to send the data we
					// just queued up.  Doing this is important on the Mac in
					// order to send out the data quickly instead of later at
					// idle time.

					TransmitTxFIFO ();
				}
			}
			else								// The host serial port is NOT open
			{
				if (config.fHwrHandshake)	// Reflects the state of the IGNORE_CTS bit.
				{
					// With hardware handshaking, data is sent only when CTS
					// is asserted.  With no host serial port, we define that
					// CTS is never asserted, so the data clogs up the FIFO.

					if (gTxFIFO.GetFree () > 0)			// There's room in the FIFO
					{
						gTxFIFO.Put (newState.TX_DATA);	// so add the data

						// Serial port is closed, so don't call Platform::TransmitTxFIFO.
					}
				}
				else
				{
					// With no hardware handshaking, data is sent whenever it's
					// ready.  With nowhere to go, we can drop it on the floor.
				}
			}
		}
		else	// We're in loopback mode.
		{
			if (gRxFIFO.GetFree () > 0)
			{
				gRxFIFO.Put (newState.TX_DATA);
			}
		}
	}
#else
	if (sendTxData && newState.UART_ENABLE && newState.TX_ENABLE)
	{
		if (newState.LOOP == 0)
		{
			// See if we've got any place to go.

			Bool	haveHost;

			if (newState.IRDA_ENABLE == 0)
			{
				haveHost = gSerial->CanWrite ();
			}
			else
			{
				haveHost = false;
			}

			// Try adding our transmit byte to the FIFO and sending it.

			if (gTxFIFO.GetFree () > 0)						// There's room in the FIFO
			{
				if (haveHost)								// The host service is open
				{
					// With or without hardware handshaking, we'll put data
					// in the FIFO, and let the host's handshaking take care
					// of when the data is removed from the FIFO.

					gTxFIFO.Put (charToSend);				// so add the data
					TransmitTxFIFO (/* forSerialOrIR */);	// Send it out.
				}
				else if (config.fHwrHandshake)	// Reflects the state of the IGNORE_CTS bit.
				{
					// With hardware handshaking, data is sent only when CTS
					// is asserted.  With no host serial port, we define that
					// CTS is never asserted, so the data clogs up the FIFO.

					gTxFIFO.Put (charToSend);
				}
				else
				{
					// With no hardware handshaking, data is sent whenever it's
					// ready.  With nowhere to go, we can drop it on the floor.
				}
			}
			else
			{
				// With no room in the FIFO, just drop the data.
			}
		}
		else	// We're in loopback mode.
		{
			if (gRxFIFO.GetFree () > 0)
			{
				gRxFIFO.Put (charToSend);
			}
			else
			{
				// With no room in the FIFO, just drop the data.
			}
		}
	}
#endif

	// ========== LOOP ==========
	//
	// This bit controls loopback for system testing purposes. When this bit is high, the receiver
	// input is internally connected to the transmitter and ignores the RXD pin. The TXD pin is
	// unaffected by this bit.

	if (gCurrentState.LOOP != newState.LOOP)
	{
		// Nothing to do here.  This bit is examined in the code
		// that reacts to TX_DATA.
	}


	// Update the state in case any of the above operations have side-effects.

	UpdateState (newState, false);


	// Remember this for next time.

	gCurrentState = newState;
}


/***********************************************************************
 *
 * FUNCTION:	UART::UpdateState
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void UART::UpdateState (State& state, Bool refreshRxData)
{
	assert (gCurrentState.UART_TYPE == state.UART_TYPE);

	// Update the RxFIFO if there's been any buffered data.

	ReceiveRxFIFO ();

	// === RX_FIFO_FULL ===
	//
	// This read-only bit indicates that the receiver FIFO is full and may generate an overrun. This
	// bit generates a maskable interrupt.
	//
	// Further, from the overview section of the manual:
	//
	// If your software has a short interrupt
	// latency time, the FIFO FULL interrupt in the Receiver register can be enabled. The FIFO has
	// one remaining space available when this interrupt is generated.

	state.RX_FIFO_FULL = gRxFIFO.GetFree () == 0;	// Interrupt generated in Foo::Cycle.


	// === RX_FIFO_HALF ===
	//
	// This read-only bit indicates that the receiver FIFO has four or fewer slots remaining in the
	// FIFO. This bit generates a maskable interrupt.

	state.RX_FIFO_HALF = gRxFIFO.GetFree () <= 4;	// Interrupt generated in Foo::Cycle.


	// === DATA_READY ===
	//
	// This read-only bit indicates that at least one byte is present in the receive FIFO. The
	// character bits are valid only while this bit is set. This bit generates a maskable interrupt.

	state.DATA_READY = gRxFIFO.GetUsed () > 0;	// Interrupt generated in Foo::Cycle.


	// === OLD_DATA ===	// 68EZ328 only
	//
	// This read-only bit indicates that data in the FIFO is older than 30 bit times. It is useful in
	// situations where the FIFO FULL or FIFO HALF interrupts are used. If there is data in the
	// FIFO, but below the interrupt threshold, a maskable interrupt can be generated to alert the
	// software that unread data is present. This bit clears when the character bits are read.

	// Not supported right now.


	// === OVRUN ===
	//
	// When this read-only bit is high, it indicates that the receiver overwrote data in the FIFO. The
	// character with this bit set is valid, but at least one previous character was lost. In normal
	// circumstances, this bit should never be set. It indicates that your software is not keeping up
	// with the incoming data rate. This bit is updated and valid for each received character.

	// !!! TBD


	// === FRAME_ERROR ===
	//
	// While high, this read-only bit indicates that the current character had a framing error
	// (missing stop bit), indicating the possibility of corrupted data. This bit is updated for each
	// character read from the FIFO.

	// !!! TBD


	// === BREAK ===
	//
	// When this read-only bit is high, it indicates that the current character was detected as a
	// BREAK. The data bits are all 0 and the stop bit was also 0. The FRAME ERROR bit will
	// always be set when this bit is set. If odd parity is selected, PARITY ERROR will also be set
	// along with this bit. This bit is updated and valid with each character read from the FIFO.

	// !!! TBD

	// === PARITY_ERROR ===
	//
	// When this read-only bit is high, it indicates that the current character was detected with a
	// parity error, indicating the possibility of corrupted data. This bit is updated and valid with
	// each character read from the FIFO. While parity is disabled, this bit always reads zero.

	// !!! TBD

	// === RX_DATA ===
	//
	// These read-only bits are the top receive character in the FIFO. They have no meaning if the
	// DATA READY bit is 0. In 7-bit mode, the MSB is forced to 0 and in 8-bit mode, all bits are
	// active.

		// !!! Should probably test against RTS, too.  Actually, that test should
		// happen before putting the byte into the Rx FIFO.

	if (state.DATA_READY && state.UART_ENABLE && state.RX_ENABLE && refreshRxData)
	{
		state.RX_DATA = gRxFIFO.Get ();

		if (state.CHAR8_7 == 0)
		{
			state.RX_DATA &= 0x07F;
		}

		PRINTF ("Serial Emulation: Put 0x%02X into RX_DATA.", (uint32) (uint8) state.RX_DATA);
	}


	// === TX_FIFO_EMPTY ===
	//
	// This read-only bit indicates that the transmit FIFO is empty. This bit generates a maskable
	// interrupt.

	state.TX_FIFO_EMPTY = gTxFIFO.GetUsed () == 0;	// Interrupt generated in Foo::Cycle.


	// === TX_FIFO_HALF ===
	//
	// This read-only bit indicates that the transmit FIFO is less than half full. This bit generates a
	// maskable interrupt.

	state.TX_FIFO_HALF = gTxFIFO.GetUsed () < 4;	// Interrupt generated in Foo::Cycle.


	// === TX_AVAIL ===
	//
	// This read-only bit indicates that the transmit FIFO has at least one slot available for data.
	// This bit generates a maskable interrupt.

	state.TX_AVAIL = gTxFIFO.GetFree () > 0;	// Interrupt generated in Foo::Cycle.


	// === BUSY ===	// 68EZ328 only
	//
	// When this bit is high, it indicates that the transmitter is busy sending a character. This signal
	// is asserted while the transmitter state machine is not idle or the FIFO has data in it.

	// Not supported right now.


	// === CTS_STATUS ===
	//
	// This bit indicates the current status of the CTS pin. A "snapshot" of the pin is taken
	// immediately before this bit is presented to the data bus. While the IGNORE CTS bit is high,
	// this bit can serve as a general-purpose input.
	//
	// Note that this pin is ACTIVE LOW!  That's why the Boolean expression is negated below.
	//
	// For now, say that it's clear to send if the FIFO is empty
	//
	// !!! TBD - could be better?

	state.CTS_STATUS = !(gTxFIFO.GetUsed () == 0);


	// === CTS_DELTA ===
	//
	// When this bit is high, it indicates that the CTS pin changed state and generates a maskable
	// interrupt. The current state of the CTS pin is available on the CTS STATUS bit. You can
	// generate an immediate interrupt by setting this bit high. The CTS interrupt is cleared by
	// writing 0 to this bit.

	// Not supported right now.


	// Remember this for next time.

	gCurrentState = state;
}


/***********************************************************************
 *
 * FUNCTION:	UART::TransmitTxFIFO
 *
 * DESCRIPTION:	Transmit any bytes in the Tx FIFO out the serial port.
 *				Assumes that the serial port is open.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void UART::TransmitTxFIFO (void)
{
	assert (gSerial);

	if (gSerial->CanWrite ())
	{
		// Write out any outgoing bytes.

		ErrCode	err = errNone;
		char	buffer[8];
		long	spaceInTxFIFO = gTxFIFO.GetUsed ();

		if (spaceInTxFIFO > 0)
		{
			for (long ii = 0; ii < spaceInTxFIFO; ++ii)
			{
				buffer[ii] = gTxFIFO.Get ();
			}

			if (LogSerialData ())
				LogAppendData (buffer, spaceInTxFIFO, "Serial Emulation: Transmitted data:");
			else
				PRINTF ("Serial Emulation: Transmitted %ld serial bytes.", spaceInTxFIFO);

			err = gSerial->Write (spaceInTxFIFO, buffer);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	UART::ReceiveRxFIFO
 *
 * DESCRIPTION:	Fills up the Rx FIFO with as many bytes as it can from
 *				the host serial port.  Assumes that the serial port is
 *				open.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void UART::ReceiveRxFIFO (void)
{
	assert (gSerial);

	if (gSerial->CanRead ())
	{
		// Buffer up any incoming bytes.

		ErrCode	err = errNone;
		char	buffer[8];
		long	spaceInRxFIFO = gRxFIFO.GetFree ();

		// See how many bytes are waiting.

		long	bytesToBuffer = gSerial->BytesInBuffer ();

		if (bytesToBuffer)
		{
			// See if we have that much room in the FIFO. If not, limit our read
			// to that many bytes.

			if (bytesToBuffer > spaceInRxFIFO)
			{
				bytesToBuffer = spaceInRxFIFO;
			}

			// If there are still bytes to be read, read them in and insert them
			// into the Rx FIFO.  If the buffer was previously empty, Hardware::Cycle
			// will notice that there are now bytes in there and update the
			// UART registers.
			//
			// !!! TBD: Hardware::Cycle is called after every opcode execution.
			// Since it is very rare that serial bytes are coming in, the overhead
			// incurred for checking to see if there are serial bytes does not pay
			// off.  We may want to figure out a way to update the UART registers
			// here, and take the check out of Hardware::Cycle.  I'd do that here,
			// but (a) Hardware::Cycle takes a "sleeping" parameter that I'd have
			// to determine here, and (b) the similar solution on Windows is more
			// difficult, as I'd have to stop the CPU thread, update the registers,
			// and restart the thread.

			if (bytesToBuffer > 0)
			{
				err = gSerial->Read (bytesToBuffer, buffer);

				if (err == errNone)
				{
					if (LogSerialData ())
						LogAppendData (buffer, bytesToBuffer, "Serial Emulation: Received data:");
					else
						PRINTF ("Serial Emulation: Received %ld serial bytes.", bytesToBuffer);

					for (long ii = 0; ii < bytesToBuffer; ++ii)
					{
						gRxFIFO.Put (buffer[ii]);
					}	// end loop that puts bytes into FIFO
				}	// end no-error-from-FSRead
			}	// end have-bytes-to-buffer
		}	// end no-error-from-SerGetBuf
	}	// end is-serial-port-open
}


/***********************************************************************
 *
 * FUNCTION:	PrvPinBaud
 *
 * DESCRIPTION:	Pins the given baud value to the test baud value if the
 *				former is sufficiently close to the latter.
 *
 * PARAMETERS:	newBaud - the value to possibly alter.
 *
 *				testBaud - the value to pin to.
 *
 * RETURNED:	newBaud is changed in place.  If it is changed, the
 *				function returns true.  Otherwise, it returns false.
 *
 ***********************************************************************/

Bool PrvPinBaud (EmTransportSerial::Baud& newBaud)
{
	Bool	pinned = false;

	if (!pinned)	pinned = PrvPinBaud (newBaud, 150);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 300);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 600);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 1200);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 1800);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 2400);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 3600);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 4800);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 7200);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 9600);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 14400);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 19200);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 28800);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 38400);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 57600);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 115200);
	if (!pinned)	pinned = PrvPinBaud (newBaud, 230400);

	return pinned;
}


/***********************************************************************
 *
 * FUNCTION:	PrvPinBaud
 *
 * DESCRIPTION:	Pins the given baud value to the test baud value if the
 *				former is sufficiently close to the latter.
 *
 * PARAMETERS:	newBaud - the value to possibly alter.
 *
 *				testBaud - the value to pin to.
 *
 * RETURNED:	newBaud is changed in place.  If it is changed, the
 *				function returns true.  Otherwise, it returns false.
 *
 ***********************************************************************/

Bool PrvPinBaud (EmTransportSerial::Baud& newBaud, EmTransportSerial::Baud testBaud)
{
	// Pin to within 2%.  The Dragonball reference says that the uBaud
	// register should be accurate to within 0.1%, but let's give it
	// some slop.

	if (newBaud > (testBaud - (testBaud / 50)) &&
		newBaud < (testBaud + (testBaud / 50)))
	{
		newBaud = testBaud;
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	PrvEmptyQueue
 *
 * DESCRIPTION:	Remove all contents of the given queue.
 *
 * PARAMETERS:	q - TByteQueue to erase.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void PrvEmptyQueue (TByteQueue& q)
{
	while (q.GetUsed () > 0)
	{
		q.Get ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvPrefsChanged
 *
 * DESCRIPTION:	Called when the preference we're interested in has
 *				changed.
 *
 * PARAMETERS:	key - the key of the preference that changed.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void PrvPrefsChanged (PrefKeyType)
{
	gNeedPortChange = true;
}
