/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Document.h"

#include "CPU_MT.h"				// CPU
#include "CPU_REG.h"			// Emulator::, others
#include "DebugMgr.h"			// Debug::HandleNewPacket
#include "EmDlg.h"				// DoHordeNew
#include "EmStreamFile.h"		// EmStreamFile, kCreateOrEraseForUpdate
#include "RAM_ROM.h"			// Memory::
#include "Hordes.h"				// Hordes::New...
#include "PreferenceMgr.h"		// Preference (kPrefKeyGremlinInfo)
#include "SocketMessaging.h"
#include "TrapPatches.h"		// Patches::fgUIInitialized

#include <limits.h>

#define down fl_down			// "down" defined both in WindowNew.h and Fl/Enumerations.h
#include <FL/Fl.H>

#include "fltk_Dialog.h"		// showNewGremlinWindow


// -----------------------------------------------------------------------------
// constructor(s) / destructor
// -----------------------------------------------------------------------------

Document::Document (void) :
	_cpu (NULL),
	_currentConfiguration ()
{
	_cpu = new CPU;
	_cpu->CreateThread ();
}

Document::~Document (void)
{
	propertyChanged (documentClosing);

	if (_cpu != NULL)
	{
		Platform::GCW_Close ();

		_cpu->DestroyThread ();
		delete _cpu;

		Emulator::Dispose ();
	}
}


// -----------------------------------------------------------------------------
// public methods
// -----------------------------------------------------------------------------

ErrCode Document::create (const Configuration& cfg)
{
	// Try initializing the emulator with all this information.
	ErrCode	result = errNone;

	try
	{
		_currentConfiguration = cfg;
		Emulator::Initialize (_currentConfiguration);
	}
	catch (ErrCode e)
	{
		result = e;
	}

	if (result == errNone)
	{
		propertyChanged (documentOpening);
	}

	return result;
}


// -----------------------------------------------------------------------------
// read the document state from the given file, then get the CPU going with it.

ErrCode Document::open (const EmFileRef& ramFileRef)
{
	ErrCode	result = errNone;

	try
	{
		Emulator::Load (ramFileRef, true, true, true);
		Hordes::PostLoad ();
	}
	catch (ErrCode e)
	{
		result = e;
	}

	if (result == errNone)
	{
		// If we loaded a session file with a Gremlin running, turn
		// it off, but let the user turn it back on.

		if (Hordes::IsOn ())
		{
			Hordes::Stop ();
			Platform::GCW_Open ();
		}

		_currentRAMRef = ramFileRef;
		propertyChanged (documentOpening);
	}

	return result;
}

// -----------------------------------------------------------------------------
// start the emulator running

void Document::emulatorStart (void)
{
	_cpu->StartThread ();
	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// stop the emulator

void Document::emulatorStop (void)
{
	Bool dummy;
	_cpu->StopThread (dummy, dummy, kStopOnADime);
	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// reset the emulator

void Document::emulatorReset (void)
{
	Bool dummy;
	_cpu->StopThread (dummy, dummy, kStopOnADime);

	Emulator::Reset ();

	_cpu->StartThread ();

	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// hotsync helper function

void _hotSyncUp (void* data)
{
	CPUStopper stopper (kStopOnADime);
	if (!stopper.Stopped())
		return;

	Hardware::HotSyncEvent(false);
}

// -----------------------------------------------------------------------------
// Start hotsync in the emulator

void Document::hotsync (void)
{
	CPUStopper stopper (kStopOnADime);
	if (!stopper.Stopped())
		return;

	Hardware::HotSyncEvent(true);

	// install a timer to have the hotsync button released in 1 second:
	Fl::add_timeout (1.0, _hotSyncUp, this);
}

// -----------------------------------------------------------------------------
// Pass a pen event to the emulator

void Document::pen (bool down, int x, int y)
{
	CPUStopper stopper (kStopOnADime);
	if (!stopper.Stopped())
		return;

	Hardware::PenEvent (down, x, y);
}

// -----------------------------------------------------------------------------
// Pass a button event to the emulator

void Document::button (bool down, SkinElementType button)
{
	CPUStopper stopper (kStopOnADime);
	if (!stopper.Stopped())
		return;

	Hardware::ButtonEvent (down, button);
}

// -----------------------------------------------------------------------------
// Pass a key event to the emulator

void Document::putkey (unsigned char key)
{
	CPUStopper stopper (kStopOnADime);
	if (!stopper.Stopped())
		return;

	Hardware::KeyboardEvent(key);
}

// -----------------------------------------------------------------------------
// load the given application file

void Document::loadApplication (const EmFileRef& appFile)
{
	EmFileRefList fileList;
	fileList.push_back (appFile);

	::LoadPalmFileList (fileList);
}

// -----------------------------------------------------------------------------
// export the application file

void Document::exportDatabase (const DatabaseInfo& db, const EmFileRef& exportFileRef)
{
	EmStreamFile		exportFile (exportFileRef, kCreateOrEraseForUpdate,
								kFileCreatorEmulator, kFileTypePalmDB);

	::SavePalmFile (exportFile, db.cardNo, db.dbName);
}

// -----------------------------------------------------------------------------
// save the memory image to current file

void Document::saveMemory (void)
{
	saveRAMImage (_currentRAMRef, false);
	propertyChanged ();
}

// -----------------------------------------------------------------------------
// save the memory image to current file

void Document::saveMemoryAs (EmFileRef& dest)
{
	saveRAMImage (dest, false);
	propertyChanged ();
}

// -----------------------------------------------------------------------------
// save the memory image to the given file ref

void Document::saveScreen (EmFileRef& dest)
{
	//! TODO
	//_cpu->saveScreenShot( dest );
}

// -----------------------------------------------------------------------------
// add the given callback to the list of PropertyChangedHandlers

void Document::addPropertyChangedHandler (PropertyChangedHandler callback, void* param)
{
	PropertyChangedHandlerSpec	spec;
	spec.func = callback;
	spec.param = param;

	_handlers.push_back (spec);
}

// -----------------------------------------------------------------------------
// remove the given callback from the list of PropertyChangedHandlers

void Document::removePropertyChangedHandler (PropertyChangedHandler callback)
{
	list<PropertyChangedHandlerSpec>::iterator	iter;

	for (iter = _handlers.begin (); iter != _handlers.end (); ++iter)
	{
		if (iter->func == callback)
		{
			_handlers.erase (iter);
			return;
		}
	}
}

// -----------------------------------------------------------------------------
// gremlins control functions

void Document::gremlinsNew (void)
{
	EmDlg::DoHordeNew ();
}

void Document::gremlinsStep (void)
{
	Hordes::Step ();
}

void Document::gremlinsResume (void)
{
	Hordes::Resume ();
}

void Document::gremlinsStop()
{
	Hordes::Stop ();
}

// -----------------------------------------------------------------------------
// private methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// save the ram image to the file

void Document::saveRAMImage (const EmFileRef& ramFileRef, Bool forAutoSave)
{
	// See if there's a given name we should use.

	if (!ramFileRef.IsSpecified())
	{
		return;
	}

	try
	{
		Emulator::Save (ramFileRef, !forAutoSave, !forAutoSave, !forAutoSave);
	}
	catch (...)
	{
	}

	_currentRAMRef = ramFileRef;
}

// -----------------------------------------------------------------------------
// broadcast changed notices to all propertyChangedHandlers

void Document::propertyChanged (int property)
{
	list<PropertyChangedHandlerSpec>::iterator	iter;

	for (iter = _handlers.begin (); iter != _handlers.end (); ++iter)
	{
		(iter->func) (property, iter->param);
	}
}
