/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#ifndef __DOCUMENT_H__
#define __DOCUMENT_H__

#include "EmulatorTypes.h"		// Configuration
#include "Platform.h"			// Platform::AutoSaveState
#include "Skins.h"				// SkinElementType

#include <list>					// list

// -----------------------------------------------------------------------------
// class Document
//
// The class encapsulates all of the DATA pertinent to an emulation session.
// Some of this data is persistent, and this class provides methods for saving
// that data.
// In the future, it would be nice if each of the data member of this class
// (in particular the CPU) would be thread safe, so that I could have several
// Documents open at once.
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// local types
// -----------------------------------------------------------------------------

typedef void (*PropertyChangedHandler)( int property, void *param );

struct PropertyChangedHandlerSpec
{
	PropertyChangedHandler	func;
	void*					param;
};

class Document
{
public:

	// -----------------------------------------------------------------------------
	// public constants
	// -----------------------------------------------------------------------------

	enum
	{
		unspecifiedChanged = 0,
		documentOpening,
		documentClosing,
		cpuStatusChanged
	};

	// -----------------------------------------------------------------------------
	// constructor / destructor
	// -----------------------------------------------------------------------------

	Document(void);
	virtual ~Document (void);

	// -----------------------------------------------------------------------------
	// public methods
	// -----------------------------------------------------------------------------

	ErrCode create (const Configuration&);
	ErrCode open (const EmFileRef&);

	void emulatorStart( void );
	void emulatorStop( void );
	void emulatorReset( void );
	void loadApplication( const EmFileRef& appFile );
	void exportDatabase (const DatabaseInfo&, const EmFileRef&);
	void saveMemory();
	void saveMemoryAs( EmFileRef& dest );
	void saveScreen( EmFileRef& dest );

	void pollDebugPort();

	void hotsync();
	void pen( bool down, int x, int y );
	void button( bool down, SkinElementType button );
	void putkey( unsigned char key );

	EmFileRef getCurrentRAMRef() { return _currentRAMRef; }

	void addPropertyChangedHandler( PropertyChangedHandler callback, void *param );
	void removePropertyChangedHandler( PropertyChangedHandler callback );

	bool initializedOk() { return _cpu != NULL; }
	Configuration& getCurrentConfiguration() { return _currentConfiguration; }

	void gremlinsNew();
	void gremlinsStep();
	void gremlinsResume();
	void gremlinsStop();

	CPU* getCPU() { return _cpu; } // Called by Platform::GetCPU.

private:

	// -----------------------------------------------------------------------------
	// private methods
	// -----------------------------------------------------------------------------

	void propertyChanged( int property = unspecifiedChanged );

	EmFileRef findPairedROMFile( EmStreamFile& docFile );
	void saveRAMImage( const EmFileRef&, Bool forAutoSave );

	// -----------------------------------------------------------------------------
	// private friends
	// -----------------------------------------------------------------------------

	friend void _hotSyncUp( void *data );
	friend void Platform::AutoSaveState (const EmFileRef&);	// saveRAMImage

	// -----------------------------------------------------------------------------
	// private data members
	// -----------------------------------------------------------------------------

	CPU*					_cpu;
	Configuration			_currentConfiguration;
	EmFileRef				_currentRAMRef;
	list<PropertyChangedHandlerSpec>	_handlers;
};


#endif /* __DOCUMENT_H__ */
